/** ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * $Id: Conversacion.java,v 1.1 2006/11/09 03:10:48 f-vela Exp $
 * Universidad de los Andes (Bogot - Colombia)
 * Departamento de Ingeniera de Sistemas y Computacin 
 * Licenciado bajo el esquema Academic Free License versin 2.1
 *
 * Proyecto Cupi2 (http://cupi2.uniandes.edu.co)
 * Ejercicio: n12_messengerAmigos
 * Autor: Mario Snchez - 21/04/2006
 * MODIFICADO: Silvia Takahashi - Verificacion de Mensajes 13/03/2009
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */

package uniandes.cupi2.messengerAmigos.cliente;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.ServerSocket;
import java.net.Socket;

import uniandes.cupi2.messengerAmigos.interfazCliente.IVentanaConversacion;
import uniandes.cupi2.messengerAmigos.verificador.*;


/**
 * Esta clase representa una conversacin que se est llevando a cabo entre dos usuarios
 */
public class Conversacion
{
    // -----------------------------------------------------------------
    // Constantes
    // -----------------------------------------------------------------

    /**
     * Mensaje que se usa para indicar el inicio de un mensaje enviado dentro de una conversacin
     */
    public static final String MENSAJE = "MENSAJE";

    /**
     * Mensaje que se usa para indicar que se quiere terminar la conversacin
     */
    public static final String TERMINAR = "TERMINAR";

    /**
     * Mensaje que se usa para indicar que se acept el fin de la conversacin
     */
    public static final String CONVERSACION_TERMINADA = "CONVERSACION_TERMINADA";
    
    /**
     * Mensaje que se usa para indicar que se cambia el chequeo
     */
    public static final String CAMBIAR_CHECK = "CAMBIAR_CHECK";

    // -----------------------------------------------------------------
    // Atributos
    // -----------------------------------------------------------------

    /**
     * El nombre del amigo con el que se va a llevar a cabo la conversacin
     */
    private String nombreAmigo;

    /**
     * El cliente en el cual se cre esta conversacin
     */
    private IClienteMessenger clienteAmigos;

    /**
     * Indica si la conversacin ya termin
     */
    private boolean conversacionTerminada;

    /**
     * La ventana donde se muestran los mensajes de esta conversacin
     */
    private IVentanaConversacion ventanaConversacion;

    /**
     * Es el socket a travs del cual se lleva a cabo la comunicacin con el amigo
     */
    private Socket socketAmigo;

    /**
     * Es el stream usuado para leer los mensajes enviados por el amigo
     */
    private BufferedReader inAmigo;

    /**
     * Es el stream usado para enviar mensajes al amigo
     */
    private PrintWriter outAmigo;

    /**
     * Es el verificador de mensajes
     */  
	
	private VerificadorMensajes checkMsg = new VerificadorMensajes();

    
    // -----------------------------------------------------------------
    // Constructores
    // -----------------------------------------------------------------

    /**
     * Crea una nueva conversacin. <br>
     * El usuario local fue el que inicio la conversacin, as que se debe crear un socket y esperar a que el amigo se conecte.
     * @param cliente Es el cliente en el cual se cre esta conversacin - cliente != null
     * @param puerto Es el puerto en el cual se debe esperar la conexin del amigo
     * @param amigo El nombre del amigo con el que se va a realizar la conversacin - amigo != null
     * @throws IOException Se lanza esta excepcin si hay problemas estableciendo la comunicacin con el amigo
     */
    public Conversacion( IClienteMessenger cliente, int puerto, String amigo ) throws IOException
    {
        conversacionTerminada = false;
        clienteAmigos = cliente;
        nombreAmigo = amigo;

        // Esperar la conexin del amigo
        ServerSocket socket = new ServerSocket( puerto );
        socketAmigo = socket.accept( );

        outAmigo = new PrintWriter( socketAmigo.getOutputStream( ), true );
        inAmigo = new BufferedReader( new InputStreamReader( socketAmigo.getInputStream( ) ) );
        
        socket.close( );

        // Iniciar el hilo que recibir los mensajes del amigo
        ThreadRecibirMensajesConversacion hilo = new ThreadRecibirMensajesConversacion( this, inAmigo );
        hilo.start( );
    }

    /**
     * Crea una nueva conversacin <br>
     * El amigo fue el que inici la conversacin, as que es necesario abrir una conexin hasta el cliente del amigo.
     * @param cliente Es el cliente en el cual se cre esta conversacin - cliente != null
     * @param amigo Es la informacin que se tiene sobre el amigo que inici la conversacin - amigo != null
     * @throws IOException Se lanza esta excepcin si hay problemas estableciendo la comunicacin con el amigo
     */
    public Conversacion( IClienteMessenger cliente, Usuario amigo ) throws IOException
    {
        conversacionTerminada = false;
        clienteAmigos = cliente;
        nombreAmigo = amigo.darNombre( );

        // Abrir la conexin
        String direccion = amigo.darDireccionIp( );
        int puerto = amigo.darPuerto( );
        socketAmigo = new Socket( direccion, puerto );

        outAmigo = new PrintWriter( socketAmigo.getOutputStream( ), true );
        inAmigo = new BufferedReader( new InputStreamReader( socketAmigo.getInputStream( ) ) );

        // Iniciar el hilo que recibir los mensajes del amigo
        ThreadRecibirMensajesConversacion hilo = new ThreadRecibirMensajesConversacion( this, inAmigo );
        hilo.start( );
    }

    // -----------------------------------------------------------------
    // Mtodos
    // -----------------------------------------------------------------

    /**
     * Indica si la conversacin est conectada
     * @return true si ya est conectada
     */
    public boolean estaConectada( )
    {
        return socketAmigo != null;
    }

    /**
     * Establece la ventana asociada a esta conversacin, en la cual se deben mostrar tanto los mensajes enviados por el usuario local como los mensajes enviados por el amigo.
     * @param ventana La ventana en la que se publicarn los mensajes - ventana != null
     */
    public void cambiarVentanaConversacion( IVentanaConversacion ventana )
    {
        ventanaConversacion = ventana;
    }

    /**
     * Indica si la conversacin ya fue marcada como terminada
     * @return conversacionTerminada
     */
    public boolean conversacionTerminada( )
    {
        return conversacionTerminada;
    }

    /**
     * Retorna el nombre del amigo con el que se est llevando a cabo la conversacin
     * @return nombreAmigo
     */
    public String darNombreAmigo( )
    {
        return nombreAmigo;
    }

    /**
     * Enva al amigo un mensaje y lo muestra en la ventana de la conversacin
     * Si el mensaje no es verificado por el verificador actual, no lo enva
     * @param mensaje El mensaje que ser enviado al amigo - mensaje != null
     */
    public void enviarMensaje( String mensaje)
    {
    	MensajeConRespuesta resp = new MensajeConRespuesta("ERROR!! "+mensaje,false);
    	mostrarMensajeEnVentana( clienteAmigos.darNombreUsuario( ), mensaje );
    		
    	resp = checkMsg.traducir(mensaje);
    	if (resp.esAceptado())  
    		outAmigo.println( MENSAJE + "-->" + resp.getMensaje() );
    	else
    		mostrarMensajeEnVentana( clienteAmigos.darNombreUsuario( ), "  MENSAJE NO ENVIADO!!" + resp.getMensaje());
    }
    

    /**
     * Muestra en la ventana de la conversacion un mensaje enviado por el amigo
     * si el verificador de mensajes activo tiene un decodificador,
     * lo decodifica y muestra el mensaje decodificado
     * @param mensaje El mensaje que fue recibido - mensaje != null
     */
    public void mostrarMensajeRecibido( String mensaje )
    {
    	String actualMsg = mensaje.substring(10);
    	MensajeConRespuesta resp = new MensajeConRespuesta(actualMsg,false);
    	
    	mostrarMensajeEnVentana( nombreAmigo, actualMsg );
    	
    	if (checkMsg.hayDeco()) {
    		resp = checkMsg.deCode(actualMsg);	
    	    if (resp.esAceptado())	
    	    	mostrarMensajeEnVentana("Traduccion de: "+nombreAmigo, resp.getMensaje() );
    	    else 
    	    	mostrarMensajeEnVentana("ERROR TRADUCIENDO: "+nombreAmigo, resp.getMensaje() );
    	}
    }

    /**
     * Muestra un mensaje en la ventana de la conversacin
     * @param autor El autor del mensaje - autor != null
     * @param mensaje El mensaje que se debe mostrar - mensaje != null
     */
    private void mostrarMensajeEnVentana( String autor, String mensaje )
    {
        if( ventanaConversacion != null )
        {
            ventanaConversacion.publicarMensaje( autor + " -> " + mensaje );
        }
    }

    
    /**
     * Envia un mensaje al amigo diciendo que se quiere terminar la conversacin
     */
    public void terminar( )
    {
        outAmigo.println( TERMINAR );
    }

    /**
     * Enva un mensaje al amigo confirmando que la conversacin ser terminada
     */
    public void enviarConfirmacionTerminarConversacion( )
    {
        outAmigo.println( CONVERSACION_TERMINADA );
    }

  /**
  * cambia el tipo de chequeo a checkType
  * Este mtodo es invocado desde la interfaz
  */
   
    public void cambiarCheck(int checkType) {
    	mostrarMensajeEnVentana( "SISTEMA:   cambiando verif "+checkType,"Local ");
    	checkMsg.setCheck(checkType);
    	outAmigo.println(CAMBIAR_CHECK+checkType);
    }
    
    /**
     * cambia el tipo de chequeo a checkType  por orden del otro
     * DEbe informar a la interfaz para que este lo cambie.
     */
   
    public void cambiarCheckRemote(String m) {
    	int index = Integer.valueOf(m.substring(13));
    	checkMsg.setCheck(index);
    	mostrarMensajeEnVentana( "SISTEMA:   cambiando verif REMOTE "+ index, " Remote");
    	if (ventanaConversacion == null) {
    		System.out.println(m);
    	}
    	else {
    		ventanaConversacion.changeCheck(index);
    	}
    	
    }
    
    /**
     * Termina la conversacion, cerrando la conexin con el amigo y la ventana. <br>
     * <b>pre: </b> No hay nadie leyendo o escribiendo en el socket
     */
    
    public void terminarConversacion( )
    {
        try
        {
            conversacionTerminada = true;

            inAmigo.close( );
            outAmigo.close( );

            socketAmigo.close( );
        }
        catch( IOException e )
        {
            e.printStackTrace( );
        }
        finally
        {
            if( ventanaConversacion != null )
                ventanaConversacion.dispose( );
        }
    }
    
 
}