/**~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * $Id: IClienteMessenger.java,v 1.3 2007/04/13 14:25:45 carl-veg Exp $ 
 * Universidad de los Andes (Bogot - Colombia)
 * Departamento de Ingeniera de Sistemas y Computacin 
 * Todos los derechos reservados 2005 
 *
 * Proyecto Cupi2 
 * Ejercicio: n12_messengerAmigos 
 * Autor: Mario Snchez - 2/05/2006 
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ 
 */

package uniandes.cupi2.messengerAmigos.cliente;

import java.io.IOException;


/**
 * Esta interfaz define los servicios que debe ofrecer el cliente para el messenger
 */
public interface IClienteMessenger
{
    // -----------------------------------------------------------------
    // Constantes
    // -----------------------------------------------------------------

    /**
     * Mensaje que se enva cuando un cliente se va a conectar
     */
    static final String LOGIN = "LOGIN";

    /**
     * Mensaje que se enva cuando un cliente se va a desconectar
     */
    static final String LOGOUT = "LOGOUT";

    /**
     * Mensaje que enva el servidor a un cliente cuando lo desconecta
     */
    static final String DESCONEXION = "DESCONEXION";

    /**
     * Mensaje que se usa para indicar que un amigo est conectado
     */
    static final String ONLINE = "ONLINE";

    /**
     * Mensaje que se usa para indicar que un amigo est desconectado
     */
    static final String OFFLINE = "OFFLINE";

    /**
     * Mensaje que se usa para agregar un amigo
     */
    static final String AGREGAR_AMIGO = "AGREGAR_AMIGO";

    /**
     * Mensaje que se usa para que el servidor le indique a un cliente que va a iniciarse una conversacin
     */
    static final String INICIO_CHARLA = "INCIO_CHARLA";

    /**
     * Mensaje que se usa para que un cliente le indique al servidor que va a iniciar una charla y que necesita ponerse en contacto con un amigo
     */
    static final String CONVERSACION = "CONVERSACION";

    /**
     * Indica si el cliente est conectado al servidor
     * @return Retorna true si el cliente est conectado
     */
    boolean estaConectado( );

    /**
     * Retorna el nombre del usuario conectado
     * @return nombreUsuario
     */
    String darNombreUsuario( );

    /**
     * Conecta el cliente al servidor y deja la aplicacin lista para enviar y recibir mensajes. <br>
     * Despus de conectarse al servidor se enva la informacin del usuario (nombre y direccin ip) y se crea un hilo especial para leer los mensajes provenientes del
     * servidor usando una instancia de la clase ThreadRecibirMensajesServidor.
     * @param usuario El nombre del usuario local que se va a conectar al servidor - usuario != null && usuario != ""
     * @throws IOException Se lanza esta excepcin si hay problemas en la comunicacin.
     */
    void conectar( String usuario ) throws IOException;

    /**
     * Enva un mensaje de LOGOUT al servidor. Este a su vez deber cerrar la conexin con el cliente y enviar un mensaje de desconexin (DESCONEXION) para que, cuando sea
     * recibido por el ThreadRecibirMensajesServidor, se inicie la desconexin del lado del cliente.
     */
    void enviarDesconexion( );

    /**
     * Cierra todas las conversaciones que estn abiertas actualmente, limpia la lista de amigos y desconecta al cliente del servidor. <br>
     * <b>pre: </b>No hay nadie escribiendo ni leyendo del socket que comunica al cliente con el servidor.
     * @throws IOException Se lanza esta excepcin si hay problemas desconectando al cliente
     */
    void desconectar( ) throws IOException;

    /**
     * Agrega un amigo al usuario
     * @param nombreAmigo El nombre del nuevo amigo - amigo != null
     */
    void agregarAmigo( String nombreAmigo );

    /**
     * Actualiza el estado de un amigo. Si no se encontraba en la lista de amigos, entonces es agregado.
     * @param amigo Los datos del amigo - amigo != null
     */
    void actualizarEstado( Usuario amigo );

    /**
     * Este mtodo es usado cuando el usuario local va a iniciar una conversacin con un amigo. <br>
     * El proceso de crear una conversacin se ejecuta en tres etapas principales. <br>
     * 1. Se enva al servidor un mensaje solicitando que se le enve un mensaje al otro cliente para que entre a hacer parte de la conversacin. <br>
     * 2. Se crea un objeto Conversacion que se encargar de recibir y manejar la conexin del cliente remoto. <br>
     * 3. Se retorna el objeto conversacin creado para que se abra la ventana correspondiente en la interfaz.
     * @param amigo El nombre del amigo con el que se va a realizar la conversacin - amigo != null && amigo hace parte de la lista de amigos
     * @return Retorna el objeto conversacin creado
     * @throws IOException Se lanza esta excepcin si hay problemas estableciendo la comunicacin con el amigo
     */
    Conversacion crearConversacionLocal( String amigo ) throws IOException;

    /**
     * Este mtodo es usado cuando se recibe una notificacin que indica que se debe participar en una conversacin con un amigo. <br>
     * El proceso para crear la conversacin es el siguiente. <br>
     * 1. Se crear un objeto Conversacion que se encargar de recibir y manejar la conexin del cliente remoto. <br>
     * 2. El objeto Conversacion recin creado establece una comunicacin con el otro cliente. <br>
     * 3. Se le notifica a la interfaz que se cre una Conversacin para que se cree la ventana correspondiente en la interfaz.
     * @param nombreAmigo El nombre del amigo con el que se va a realizar la conversacin - nombreAmigo != null
     * @param ipAmigo La direccin del amigo con el que se va a realizar la conversacin - ipAmigo != null && ipAmigo != ""
     * @param puertoAmigo El puerto del amigo al que se debe conectar el cliente
     * @throws IOException Se lanza esta excepcin si hay problemas estableciendo la comunicacin con el amigo
     */
    void conectarAConversacion( String nombreAmigo, String ipAmigo, int puertoAmigo ) throws IOException;

    /**
     * Elimina de la coleccin de conversaciones la conversacin indicada
     * @param conv La conversacin que se va a eliminar - conv != null
     */
    void eliminarConversacion( Conversacion conv );

    /**
     * Mtodo para la extensin 1
     * @return respuesta1
     */
    String metodo1( );

    /**
     * Mtodo para la extensin2
     * @return respuesta2
     */
    String metodo2( );
}