/** ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * $Id: ManejadorCliente.java,v 1.2 2006/12/11 07:30:44 man-muno Exp $
 * Universidad de los Andes (Bogot - Colombia)
 * Departamento de Ingeniera de Sistemas y Computacin 
 * Licenciado bajo el esquema Academic Free License versin 2.1
 *
 * Proyecto Cupi2 (http://cupi2.uniandes.edu.co)
 * Ejercicio: n12_messengerAmigos
 * Autor: Mario Snchez - 21/04/2006
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */


package uniandes.cupi2.messengerAmigos.servidor;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.Socket;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

/**
 * Esta clase se encarga de manejar la comunicacin entre el servidor y un cliente. <br>
 * A travs de esta clase el servidor puede enviar mensajes a un cliente que est conectado. <br>
 * Una instancia de la clase ThreadRecibirMensajesCliente es la encargada de recibir los mensajes enviados por el cliente, pero en esta clase es donde se implementa la lgica
 * para ejecutar las acciones que corresponden a los mensajes enviados.
 */
public class ManejadorCliente
{
    // -----------------------------------------------------------------
    // Atributos
    // -----------------------------------------------------------------

    /**
     * Es el servidor en el que se encuentra este manejador
     */
    private IServidorAmigos servidor;

    /**
     * Referencia al objeto que permite el acceso a la base de datos
     */
    private IAdministradorAmigos adminAmigos;

    /**
     * Nombre del usuario al que est asociado este manejador
     */
    private String nombreUsuario;

    /**
     * Direccin IP del usuario
     */
    private String direccionIp;

    /**
     * Puerto que el cliente estableci para que se lleven a cabo las conversaciones
     */
    private int puerto;

    /**
     * El socket que est conectado al cliente
     */
    private Socket socketCliente;

    /**
     * Es el stream usuado para leer los mensajes enviados por el cliente
     */
    private BufferedReader inCliente;

    /**
     * Es el stream usado para enviar mensajes al cliente
     */
    private PrintWriter outCliente;

    // -----------------------------------------------------------------
    // Constructores
    // -----------------------------------------------------------------

    /**
     * Construye el manejador que se encargar de la comunicacin con un cliente. <br>
     * Al construir el manejador se preparan tambin los streams que sern usados para la comunicacin con el cliente.
     * @param servidorAmigos Es el servidor en el que se encuentra este manejador - servidorAmigos != null
     * @param socket Es el socket que est conectado al cliente - socket != null
     * @throws IOException Se lanza esta excepcin si hay problemas estableciendo la comunicacin con el cliente
     */
    public ManejadorCliente( IServidorAmigos servidorAmigos, Socket socket ) throws IOException
    {
        servidor = servidorAmigos;
        adminAmigos = servidor.darAministradorAmigos( );

        socketCliente = socket;

        outCliente = new PrintWriter( socketCliente.getOutputStream( ), true );
        inCliente = new BufferedReader( new InputStreamReader( socketCliente.getInputStream( ) ) );
    }

    // -----------------------------------------------------------------
    // Mtodos
    // -----------------------------------------------------------------

    /**
     * Retorna el nombre del usuario asociado a este manejador
     * @return nombreUsuario
     */
    public String darNombre( )
    {
        return nombreUsuario;
    }

    /**
     * Inicia el manejador, leyendo el mensaje inicial del cliente, enviandole la informacin de sus amigos y luego activando el Thread que va a recibir los otros mensajes que
     * enve. <br>
     * Los pasos de la inicializacin son: <br>
     * 1. Recibir el mensaje inicial del cliente (que tiene la informacin de conexin del usuario) <br>
     * 2. Iniciar el hilo que se encargar de recibir los mensajes enviados por el cliente (ThreadRecibirMensajesCliente) <br>
     * 3. Revisar en la base de datos si el usuario ya existe y en caso negativo crear el nuevo usuario <br>
     * 4. Cambiar el estado del usuario en la base de datos <br>
     * 5. Enviar a las personas que conocen al usuario la notificacin de que ahora est online <br>
     * 6. Enviar al usuario notificaciones con el estado de sus amigos
     * @throws MessengerException Se lanza esta excepcin si hay un problema con los mensajes del cliente
     * @throws IOException Se lanza esta excepcin si hay problemas en la comunicacin
     */
    public void iniciarManejador( ) throws MessengerException, IOException
    {
        try
        {
            // 1. Recibir el mensaje inicial
            String mensajeCliente = inCliente.readLine( );
            if( !mensajeCliente.startsWith( IServidorAmigos.LOGIN + ":" ) )
                throw new MessengerException( "El formato del mensaje no es el esperado: " + mensajeCliente );

            String strDatosCliente = mensajeCliente.split( ":" )[ 1 ];
            String[] datosCliente = strDatosCliente.split( ";" );
            if( datosCliente.length != 3 )
                throw new MessengerException( "El formato del mensaje no es el esperado: " + mensajeCliente );

            // Sacar del mensaje la informacin del cliente
            nombreUsuario = datosCliente[ 0 ];
            direccionIp = datosCliente[ 1 ];
            puerto = Integer.parseInt( datosCliente[ 2 ] );

            // 2. Iniciar el hilo que recibir los siguientes mensajes
            ThreadRecibirMensajesCliente hiloRecibirMensajes = new ThreadRecibirMensajesCliente( this, inCliente );
            hiloRecibirMensajes.start( );

            // 3. Consultar en la base de datos los amigos del usuario.
            // Si el usuario no existe hay que crearlo
            Collection amigos = null;
            Collection conocen = null;

            if( !adminAmigos.existeUsuario( nombreUsuario ) )
            {
                adminAmigos.crearUsuario( nombreUsuario );
                amigos = new ArrayList( );
                conocen = new ArrayList( );
            }
            else
            {
                amigos = adminAmigos.darAmigos( nombreUsuario );
                conocen = adminAmigos.darPersonasConocen( nombreUsuario );
            }

            // 4. Cambiar el estado del usuario en la base de datos
            adminAmigos.cambiarEstadoUsuario( nombreUsuario, Usuario.STR_ONLINE, direccionIp, puerto );

            // 5. Enviar a las personas que conocen al usuario la notificacin de que ahora est online
            enviarNotificacionEstado( conocen, Usuario.ONLINE );

            // 6. Enviar al usuario notificaciones con el estado de sus amigos
            enviarNotificacionEstadoAmigos( amigos );

        }
        catch( SQLException e )
        {
            e.printStackTrace( );
        }
    }

    /**
     * Desconecta al usuario, notifica a todas las personas que lo conocen y destruye el manejador de clientes. <br>
     * <b>pre: </b> No hay nadie leyendo ni escribiendo en los streams; el thread que recibe los mensajes del cliente ya va a terminar su ejecucin
     * @throws IOException Se lanza esta excepcin si hay problemas desconectando el socket
     */
    public void desconectarUsuario( ) throws IOException
    {
        try
        {
            PrintWriter tempOut = outCliente;
            outCliente = null;

            BufferedReader tempIn = inCliente;
            inCliente = null;

            // Notificar al servidor de la desconexin del cliente y cambiar el estado del usuario en la base de datos
            adminAmigos.cambiarEstadoUsuario( nombreUsuario, Usuario.STR_OFFLINE, "", 0 );

            servidor.desconectarCliente( nombreUsuario );

            // Enviar la notificacion a las personas que conocen al usuario
            Collection conocen = adminAmigos.darPersonasConocen( nombreUsuario );
            enviarNotificacionEstado( conocen, Usuario.OFFLINE );

            // Enviar al cliente un mensaje avisando de la desconexin
            tempOut.println( IServidorAmigos.DESCONEXION );

            // Desconectar al cliente
            tempOut.close( );
            tempIn.close( );
            socketCliente.close( );

            socketCliente = null;
        }
        catch( SQLException e )
        {
            e.printStackTrace( );
        }
    }

    /**
     * Agrega un amigo al usuario. Actualiza al base de datos y enva al cliente el estado del amigo.
     * @param nombreAmigo El nombre del amigo que se va a agregar - nombreAmigo != null && nombreAmigo != ""
     */
    public void agregarAmigo( String nombreAmigo )
    {
        try
        {
            // Revisar si el usuario ya es su amigo
            boolean encontroAmigo = adminAmigos.existeAmigo( nombreUsuario, nombreAmigo );

            if( !encontroAmigo )
            {
                adminAmigos.agregarAmigo( nombreUsuario, nombreAmigo );
                Usuario amigo = adminAmigos.darEstadoUsuario( nombreAmigo );

                if( amigo != null )
                    enviarEstadoAmigo( amigo );
            }
        }
        catch( SQLException e )
        {
            e.printStackTrace( );
        }
    }

    /**
     * Enva a un amigo, a travs del servidor, un mensaje para iniciar una conversacin
     * @param nombreAmigo El nombre del amigo con el que se va a establecer la conversacin - nombreAmigo != null && nombreAmigo != ""
     */
    public void enviarConversacionAmigo( String nombreAmigo )
    {
        servidor.iniciarConversacion( nombreAmigo, nombreUsuario, direccionIp, puerto );
    }

    /**
     * Enva al usuario un mensaje indicando que un amigo quiere que se inicie una conversacin
     * @param nombreAmigo El nombre del amigo al que se debe conectar el usuario - amigo != null
     * @param direccionIp La direccion ip del amigo con el que se va a establecer la conversacin - direccionIp != null
     * @param puerto El puerto al que se debe conectar el usuario
     */
    public void iniciarConversacion( String nombreAmigo, String direccionIp, int puerto )
    {
        if( outCliente != null )
        {
            outCliente.println( IServidorAmigos.INCIAR_CHARLA + ":" + nombreAmigo + ";" + direccionIp + ";" + puerto );
        }
    }

    /**
     * Enva al usuario el estado de uno de sus amigos
     * @param amigo El amigo del que se quiere enviar la informacin - amigo != null
     */
    public void enviarEstadoAmigo( Usuario amigo )
    {
        if( outCliente != null )
        {
            if( amigo.darEstado( ) == Usuario.ONLINE )
            {
                outCliente.println( IServidorAmigos.ONLINE + ":" + amigo.darNombre( ) + ";" + amigo.darDireccionIp( ) + ";" + amigo.darPuerto( ) );
            }
            else
            {
                outCliente.println( IServidorAmigos.OFFLINE + ":" + amigo.darNombre( ) );
            }
        }
    }

    /**
     * Enva al usuario mensajes indicando el estado de cada uno de sus amigos
     * @param amigos La coleccin de amigos del usuario - amigos != null y amigos contiene objetos de tipo Usuario
     */
    private void enviarNotificacionEstadoAmigos( Collection amigos )
    {
        for( Iterator iter = amigos.iterator( ); iter.hasNext( ); )
        {
            Usuario amigo = ( Usuario )iter.next( );
            enviarEstadoAmigo( amigo );
        }
    }

    /**
     * Enva un mensaje a las personas que conocen al usuario indicndole su estado actual
     * @param personas La coleccin con las personas a los que se les va a mandar la notificacin - personas != null y personas contiene objetos de tipo Usuario
     * @param estado El estado que se va a notificar - estado = Usuario.ONLINE || estado = Usuario.OFFLINE
     */
    private void enviarNotificacionEstado( Collection personas, int estado )
    {
        Usuario usuario = new Usuario( nombreUsuario );
        if( estado == Usuario.ONLINE )
            usuario.cambiarEstado( direccionIp, puerto );

        for( Iterator iter = personas.iterator( ); iter.hasNext( ); )
        {
            Usuario amigo = ( Usuario )iter.next( );

            // La notificacin solamente se enva a los usuarios conectados
            if( amigo.darEstado( ) == Usuario.ONLINE )
            {
                servidor.enviarNotificacionAmigo( amigo.darNombre( ), usuario );
            }
        }
    }

    /**
     * Retorna una cadena con el nombre y la direccin IP de los clientes que estn conectados actualmente
     * @return cadena
     */
    public String toString( )
    {
        return nombreUsuario + " - " + direccionIp + ":" + puerto;
    }
}
