package gold.structures.automaton;

import gold.structures.map.*;
import gold.structures.tuple.*;
import gold.util.*;

public class GDeterministicTransducer<V> extends GDeterministicAutomaton<V> implements ITransducer<V> {
  
  private static final long serialVersionUID=1958441227527831771L;
  
  protected final IAlphabet outputAlphabet;
  protected final IMap<V,String> stateOutputMap=new GHashTableMap<V,String>();
  protected final IMap<ICouple<V,Character>,String> transitionOutputMap=new GHashTableMap<ICouple<V,Character>,String>();
  protected transient IMethod stateOutputFunction=null;
  protected transient IMethod transitionOutputFunction=null;
  protected final StringBuffer outputString=new StringBuffer();
  
  public GDeterministicTransducer(Iterable<? extends V> pStates, Object pInputAlphabet, Object pOutputAlphabet, V pInitialState, Iterable<? extends V> pAcceptStates, IMethod pTransitionFunction, IMethod pStateOutputFunction, IMethod pTransitionOutputFunction) {
    super(pStates,pInputAlphabet,pInitialState,pAcceptStates,pTransitionFunction);
    outputAlphabet=new GAlphabet(pOutputAlphabet);
    setStateOutputFunction(pStateOutputFunction);
    setTransitionOutputFunction(pTransitionOutputFunction);
  }
  
  public GDeterministicTransducer(Iterable<? extends V> pStates, Object pInputAlphabet, Object pOutputAlphabet, V pInitialState, Iterable<? extends V> pAcceptStates) {
    super(pStates,pInputAlphabet,pInitialState,pAcceptStates);
    outputAlphabet=new GAlphabet(pOutputAlphabet);
  }
  
  public GDeterministicTransducer(Iterable<? extends V> pStates, Object pInputAlphabet, Object pOutputAlphabet) {
    super(pStates,pInputAlphabet);
    outputAlphabet=new GAlphabet(pOutputAlphabet);
  }
  
  public <W extends V> GDeterministicTransducer(IAutomaton<W> pAutomaton) {
    super(pAutomaton);
    if (pAutomaton instanceof ITransducer) {
      ITransducer<W> transducer=(ITransducer<W>)pAutomaton;
      outputAlphabet=new GAlphabet(transducer.getOutputAlphabet());
      for (W input:transducer.getStates()) {
        setStateOutput(input,transducer.getStateOutput(input));
        for (char action:getActions()) {
          setTransitionOutput(input,action,transducer.getTransitionOutput(input,action));
        }
      }
    }
    else {
      outputAlphabet=new GAlphabet("");
    }
  }
  
  protected String checkString(String pString) {
    if (!outputAlphabet.containsAll(pString)) throw new IllegalArgumentException("The string '"+pString+"' contains symbols that are not in the automaton's output alphabet.");
    return pString;
  }
  
  @Override
  public IAlphabet getInputAlphabet() {
    return getAlphabet();
  }
  
  @Override
  public IAlphabet getOutputAlphabet() {
    return outputAlphabet;
  }
  
  @Override
  public String getStateOutput(V pInput) {
    try {
      if (stateOutputFunction==null&&stateOutputMap.isEmpty()) return "";
      Object result=stateOutputFunction!=null?stateOutputFunction.invoke(checkState(pInput)):stateOutputMap.get(checkState(pInput));
      return checkString(result!=null?GToolkit.toString(result):"");
    }
    catch (RuntimeException ex) {
      throw new RuntimeException("There was an error evaluating the state output function for the state '"+pInput+"'.",ex);
    }
  }
  
  @Override
  public String getTransitionOutput(V pInput, char pAction) {
    try {
      if (transitionOutputFunction==null&&transitionOutputMap.isEmpty()) return "";
      Object result=transitionOutputFunction!=null?transitionOutputFunction.invoke(checkState(pInput),checkAction(pAction)):transitionOutputMap.get(new GCouple<V,Character>(checkState(pInput),checkAction(pAction)));
      return checkString(result!=null?GToolkit.toString(result):"");
    }
    catch (RuntimeException ex) {
      throw new RuntimeException("There was an error evaluating the transition output function for the state '"+pInput+"' and symbol '"+pAction+"'.",ex);
    }
  }
  
  @Override
  public String getOutputString() {
    return checkString(outputString.toString());
  }
  
  @Override
  public boolean acceptsString(String pInputString) {
    outputString.setLength(0);
    outputString.append(getStateOutput(initialState));
    V current=initialState;
    for (char symbol:pInputString.toCharArray()) {
      if (symbol==IAutomaton.LAMBDA) throw new IllegalArgumentException("The input string cannot have lambda symbols.");
      outputString.append(getTransitionOutput(current,symbol));
      current=getDelta(current,symbol);
      outputString.append(getStateOutput(current));
    }
    return acceptStates.contains(current);
  }
  
  public void setStateOutput(V pInput, String pString) {
    stateOutputMap.put(checkState(pInput),checkString(pString));
  }
  
  public void setTransitionOutput(V pInput, char pAction, String pString) {
    transitionOutputMap.put(new GCouple<V,Character>(checkState(pInput),checkAction(pAction)),checkString(pString));
  }
  
  public void setStateOutputFunction(IMethod pStateOutputFunction) {
    stateOutputFunction=pStateOutputFunction;
    if (stateOutputFunction!=null) {
      for (V input:getStates()) {
        getStateOutput(input); // Check function
      }
    }
  }
  
  public void setTransitionOutputFunction(IMethod pTransitionOutputFunction) {
    transitionOutputFunction=pTransitionOutputFunction;
    if (transitionOutputFunction!=null) {
      for (V input:getStates()) {
        for (char action:getActions()) {
          getTransitionOutput(input,action); // Check function
        }
      }
    }
  }
  
  @Override
  public GDeterministicTransducer<V> clone() {
    return new GDeterministicTransducer<V>(this);
  }
  
  private Object writeReplace() {
    return clone();
  }
  
}
