package gold.structures.automaton;

import java.io.*;
import java.util.*;
import gold.structures.collection.*;
import gold.structures.graph.*;
import gold.structures.multimap.*;
import gold.structures.queue.*;
import gold.structures.set.*;
import gold.structures.tuple.*;
import gold.util.*;

public class GNondeterministicAutomaton<V> implements IAutomaton<V> {
  
  private static final long serialVersionUID=-6510344592884913437L;
  
  protected final ISet<V> states;
  protected final IAlphabet alphabet;
  protected V initialState=null;
  protected final ISet<V> acceptStates=new GHashTableSet<V>();
  protected final IMultiMap<ICouple<V,Character>,V> transitionMap=new GHashTableMultiMap<ICouple<V,Character>,V>();
  protected transient IMethod transitionFunction=null;
  
  public GNondeterministicAutomaton(Iterable<? extends V> pStates, Object pAlphabet, V pInitialState, Iterable<? extends V> pAcceptStates, IMethod pTransitionFunction) {
    if (!GCollections.isFinite(pStates)) throw new IllegalArgumentException("The set of states of the automaton cannot be infinite.");
    if (GCollections.isEmpty(pStates)) throw new IllegalArgumentException("The automaton must have at least one state.");
    if (GCollections.asJavaCollection(pStates).contains(null)) throw new IllegalArgumentException("The automaton cannot contain null states.");
    states=new GHashTableSet<V>(GCollections.unmodifiableCollection(pStates));
    alphabet=new GAlphabet(pAlphabet);
    if (alphabet.isEmpty()) throw new IllegalArgumentException("The input alphabet cannot be empty.");
    if (alphabet.contains(IAutomaton.LAMBDA)) throw new IllegalArgumentException("The input alphabet cannot contain the lambda greek letter ('"+IAutomaton.LAMBDA+").");
    setInitialState(pInitialState);
    setAcceptStates(pAcceptStates);
    setTransitionFunction(pTransitionFunction);
  }
  
  @SuppressWarnings("unchecked")
  public GNondeterministicAutomaton(Iterable<? extends V> pStates, Object pAlphabet, V pInitialState, Iterable<? extends V> pAcceptStates, Object[] pTransitionFunctionTable) {
    this(pStates,pAlphabet,pInitialState,pAcceptStates,(IMethod)null);
    for (Object row:pTransitionFunctionTable) {
      Iterator<?> it=GCollections.unmodifiableCollection(row).iterator();
      addDelta((V)(it.next()),(Character)(it.next()),(V)(it.next()));
    }
  }
  
  public GNondeterministicAutomaton(Iterable<? extends V> pStates, Object pAlphabet, V pInitialState, Iterable<? extends V> pAcceptStates) {
    this(pStates,pAlphabet,pInitialState,pAcceptStates,(IMethod)null);
  }
  
  public GNondeterministicAutomaton(Iterable<? extends V> pStates, Object pAlphabet) {
    this(pStates,pAlphabet,pStates.iterator().hasNext()?pStates.iterator().next():null,GCollections.<V>emptySet(),(IMethod)null);
  }
  
  public <W extends V> GNondeterministicAutomaton(IAutomaton<W> pAutomaton) {
    this(pAutomaton.getStates(),pAutomaton.getAlphabet(),pAutomaton.getInitialState(),pAutomaton.getAcceptStates(),(IMethod)null);
    for (W input:pAutomaton.getStates()) {
      for (char action:getActions()) {
        setDeltaSet(input,action,pAutomaton.getDeltaSet(input,action));
      }
    }
  }
  
  protected V checkState(V pState) {
    checkStates(GCollections.singletonCollection(pState));
    return pState;
  }
  
  protected <T extends Iterable<? extends V>> T checkStates(T pStates) {
    for (V state:pStates) {
      if (!states.contains(state)) throw new IllegalArgumentException("The state "+state+" doesn't belong to the set of states of the automaton.");
    }
    return pStates;
  }
  
  protected char checkAction(char pAction) {
    if (pAction!=IAutomaton.LAMBDA&&!alphabet.contains(pAction)) throw new IllegalArgumentException("The character '"+pAction+"' isn't a symbol of the automaton's alphabet.");
    return pAction;
  }
  
  protected Object invokeTransitionFunction(V pInput, char pAction) {
    if (transitionFunction==null) throw new IllegalStateException("The transition function of the automaton is not defined.");
    return transitionFunction.invoke(checkState(pInput),checkAction(pAction));
  }
  
  protected static <V> ISet<String> search(IAutomaton<V> pAutomaton, int pMaximumLength, boolean pSearchAcceptedStrings) {
    assert pAutomaton.isDeterministic();
    if (pMaximumLength<0) throw new IllegalArgumentException("The maximum length cannot be negative.");
    ISet<String> result=new GRedBlackTreeSet<String>(GComparators.shortlexStringComparator());
    IQueue<ICouple<V,String>> queue=new GArrayQueue<ICouple<V,String>>(pAutomaton.getStates().size()/2);
    queue.enqueue(new GCouple<V,String>(pAutomaton.getInitialState(),""));
    while (!queue.isEmpty()) {
      ICouple<V,String> couple=queue.dequeue();
      V input=couple.getFirst();
      String string=couple.getSecond();
      if (string.length()>pMaximumLength) break;
      if (pSearchAcceptedStrings==pAutomaton.isAcceptState(input)) result.add(string);
      for (char symbol:pAutomaton.getAlphabet()) {
        queue.enqueue(new GCouple<V,String>(pAutomaton.getDelta(input,symbol),string+symbol));
      }
    }
    return result;
  }
  
  @Override
  public ISet<V> getStates() {
    return GCollections.unmodifiableSet(states);
  }
  
  @Override
  public IAlphabet getAlphabet() {
    return alphabet;
  }
  
  @Override
  public V getInitialState() {
    return initialState;
  }
  
  @Override
  public ISet<V> getAcceptStates() {
    return GCollections.unmodifiableSet(acceptStates);
  }
  
  @Override
  public V getDelta(V pInput, char pAction) {
    if (pAction==IAutomaton.LAMBDA) {
      throw new IllegalArgumentException("A deterministic automaton doesn't must have lambda transitions.");
    }
    else {
      ISet<V> outputs=getDeltaSet(checkState(pInput),checkAction(pAction));
      if (outputs.size()==0) throw new IllegalStateException("The automaton doesn't have a transition for the input '"+pInput+"' and action '"+pAction+"'.");
      if (outputs.size()!=1) throw new IllegalStateException("The automaton has multiple transitions for the input '"+pInput+"' and action '"+pAction+"'.");
      return outputs.pick();
    }
  }
  
  @Override
  @SuppressWarnings("unchecked")
  public ISet<V> getDeltaSet(V pInput, char pAction) {
    try {
      if (transitionFunction!=null) {
        return checkStates(GCollections.unmodifiableSet((Iterable<V>)(invokeTransitionFunction(pInput,pAction))));
      }
      else {
        return GCollections.unmodifiableSet(transitionMap.get(new GCouple<V,Character>(checkState(pInput),checkAction(pAction))));
      }
    }
    catch (RuntimeException ex) {
      throw new RuntimeException("There was an error evaluating the transition function for the state '"+pInput+"' and symbol '"+pAction+"'.",ex);
    }
  }
  
  @Override
  public Iterable<Character> getActions() {
    return new Iterable<Character>() {
      @Override
      public Iterator<Character> iterator() {
        return new GAbstractIterator<Character>() {
          private int index=0;
          @Override
          public boolean hasNext() {
            return index<=alphabet.size();
          }
          @Override
          public Character next() {
            Character ch=index==alphabet.size()?IAutomaton.LAMBDA:alphabet.getSymbol(index);
            index++;
            return ch;
          }
        };
      }
    };
  }
  
  @Override
  public boolean isAcceptState(V pState) {
    return acceptStates.contains(pState);
  }
  
  @Override
  public boolean isDeterministic() {
    for (V state:states) {
      for (char symbol:getAlphabet()) {
        if (getDeltaSet(state,symbol).size()!=1) return false;
      }
      if (!getDeltaSet(state,IAutomaton.LAMBDA).isEmpty()) return false;
    }
    return true;
  }
  
  @Override
  public ISet<V> applyLambdasRecursively(Iterable<? extends V> pStates) {
    int size=GCollections.sizeOf(pStates);
    ISet<V> result=new GHashTableSet<V>(size*2);
    IQueue<V> queue=new GArrayQueue<V>(size);
    for (V state:pStates) {
      result.add(state);
      queue.enqueue(state);
    }
    while (!queue.isEmpty()) {
      V input=queue.dequeue();
      for (V output:getDeltaSet(input,IAutomaton.LAMBDA)) {
        if (!result.contains(output)) {
          result.add(output);
          queue.enqueue(output);
        }
      }
    }
    return GCollections.unmodifiableSet(result);
  }
  
  @Override
  public ISet<V> applyAction(Iterable<? extends V> pStates, char pAction) {
    char action=checkAction(pAction);
    ISet<V> result=new GHashTableSet<V>(GCollections.sizeOf(pStates)*2);
    for (V input:pStates) {
      result.union(getDeltaSet(input,action));
    }
    return applyLambdasRecursively(result);
  }
  
  @Override
  public ISet<V> getReachableStates() {
    ISet<V> result=new GHashTableSet<V>(states.size());
    IQueue<V> queue=new GArrayQueue<V>(states.size()/2);
    result.add(initialState);
    queue.enqueue(initialState);
    while (!queue.isEmpty()) {
      V input=queue.dequeue();
      for (char action:getActions()) {
        for (V output:getDeltaSet(input,action)) {
          if (!result.contains(output)) {
            result.add(output);
            queue.enqueue(output);
          }
        }
      }
    }
    return GCollections.unmodifiableSet(result);
  }
  
  @Override
  public ISet<V> getLiveStates() {
    IMultiMap<V,V> predecessors=new GHashTableMultiMap<V,V>(states.size());
    for (V input:states) {
      for (char action:getActions()) {
        for (V output:getDeltaSet(input,action)) {
          predecessors.add(output,input);
        }
      }
    }
    ISet<V> result=new GHashTableSet<V>(states.size());
    IQueue<V> queue=new GArrayQueue<V>(states.size()/2);
    for (V acceptState:acceptStates) {
      result.add(acceptState);
      queue.enqueue(acceptState);
    }
    while (!queue.isEmpty()) {
      V output=queue.dequeue();
      for (V input:predecessors.get(output)) {
        if (!result.contains(input)) {
          result.add(input);
          queue.enqueue(input);
        }
      }
    }
    return GCollections.unmodifiableSet(result);
  }
  
  @Override
  public ISet<V> getUsefulStates() {
    return GCollections.unmodifiableSet(GCollections.intersection(getReachableStates(),getLiveStates()));
  }
  
  @Override
  public ISet<String> getAcceptedStrings(int pMaximumLength) {
    return search(determinize(),pMaximumLength,true);
  }
  
  @Override
  public ISet<String> getRejectedStrings(int pMaximumLength) {
    return search(determinize(),pMaximumLength,false);
  }
  
  @Override
  public boolean acceptsNoStrings() {
    return getReachableStates().isDisjointFrom(acceptStates);
  }
  
  @Override
  public boolean acceptsAllStrings() {
    return complement().acceptsNoStrings();
  }
  
  @Override
  public boolean acceptsEmptyString() {
    return acceptsString("");
  }
  
  @Override
  public boolean acceptsString(String pInputString) {
    ISet<V> current=applyLambdasRecursively(GCollections.singletonSet(initialState));
    for (char symbol:pInputString.toCharArray()) {
      if (symbol==IAutomaton.LAMBDA) throw new IllegalArgumentException("The input string cannot have lambda symbols.");
      current=applyAction(current,symbol);
    }
    return !current.isDisjointFrom(acceptStates);
  }
  
  @Override
  public IAutomaton<?> simplify() {
    return GAutomata.simplify(this);
  }
  
  @Override
  public IAutomaton<?> determinize() {
    return GAutomata.determinize(this);
  }
  
  @Override
  public IAutomaton<?> minimize() {
    return GAutomata.minimize(this);
  }
  
  @Override
  public IAutomaton<?> union(IAutomaton<?> pAutomaton) {
    return GAutomata.union(this,pAutomaton);
  }
  
  @Override
  public IAutomaton<?> intersection(IAutomaton<?> pAutomaton) {
    return GAutomata.intersection(this,pAutomaton);
  }
  
  @Override
  public IAutomaton<?> concatenation(IAutomaton<?> pAutomaton) {
    return GAutomata.concatenation(this,pAutomaton);
  }
  
  @Override
  public IAutomaton<?> difference(IAutomaton<?> pAutomaton) {
    return GAutomata.difference(this,pAutomaton);
  }
  
  @Override
  public IAutomaton<?> transpose() {
    return GAutomata.transpose(this);
  }
  
  @Override
  public IAutomaton<?> complement() {
    return GAutomata.complement(this);
  }
  
  @Override
  public IAutomaton<?> optionality() {
    return GAutomata.optionality(this);
  }
  
  @Override
  public IAutomaton<?> kleenePlus() {
    return GAutomata.kleenePlus(this);
  }
  
  @Override
  public IAutomaton<?> kleeneStar() {
    return GAutomata.kleeneStar(this);
  }
  
  @Override
  public boolean isEquivalentTo(IAutomaton<?> pAutomaton) {
    return GAutomata.areEquivalent(this,pAutomaton);
  }
  
  @Override
  public boolean isDisjointFrom(IAutomaton<?> pAutomaton) {
    return GAutomata.areDisjoint(this,pAutomaton);
  }
  
  @Override
  public boolean isSubset(IAutomaton<?> pAutomaton) {
    return GAutomata.isSubset(this,pAutomaton);
  }
  
  @Override
  public IGraph<V,ISet<?>> toGraphCopy() {
    IGraph<V,ISet<?>> graph=new GDirectedGraph<V,ISet<?>>(states) {
      private static final long serialVersionUID=4839551200822649694L;
      @Override
      public boolean isUndirected() {
        return false;
      }
    };
    IMultiMap<IPair<V>,Character> multimap=new GHashTableMultiMap<IPair<V>,Character>(states.size());
    for (V input:states) {
      for (char action:getActions()) {
        for (V output:getDeltaSet(input,action)) {
          multimap.add(new GPair<V>(input,output),action);
        }
      }
    }
    for (Map.Entry<IPair<V>,ISet<Character>> entry:multimap.toJavaView().entrySet()) {
      graph.addEdge(entry.getKey().getFirst(),entry.getKey().getSecond(),1.0,entry.getValue());
    }
    return graph;
  }
  
  @Override
  public GNondeterministicAutomaton<V> clone() {
    return new GNondeterministicAutomaton<V>(this);
  }
  
  @Override
  public boolean equals(Object pObject) {
    return this==pObject;
  }
  
  @Override
  public int hashCode() {
    return toString().hashCode();
  }
  
  @Override
  public String toString() {
    StringWriter writer=new StringWriter(256);
    try {
      GAutomata.write(writer,this);
    }
    catch (IOException ex) {
      throw new RuntimeException(ex);
    }
    return writer.toString();
  }
  
  public void setInitialState(V pInitialState) {
    initialState=checkState(pInitialState);
  }
  
  public void addAcceptState(V pAcceptState) {
    acceptStates.add(checkState(pAcceptState));
  }
  
  public void removeAcceptState(V pAcceptState) {
    acceptStates.remove(checkState(pAcceptState));
  }
  
  public void setAcceptStates(Iterable<? extends V> pAcceptStates) {
    acceptStates.clear();
    acceptStates.union(GCollections.unmodifiableSet(checkStates(pAcceptStates)));
  }
  
  public void addDelta(V pInput, char pAction, V pOutput) {
    transitionMap.add(new GCouple<V,Character>(checkState(pInput),checkAction(pAction)),checkState(pOutput));
  }
  
  public void removeDelta(V pInput, char pAction, V pOutput) {
    transitionMap.remove(new GCouple<V,Character>(checkState(pInput),checkAction(pAction)),checkState(pOutput));
  }
  
  public void setDelta(V pInput, char pAction, V pOutput) {
    setDeltaSet(pInput,pAction,GCollections.singletonSet(pOutput));
  }
  
  public void setDeltaSet(V pInput, char pAction, Iterable<? extends V> pOutputs) {
    transitionMap.put(new GCouple<V,Character>(checkState(pInput),checkAction(pAction)),new GHashTableSet<V>(GCollections.unmodifiableSet(checkStates(pOutputs))));
  }
  
  public void setTransitionFunction(IMethod pTransitionFunction) {
    transitionFunction=pTransitionFunction;
    if (transitionFunction!=null) {
      for (V input:getStates()) {
        for (char action:getActions()) {
          getDeltaSet(input,action); // Check function
        }
      }
    }
  }
  
  private Object writeReplace() {
    return clone();
  }
  
}
