package gold.structures.automaton;

import java.io.*;
import java.util.*;
import gold.structures.graph.*;
import gold.structures.multimap.*;
import gold.structures.queue.*;
import gold.structures.set.*;
import gold.structures.tuple.*;
import gold.util.*;

public class GPushdownAutomaton<V> implements IPushdownAutomaton<V> {
  
  private static final long serialVersionUID=-5361953504876579984L;
  
  protected final ISet<V> states;
  protected final IAlphabet inputAlphabet;
  protected final IAlphabet stackAlphabet;
  protected final V initialState;
  protected final ISet<V> acceptStates;
  protected final IMultiMap<V,IPushdownTransition<V>> transitionMap=new GHashTableMultiMap<V,IPushdownTransition<V>>();
  
  public GPushdownAutomaton(Iterable<? extends V> pStates, Object pInputAlphabet, Object pStackAlphabet, V pInitialState, Iterable<? extends V> pAcceptStates) {
    if (!GCollections.isFinite(pStates)) throw new IllegalArgumentException("The set of states of the automaton cannot be infinite.");
    if (GCollections.isEmpty(pStates)) throw new IllegalArgumentException("The automaton must have at least one state.");
    if (GCollections.asJavaCollection(pStates).contains(null)) throw new IllegalArgumentException("The automaton cannot contain null states.");
    states=GCollections.unmodifiableSet(new GHashTableSet<V>(GCollections.unmodifiableCollection(pStates)));
    inputAlphabet=new GAlphabet(pInputAlphabet);
    if (inputAlphabet.isEmpty()) throw new IllegalArgumentException("The input alphabet cannot be empty.");
    if (inputAlphabet.contains(IAutomaton.LAMBDA)) throw new IllegalArgumentException("The input alphabet cannot contain the lambda greek letter ('"+IAutomaton.LAMBDA+").");
    stackAlphabet=new GAlphabet(pStackAlphabet);
    if (stackAlphabet.contains(IAutomaton.LAMBDA)) throw new IllegalArgumentException("The stack alphabet cannot contain the lambda greek letter ('"+IAutomaton.LAMBDA+").");
    initialState=checkState(pInitialState);
    acceptStates=GCollections.unmodifiableSet(new GHashTableSet<V>(GCollections.unmodifiableCollection(checkStates(pAcceptStates))));
  }
  
  public <W extends V> GPushdownAutomaton(IPushdownAutomaton<W> pAutomaton) {
    this(pAutomaton.getStates(),pAutomaton.getInputAlphabet(),pAutomaton.getStackAlphabet(),pAutomaton.getInitialState(),pAutomaton.getAcceptStates());
    for (W state:pAutomaton.getStates()) {
      for (IPushdownTransition<W> transition:pAutomaton.getTransitions(state)) {
        transitionMap.add(transition.getInput(),new GPushdownTransition(transition.getInput(),transition.getPhi(),transition.getPrefix(),transition.getOutput(),transition.getRho()));
      }
    }
  }
  
  protected V checkState(V pState) {
    checkStates(GCollections.singletonCollection(pState));
    return pState;
  }
  
  protected <T extends Iterable<? extends V>> T checkStates(T pStates) {
    for (V state:pStates) {
      if (!states.contains(state)) throw new IllegalArgumentException("The state "+state+" doesn't belong to the set of states of the automaton.");
    }
    return pStates;
  }
  
  protected String checkInputString(String pInputString) {
    for (int i=0; i<pInputString.length(); i++) {
      char ch=pInputString.charAt(i);
      if (!inputAlphabet.contains(ch)) throw new IllegalArgumentException("The character '"+ch+"' isn't a symbol of the automaton's input alphabet.");
      if (ch==IAutomaton.LAMBDA) throw new IllegalArgumentException("The string '"+pInputString+"' cannot contain the lambda symbol ("+IAutomaton.LAMBDA+").");
    }
    return pInputString;
  }
  
  protected String checkStackString(String pStackString) {
    for (int i=0; i<pStackString.length(); i++) {
      char ch=pStackString.charAt(i);
      if (!stackAlphabet.contains(ch)) throw new IllegalArgumentException("The character '"+ch+"' isn't a symbol of the automaton's stack alphabet.");
      if (ch==IAutomaton.LAMBDA) throw new IllegalArgumentException("The string '"+pStackString+"' cannot contain the lambda symbol ("+IAutomaton.LAMBDA+").");
    }
    return pStackString;
  }
  
  @Override
  public ISet<V> getStates() {
    return states;
  }
  
  @Override
  public IAlphabet getAlphabet() {
    return inputAlphabet;
  }
  
  @Override
  public IAlphabet getInputAlphabet() {
    return inputAlphabet;
  }
  
  @Override
  public IAlphabet getStackAlphabet() {
    return stackAlphabet;
  }
  
  @Override
  public V getInitialState() {
    return initialState;
  }
  
  @Override
  public ISet<V> getAcceptStates() {
    return acceptStates;
  }
  
  @Override
  public V getDelta(V pInput, char pAction) {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public ISet<V> getDeltaSet(V pInput, char pAction) {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public Iterable<IPushdownTransition<V>> getTransitions(V pInput) {
    return GCollections.unmodifiableCollection(transitionMap.get(checkState(pInput)));
  }
  
  @Override
  public Iterable<Character> getActions() {
    return inputAlphabet;
  }
  
  @Override
  public boolean isAcceptState(V pState) {
    return acceptStates.contains(pState);
  }
  
  @Override
  public boolean isDeterministic() {
    return false;
  }
  
  @Override
  public IPushdownDelta<V> delta(V pInput, V pOutput, String pPrefix) {
    return new GPushdownDelta(pInput,pOutput,pPrefix);
  }
  
  @Override
  public ISet<V> applyLambdasRecursively(Iterable<? extends V> pStates) {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public ISet<V> applyAction(Iterable<? extends V> pStates, char pAction) {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public ISet<V> getReachableStates() {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public ISet<V> getLiveStates() {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public ISet<V> getUsefulStates() {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public ISet<String> getAcceptedStrings(int pMaximumLength) {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public ISet<String> getRejectedStrings(int pMaximumLength) {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public boolean acceptsNoStrings() {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public boolean acceptsAllStrings() {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public boolean acceptsEmptyString() {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public IPushdownConfiguration<V> trace(String pInputString) {
    int MAX_STEPS=262144;
    IPushdownConfiguration<V> initial=new GPushdownConfiguration(initialState,"",pInputString,null,null);
    if (initial.isAcceptConfiguration()) return initial;
    ISet<IPushdownConfiguration<V>> visited=new GHashTableSet<IPushdownConfiguration<V>>(128);
    IQueue<IPushdownConfiguration<V>> queue=new GArrayQueue<IPushdownConfiguration<V>>(128);
    visited.add(initial);
    queue.enqueue(initial);
    while (!queue.isEmpty()) {
      IPushdownConfiguration<V> current=queue.dequeue();
      for (IPushdownTransition<V> transition:getTransitions(current.getState())) {
        if (current.canApply(transition)) {
          IPushdownConfiguration<V> next=current.apply(transition);
          if (next.isAcceptConfiguration()) return next;
          if (!visited.contains(next)) {
            visited.add(next);
            queue.enqueue(next);
            if (visited.size()==MAX_STEPS) throw new RuntimeException("Too many steps in computation.");
          }
        }
      }
    }
    return null;
  }
  
  @Override
  public boolean acceptsString(String pInputString) {
    return trace(pInputString)!=null;
  }
  
  @Override
  public IAutomaton<?> simplify() {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public IAutomaton<?> determinize() {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public IAutomaton<?> minimize() {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public IAutomaton<?> union(IAutomaton<?> pAutomaton) {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public IAutomaton<?> intersection(IAutomaton<?> pAutomaton) {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public IAutomaton<?> concatenation(IAutomaton<?> pAutomaton) {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public IAutomaton<?> difference(IAutomaton<?> pAutomaton) {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public IAutomaton<?> transpose() {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public IAutomaton<?> complement() {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public IAutomaton<?> optionality() {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public IAutomaton<?> kleenePlus() {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public IAutomaton<?> kleeneStar() {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public boolean isEquivalentTo(IAutomaton<?> pAutomaton) {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public boolean isDisjointFrom(IAutomaton<?> pAutomaton) {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public boolean isSubset(IAutomaton<?> pAutomaton) {
    throw getUnsupportedOperationException();
  }
  
  @Override
  public IGraph<V,ISet<?>> toGraphCopy() {
    IGraph<V,ISet<?>> graph=new GDirectedGraph<V,ISet<?>>(states) {
      private static final long serialVersionUID=-3256387616856200552L;
      @Override
      public boolean isUndirected() {
        return false;
      }
    };
    IMultiMap<IPair<V>,IPushdownTransition<V>> multimap=new GHashTableMultiMap<IPair<V>,IPushdownTransition<V>>(states.size());
    for (ISet<IPushdownTransition<V>> set:transitionMap) {
      for (IPushdownTransition<V> transition:set) {
        multimap.add(new GPair<V>(transition.getInput(),transition.getOutput()),transition);
      }
    }
    for (Map.Entry<IPair<V>,ISet<IPushdownTransition<V>>> entry:multimap.toJavaView().entrySet()) {
      graph.addEdge(entry.getKey().getFirst(),entry.getKey().getSecond(),1.0,entry.getValue());
    }
    return graph;
  }
  
  @Override
  public GPushdownAutomaton<V> clone() {
    return new GPushdownAutomaton<V>(this);
  }
  
  @Override
  public boolean equals(Object pObject) {
    return this==pObject;
  }
  
  @Override
  public int hashCode() {
    return toString().hashCode();
  }
  
  @Override
  public String toString() {
    StringWriter writer=new StringWriter(256);
    try {
      GAutomata.write(writer,this);
    }
    catch (IOException ex) {
      throw new RuntimeException(ex);
    }
    return writer.toString();
  }
  
  private Object writeReplace() {
    return clone();
  }
  
  private static RuntimeException getUnsupportedOperationException() {
    return new UnsupportedOperationException("The operation is not implemented over pushdown automata.");
  }
  
  private class GPushdownTransition implements IPushdownTransition<V> {
    
    private final V input;
    private final String phi;
    private final String prefix;
    private final V output;
    private final String rho;
    
    public GPushdownTransition(V pInput, String pPhi, String pPrefix, V pOutput, String pRho) {
      if (pPhi==null) throw new IllegalArgumentException("The string to be popped cannot be null.");
      if (pPrefix==null) throw new IllegalArgumentException("The prefix cannot be null.");
      if (pRho==null) throw new IllegalArgumentException("The string to be pushed cannot be null.");
      input=checkState(pInput);
      phi=checkStackString(pPhi);
      prefix=checkInputString(pPrefix);
      output=checkState(pOutput);
      rho=checkStackString(pRho);
    }
    
    @Override
    public V getInput() {
      return input;
    }
    
    @Override
    public String getPhi() {
      return phi;
    }
    
    @Override
    public String getPrefix() {
      return prefix;
    }
    
    @Override
    public V getOutput() {
      return output;
    }
    
    @Override
    public String getRho() {
      return rho;
    }
    
    @Override
    public boolean equals(Object pObject) {
      if (!(pObject instanceof IPushdownTransition)) return false;
      IPushdownTransition<?> t=(IPushdownTransition<?>)pObject;
      return GToolkit.equals(input,t.getInput())&&phi.equals(t.getPhi())&&prefix.equals(t.getPrefix())&&GToolkit.equals(output,t.getOutput())&&rho.equals(t.getRho());
    }
    
    @Override
    public int hashCode() {
      return input.hashCode()^phi.hashCode()^prefix.hashCode()^output.hashCode()^rho.hashCode();
    }
    
    @Override
    public String toString() {
      String s=prefix.isEmpty()?""+IAutomaton.LAMBDA:GToolkit.escape(prefix);
      if (phi.isEmpty()&&rho.isEmpty()) {
        return s+",ignore";
      }
      else if (phi.equals(rho)) {
        return s+",skip("+GToolkit.escape(phi)+")";
      }
      else if (phi.isEmpty()&&!rho.isEmpty()) {
        return s+",push("+GToolkit.escape(rho)+")";
      }
      else if (!phi.isEmpty()&&rho.isEmpty()) {
        return s+",pop("+GToolkit.escape(phi)+")";
      }
      else if (rho.startsWith(phi)) {
        return s+",pushOn("+GToolkit.escape(phi)+","+GToolkit.escape(rho.substring(phi.length()))+")";
      }
      else {
        return s+",changeTop("+GToolkit.escape(phi)+","+GToolkit.escape(rho)+")";
      }
    }
    
  }
  
  private class GPushdownConfiguration implements IPushdownConfiguration<V> {
    
    private final V state;
    private final String stack;
    private final String inputString;
    private final IPushdownTransition<V> transition;
    private final IPushdownConfiguration<V> previousConfiguration;
    
    public GPushdownConfiguration(V pState, String pStack, String pInputString, IPushdownTransition<V> pTransition, IPushdownConfiguration<V> pPreviousConfiguration) {
      state=pState;
      stack=pStack;
      inputString=pInputString;
      transition=pTransition;
      previousConfiguration=pPreviousConfiguration;
    }
    
    @Override
    public boolean isAcceptConfiguration() {
      return inputString.isEmpty()&&stack.isEmpty()&&isAcceptState(state);
    }
    
    @Override
    public boolean canApply(IPushdownTransition<V> pTransition) {
      return inputString.startsWith(pTransition.getPrefix())&&stack.endsWith(pTransition.getPhi());
    }
    
    @Override
    public IPushdownConfiguration<V> apply(IPushdownTransition<V> pTransition) {
      String phi=pTransition.getPhi(),prefix=pTransition.getPrefix(),rho=pTransition.getRho();
      return new GPushdownConfiguration(pTransition.getOutput(),stack.substring(0,stack.length()-phi.length())+rho,inputString.substring(prefix.length()),pTransition,this);
    }
    
    @Override
    public V getState() {
      return state;
    }
    
    @Override
    public String getStack() {
      return stack;
    }
    
    @Override
    public String getInputString() {
      return inputString;
    }
    
    @Override
    public IPushdownTransition<V> getTransition() {
      return transition;
    }
    
    @Override
    public IPushdownConfiguration<V> getPreviousConfiguration() {
      return previousConfiguration;
    }
    
    @Override
    public boolean equals(Object pObject) {
      if (!(pObject instanceof IPushdownConfiguration)) return false;
      IPushdownConfiguration<?> c=(IPushdownConfiguration<?>)pObject;
      return GToolkit.equals(state,c.getState())&&stack.equals(c.getStack())&&inputString.equals(c.getInputString());
    }
    
    @Override
    public int hashCode() {
      return state.hashCode()^stack.hashCode()^inputString.hashCode();
    }
    
  }
  
  private class GPushdownDelta implements IPushdownDelta<V> {
    
    private final V input;
    private final V output;
    private final String prefix;
    
    public GPushdownDelta(V pInput, V pOutput, String pPrefix) {
      if (pPrefix==null) throw new IllegalArgumentException("The prefix cannot be null.");
      input=checkState(pInput);
      output=checkState(pOutput);
      prefix=checkInputString(pPrefix);
    }
    
    @Override
    public V getInput() {
      return input;
    }
    
    @Override
    public V getOutput() {
      return output;
    }
    
    @Override
    public String getPrefix() {
      return prefix;
    }
    
    @Override
    public IPushdownTransition<V> removeTransition(String pPhi, String pRho) {
      IPushdownTransition<V> transition=new GPushdownTransition(input,pPhi,prefix,output,pRho);
      return transitionMap.remove(input,transition);
    }
    
    @Override
    public IPushdownTransition<V> insertTransition(String pPhi, String pRho) {
      IPushdownTransition<V> transition=new GPushdownTransition(input,pPhi,prefix,output,pRho);
      transitionMap.add(input,transition);
      return transition;
    }
    
    @Override
    public IPushdownTransition<V> push(String pRho) {
      return insertTransition("",pRho);
    }
    
    @Override
    public IPushdownTransition<V> pop(String pPhi) {
      return insertTransition(pPhi,"");
    }
    
    @Override
    public IPushdownTransition<V> ignore() {
      return insertTransition("","");
    }
    
    @Override
    public IPushdownTransition<V> pushOn(String pPhi, String pRho) {
      return insertTransition(pPhi,pPhi+pRho);
    }
    
    @Override
    public IPushdownTransition<V> skip(String pPhi) {
      return insertTransition(pPhi,pPhi);
    }
    
    @Override
    public IPushdownTransition<V> changeTop(String pPhi, String pRho) {
      return insertTransition(pPhi,pRho);
    }
    
    @Override
    public boolean equals(Object pObject) {
      if (!(pObject instanceof IPushdownDelta)) return false;
      IPushdownDelta<?> d=(IPushdownDelta<?>)pObject;
      return GToolkit.equals(input,d.getInput())&&prefix.equals(d.getPrefix())&&GToolkit.equals(output,d.getOutput());
    }
    
    @Override
    public int hashCode() {
      return input.hashCode()^prefix.hashCode()^output.hashCode();
    }
    
  }
  
}
