package gold.structures.bag;

import java.util.*;
import gold.structures.collection.*;
import gold.util.*;

public class GAdaptorBag<E> extends GAbstractCollection<E> implements IBag<E> {
  
  private static final long serialVersionUID=-3954870174308099072L;
  
  protected final Map<E,Integer> adaptee;
  protected int cardinality;
  
  public GAdaptorBag(Map<E,Integer> pAdaptee) {
    adaptee=pAdaptee;
    cardinality=0;
    if (!adaptee.isEmpty()) {
      List<E> invalidKeys=new ArrayList<E>();
      for (Map.Entry<E,Integer> entry:adaptee.entrySet()) {
        Integer c=entry.getValue();
        if (c!=null&&c>0) {
          cardinality+=c;
        }
        else {
          invalidKeys.add(entry.getKey());
        }
      }
      for (E invalidKey:invalidKeys) {
        adaptee.remove(invalidKey);
      }
    }
  }
  
  protected int getFrequency(Object pElement) {
    Integer c=adaptee.get(pElement);
    return c!=null?c:0;
  }
  
  protected int setFrequency(E pElement, int pFrequency, boolean pAccumulate) {
    int oldFrequency=getFrequency(pElement),newFrequency=Math.max(pFrequency+(pAccumulate?oldFrequency:0),0);
    if (newFrequency!=oldFrequency) {
      if (newFrequency==0) {
        adaptee.remove(pElement);
      }
      else {
        adaptee.put(pElement,newFrequency);
      }
      cardinality+=newFrequency-oldFrequency;
    }
    return newFrequency;
  }
  
  protected int incrementFrequency(E pElement, int pDeltaFrequency) {
    return setFrequency(pElement,pDeltaFrequency,true);
  }
  
  protected void setFrequency(E pElement, int pFrequency) {
    setFrequency(pElement,pFrequency,false);
  }
  
  @Override
  public int size() {
    return cardinality;
  }
  
  @Override
  public boolean isEmpty() {
    return cardinality==0;
  }
  
  @Override
  public void clear() {
    adaptee.clear();
    cardinality=0;
  }
  
  @Override
  public Iterator<E> iterator() {
    return toJavaView().iterator();
  }
  
  @Override
  public boolean contains(Object pElement) {
    return adaptee.containsKey(pElement);
  }
  
  @Override
  public boolean equals(Object pObject) {
    if (pObject instanceof IBag) {
      IBag<?> bag=(IBag<?>)pObject;
      return adaptee.equals(bag.toJavaMap());
    }
    else if (pObject instanceof Map) {
      return adaptee.equals(pObject);
    }
    else if (pObject instanceof ICollection) {
      ICollection<?> collection=(ICollection<?>)pObject;
      return toJavaView().equals(collection.toJavaView());
    }
    else if (pObject instanceof Collection) {
      Collection<?> collection=(Collection<?>)pObject;
      return toJavaView().equals(collection);
    }
    else {
      return super.equals(pObject);
    }
  }
  
  @Override
  public int hashCode() {
    return adaptee.hashCode();
  }
  
  @Override
  public String toString() {
    return super.toString();
  }
  
  @Override
  public Collection<E> toJavaView() {
    return new AbstractCollection<E>() {
      @Override
      public boolean add(E pElement) {
        GAdaptorBag.this.incrementFrequency(pElement,1);
        return true;
      }
      @Override
      public int size() {
        return GAdaptorBag.this.cardinality;
      }
      @Override
      public Iterator<E> iterator() {
        return new GAbstractIterator<E>() {
          private final Iterator<E> iterator=new LinkedList<E>(GAdaptorBag.this.adaptee.keySet()).iterator();
          private boolean removed=false;
          private E current=null;
          private int count=0;
          @Override
          public boolean hasNext() {
            return iterator.hasNext()||count>0;
          }
          @Override
          public E next() {
            if (count==0) {
              current=iterator.next();
              count=GAdaptorBag.this.count(current);
            }
            removed=false;
            count--;
            return current;
          }
          @Override
          public void remove() {
            if (removed) throw new IllegalStateException();
            removed=true;
            if (GAdaptorBag.this.incrementFrequency(current,-1)==0) {
              iterator.remove();
            }
          }
        };
      }
    };
  }
  
  @Override
  public Map<E,Integer> toJavaMap() {
    return adaptee;
  }
  
  @Override
  @SuppressWarnings("unchecked")
  public IBag<E> clone() {
    try {
      return new GAdaptorBag<E>((Map<E,Integer>)(GToolkit.clone(adaptee)));
    }
    catch (Throwable th) {
      throw new InternalError("The class "+adaptee.getClass().getCanonicalName()+" is not cloneable.");
    }
  }
  
  @Override
  public void add(E pElement) {
    add(pElement,1);
  }
  
  @Override
  public void add(E pElement, int pFrequency) {
    incrementFrequency(pElement,pFrequency);
  }
  
  @Override
  public void remove(E pElement) {
    remove(pElement,1);
  }
  
  @Override
  public void remove(E pElement, int pFrequency) {
    incrementFrequency(pElement,-pFrequency);
  }
  
  @Override
  public int count(Object pElement) {
    return getFrequency(pElement);
  }
  
  @Override
  public void union(IBag<? extends E> pBag) {
    for (Map.Entry<? extends E,Integer> entry:pBag.toJavaMap().entrySet()) {
      incrementFrequency(entry.getKey(),entry.getValue());
    }
  }
  
  @Override
  public void intersection(IBag<? extends E> pBag) {
    for (Map.Entry<? extends E,Integer> entry:pBag.toJavaMap().entrySet()) {
      setFrequency(entry.getKey(),Math.min(count(entry.getKey()),entry.getValue()));
    }
  }
  
  @Override
  public void difference(IBag<? extends E> pBag) {
    for (Map.Entry<? extends E,Integer> entry:pBag.toJavaMap().entrySet()) {
      incrementFrequency(entry.getKey(),-entry.getValue());
    }
  }
  
  @Override
  public void symmetricDifference(IBag<? extends E> pBag) {
    for (Map.Entry<? extends E,Integer> entry:pBag.toJavaMap().entrySet()) {
      E element=entry.getKey();
      int count1=count(element),count2=entry.getValue();
      setFrequency(element,Math.max(count1,count2)-Math.min(count1,count2));
    }
  }
  
  @Override
  public boolean isSubbag(IBag<?> pBag) {
    for (Map.Entry<? extends E,Integer> entry:adaptee.entrySet()) {
      if (entry.getValue()>pBag.count(entry.getKey())) return false;
    }
    return true;
  }
  
  @Override
  public boolean isSuperbag(IBag<?> pBag) {
    return pBag.isSubbag(this);
  }
  
  @Override
  public boolean isProperSubbag(IBag<?> pBag) {
    return size()<pBag.size()&&isSubbag(pBag);
  }
  
  @Override
  public boolean isProperSuperbag(IBag<?> pBag) {
    return pBag.isProperSubbag(this);
  }
  
  @Override
  public boolean isDisjointFrom(IBag<?> pBag) {
    for (Map.Entry<? extends E,Integer> entry:adaptee.entrySet()) {
      if (pBag.count(entry.getKey())>0) return false;
    }
    return true;
  }
  
}
