package gold.structures.graph;

import java.util.*;
import gold.structures.collection.*;
import gold.structures.list.*;
import gold.structures.set.*;
import gold.util.*;
import edu.uci.ics.jung.graph.*;
import edu.uci.ics.jung.graph.util.*;

public abstract class GAbstractGraph<V,E> implements IGraph<V,E> {
  
  private static final long serialVersionUID=8632738996927679821L;
  
  @Override
  public boolean isEmpty() {
    return getVertexCount()==0;
  }
  
  @SuppressWarnings("unchecked")
  protected IEdge<? extends V,? extends E> buildEdge(Object pObject) {
    if (pObject instanceof IEdge) {
      return (IEdge<? extends V,? extends E>)pObject;
    }
    else if ((pObject instanceof Iterable)&&GCollections.sizeOf((Iterable<?>)pObject)>=2) {
      Iterator<?> it=((Iterable<?>)pObject).iterator();
      return new GEdge<V,E>((V)(it.next()),(V)(it.next()),it.hasNext()?((Number)(it.next())).doubleValue():0.0,it.hasNext()?(E)(it.next()):null);
    }
    else {
      throw new IllegalArgumentException("Invalid edge: "+pObject);
    }
  }
  
  @Override
  public boolean isUndirected() {
    /**
     * <code>
     * if (isUndirectedAnnotationPresent()) return true;
     * for (IEdge<V,E> edge1:getEdges()) {
     *   IEdge<V,E> edge2=getEdge(edge1.getTarget(),edge1.getSource());
     *   if (edge2==null||!GToolkit.equals(edge1.getData(),edge2.getData())||edge1.getCost()!=edge2.getCost()) return false;
     * }
     * return true;
     * </code>
     */
    return getClass().isAnnotationPresent(IIsUndirected.class);
  }
  
  @Override
  public int getVertexCount() {
    return getVertices().size();
  }
  
  @Override
  public abstract ISet<V> getVertices();
  
  @Override
  public int getEdgeCount() {
    return getEdges().size();
  }
  
  @Override
  public ISet<IEdge<V,E>> getEdges() {
    return new GAdaptorSet<IEdge<V,E>>(new AbstractSet<IEdge<V,E>>() {
      @Override
      public Iterator<IEdge<V,E>> iterator() {
        return new GAbstractIterator<IEdge<V,E>>() {
          private final Iterator<V> iteratorV=getVertices().iterator();
          private Iterator<IEdge<V,E>> iteratorE=null;
          @Override
          public boolean hasNext() {
            while ((iteratorE==null||!iteratorE.hasNext())&&iteratorV.hasNext()) {
              iteratorE=getOutEdges(iteratorV.next()).iterator();
            }
            return iteratorE!=null&&iteratorE.hasNext();
          }
          @Override
          public IEdge<V,E> next() {
            return iteratorE.next();
          }
        };
      }
      @Override
      public int size() {
        int c=0;
        for (V vertex:getVertices()) {
          c+=getOutDegree(vertex);
        }
        return c;
      }
    });
  }
  
  @Override
  public IEdge<V,E> getEdge(V pSource, V pTarget) {
    double cost=getCost(pSource,pTarget);
    if (cost==IEdge.INFINITE_COST) {
      return null;
    }
    else {
      E data=getData(pSource,pTarget);
      return new GEdge<V,E>(pSource,pTarget,cost,data);
    }
  }
  
  @Override
  public abstract double getCost(V pSource, V pTarget);
  
  @Override
  public E getData(V pSource, V pTarget) {
    return null;
  }
  
  @Override
  public boolean containsVertex(V pVertex) {
    return getVertices().contains(pVertex);
  }
  
  @Override
  public boolean containsEdge(V pSource, V pTarget) {
    return getCost(pSource,pTarget)!=IEdge.INFINITE_COST;
  }
  
  @Override
  public void addEdge(IEdge<? extends V,? extends E> pEdge) {
    addEdge(pEdge.getSource(),pEdge.getTarget(),pEdge.getCost(),pEdge.getData());
  }
  
  @Override
  public void addEdge(V pSource, V pTarget) {
    addEdge(pSource,pTarget,0.0);
  }
  
  @Override
  public void addEdge(V pSource, V pTarget, double pCost) {
    addEdge(pSource,pTarget,pCost,null);
  }
  
  @Override
  public void addEdge(V pSource, V pTarget, double pCost, E pData) {
    throw new UnsupportedOperationException();
  }
  
  @Override
  public void addEdges(Iterable<?> pEdges) {
    for (Object edge:pEdges) {
      addEdge(buildEdge(edge));
    }
  }
  
  @Override
  public void removeEdge(IEdge<? extends V,? extends E> pEdge) {
    removeEdge(pEdge.getSource(),pEdge.getTarget());
  }
  
  @Override
  public void removeEdge(V pSource, V pTarget) {
    throw new UnsupportedOperationException();
  }
  
  @Override
  public void removeEdges(Iterable<?> pEdges) {
    for (Object edge:pEdges) {
      removeEdge(buildEdge(edge));
    }
  }
  
  @Override
  public abstract ISet<V> getSuccessors(V pVertex);
  
  @Override
  public ISet<V> getPredecessors(V pVertex) {
    if (isUndirected()) return getSuccessors(pVertex);
    final V node=pVertex;
    return new GAdaptorSet<V>(new AbstractSet<V>() {
      @Override
      public Iterator<V> iterator() {
        return new GAbstractIterator<V>() {
          private final Iterator<V> iteratorV=getVertices().iterator();
          private V nextV=null;
          @Override
          public boolean hasNext() {
            if (nextV!=null) return true;
            while (iteratorV.hasNext()) {
              nextV=iteratorV.next();
              if (getSuccessors(nextV).contains(node)) return true;
            }
            nextV=null;
            return false;
          }
          @Override
          public V next() {
            V result=nextV;
            nextV=null;
            return result;
          }
        };
      }
      @Override
      public int size() {
        int c=0;
        for (V vertex:getVertices()) {
          if (getSuccessors(vertex).contains(node)) c++;
        }
        return c;
      }
    });
  }
  
  @Override
  public ISet<V> getNeighbors(V pVertex) {
    if (isUndirected()) return getSuccessors(pVertex);
    ISet<V> result=new GHashTableSet<V>();
    result.union(getSuccessors(pVertex));
    result.union(getPredecessors(pVertex));
    return GCollections.unmodifiableSet(result);
  }
  
  @Override
  public ISet<IEdge<V,E>> getOutEdges(V pVertex) {
    final V node=pVertex;
    return new GAdaptorSet<IEdge<V,E>>(new AbstractSet<IEdge<V,E>>() {
      @Override
      public Iterator<IEdge<V,E>> iterator() {
        return new GAbstractIterator<IEdge<V,E>>() {
          private final Iterator<V> iteratorV=getSuccessors(node).iterator();
          @Override
          public boolean hasNext() {
            return iteratorV.hasNext();
          }
          @Override
          public IEdge<V,E> next() {
            return getEdge(node,iteratorV.next());
          }
        };
      }
      @Override
      public int size() {
        return getOutDegree(node);
      }
    });
  }
  
  @Override
  public ISet<IEdge<V,E>> getInEdges(V pVertex) {
    if (isUndirected()) return getOutEdges(pVertex);
    final V node=pVertex;
    return new GAdaptorSet<IEdge<V,E>>(new AbstractSet<IEdge<V,E>>() {
      @Override
      public Iterator<IEdge<V,E>> iterator() {
        return new GAbstractIterator<IEdge<V,E>>() {
          private final Iterator<V> iteratorV=getPredecessors(node).iterator();
          @Override
          public boolean hasNext() {
            return iteratorV.hasNext();
          }
          @Override
          public IEdge<V,E> next() {
            return getEdge(iteratorV.next(),node);
          }
        };
      }
      @Override
      public int size() {
        return getInDegree(node);
      }
    });
  }
  
  @Override
  public int getOutDegree(V pVertex) {
    return getSuccessors(pVertex).size();
  }
  
  @Override
  public int getInDegree(V pVertex) {
    if (isUndirected()) return getOutDegree(pVertex);
    return getPredecessors(pVertex).size();
  }
  
  @Override
  public double[][] getCostMatrix() {
    return getCostMatrix(getVertices());
  }
  
  @Override
  public double[][] getCostMatrix(Iterable<? extends V> pVertexOrdering) {
    if (!getVertices().toJavaView().containsAll(GCollections.asJavaCollection(pVertexOrdering))) {
      throw new IllegalArgumentException("The given ordering contains objects that aren't vertices of the graph.");
    }
    IList<V> list=new GArrayList<V>(GCollections.unmodifiableList(pVertexOrdering));
    int n=list.size();
    double matrix[][]=new double[n][n];
    for (int i=0; i<n; i++) {
      for (int j=0; j<n; j++) {
        matrix[i][j]=getCost(list.get(i),list.get(j));
      }
    }
    return matrix;
  }
  
  @Override
  public boolean[][] getAdjacencyMatrix() {
    return getAdjacencyMatrix(getVertices());
  }
  
  @Override
  public boolean[][] getAdjacencyMatrix(Iterable<? extends V> pVertexOrdering) {
    if (!getVertices().toJavaView().containsAll(GCollections.asJavaCollection(pVertexOrdering))) {
      throw new IllegalArgumentException("The given ordering contains objects that aren't vertices of the graph.");
    }
    IList<V> list=new GArrayList<V>(GCollections.unmodifiableList(pVertexOrdering));
    int n=list.size();
    boolean matrix[][]=new boolean[n][n];
    for (int i=0; i<n; i++) {
      for (int j=0; j<n; j++) {
        matrix[i][j]=containsEdge(list.get(i),list.get(j));
      }
    }
    return matrix;
  }
  
  @Override
  public boolean equals(Object pObject) {
    return (pObject instanceof IGraph)&&GToolkit.equals(getVertices(),((IGraph<?,?>)pObject).getVertices())&&GToolkit.equals(getEdges(),((IGraph<?,?>)pObject).getEdges());
  }
  
  @Override
  public int hashCode() {
    return getVertices().hashCode()^getEdges().hashCode();
  }
  
  @Override
  public String toString() {
    return getVertices().toString()+"\n"+getEdges().toString();
  }
  
  @Override
  public Graph<V,IEdge<V,E>> toJungView() {
    return toJungView(isUndirected()?EdgeType.UNDIRECTED:EdgeType.DIRECTED);
  }
  
  protected Graph<V,IEdge<V,E>> toJungView(EdgeType pEdgeType) {
    return new GViewJungGraph<V,E>(this,pEdgeType);
  }
  
  @Override
  public Graph<V,IEdge<V,E>> toJungCopy() {
    return toJungCopy(isUndirected()?EdgeType.UNDIRECTED:EdgeType.DIRECTED);
  }
  
  protected Graph<V,IEdge<V,E>> toJungCopy(EdgeType pEdgeType) {
    Graph<V,IEdge<V,E>> result=pEdgeType.equals(EdgeType.UNDIRECTED)?new UndirectedSparseGraph<V,IEdge<V,E>>():new DirectedSparseGraph<V,IEdge<V,E>>();
    for (V vertex:getVertices()) {
      result.addVertex(vertex);
    }
    for (IEdge<V,E> edge:getEdges()) {
      result.addEdge(edge,edge.getSource(),edge.getTarget(),pEdgeType);
    }
    return result;
  }
  
  @Override
  public GAbstractGraph<V,E> clone() {
    throw new UnsupportedOperationException();
  }
  
}
