package gold.structures.graph;

import java.util.*;
import gold.structures.collection.*;
import gold.structures.map.*;
import gold.structures.multimap.*;
import gold.structures.set.*;
import gold.structures.tuple.*;
import gold.util.*;

import edu.uci.ics.jung.graph.*;

public class GDirectedGraph<V,E> extends GAbstractGraph<V,E> {
  
  private static final long serialVersionUID=-511393285200798907L;
  
  protected final ISet<V> vertices;
  protected final IMap<IPair<V>,IEdge<V,E>> edges;
  protected final IMultiMap<V,IEdge<V,E>> adjacencyLists;
  
  public GDirectedGraph(Iterable<? extends V> pVertices) {
    if ((pVertices instanceof ICollection)&&!((ICollection<?>)pVertices).isFinite()) throw new IllegalArgumentException("The set of vertices of the graph cannot be infinite.");
    vertices=new GLinkedHashTableSet<V>(GCollections.unmodifiableCollection(pVertices));
    edges=new GHashTableMap<IPair<V>,IEdge<V,E>>(vertices.size()*4);
    adjacencyLists=new GHashTableMultiMap<V,IEdge<V,E>>(vertices.size());
  }
  
  public GDirectedGraph(Iterable<? extends V> pVertices, Iterable<?> pEdges) {
    this(pVertices);
    addEdges(pEdges);
  }
  
  public GDirectedGraph(IGraph<? extends V,? extends E> pGraph) {
    this(pGraph.getVertices());
    for (IEdge<? extends V,? extends E> edge:pGraph.getEdges()) {
      addEdge(edge.getSource(),edge.getTarget(),edge.getCost(),edge.getData());
    }
  }
  
  public <F extends E> GDirectedGraph(Graph<? extends V,F> pGraph) {
    this(pGraph.getVertices());
    for (F edge:pGraph.getEdges()) {
      addEdge(pGraph.getSource(edge),pGraph.getDest(edge),(edge instanceof Double)?(Double)edge:1.0,edge);
    }
  }
  
  @Override
  public boolean isEmpty() {
    return vertices.isEmpty();
  }
  
  @Override
  public int getVertexCount() {
    return vertices.size();
  }
  
  @Override
  public ISet<V> getVertices() {
    return GCollections.unmodifiableSet(vertices);
  }
  
  @Override
  public int getEdgeCount() {
    return edges.size();
  }
  
  @Override
  public ISet<IEdge<V,E>> getEdges() {
    return GCollections.unmodifiableSet(edges.values());
  }
  
  @Override
  public IEdge<V,E> getEdge(V pSource, V pTarget) {
    return edges.get(new GPair<V>(pSource,pTarget));
  }
  
  @Override
  public double getCost(V pSource, V pTarget) {
    IEdge<V,E> edge=getEdge(pSource,pTarget);
    return edge!=null?edge.getCost():IEdge.INFINITE_COST;
  }
  
  @Override
  public E getData(V pSource, V pTarget) {
    IEdge<V,E> edge=getEdge(pSource,pTarget);
    return edge!=null?edge.getData():null;
  }
  
  @Override
  public boolean containsVertex(V pVertex) {
    return vertices.contains(pVertex);
  }
  
  @Override
  public boolean containsEdge(V pSource, V pTarget) {
    return edges.containsKey(new GPair<V>(pSource,pTarget));
  }
  
  @Override
  public void addEdge(V pSource, V pTarget, double pCost, E pData) {
    IEdge<V,E> edge=new GEdge<V,E>(pSource,pTarget,pCost,pData);
    adjacencyLists.remove(pSource,edge);
    adjacencyLists.add(pSource,edge);
    edges.put(new GPair<V>(pSource,pTarget),edge);
  }
  
  @Override
  public void removeEdge(V pSource, V pTarget) {
    IEdge<V,E> edge=edges.remove(new GPair<V>(pSource,pTarget));
    if (edge!=null) adjacencyLists.remove(pSource,edge);
  }
  
  @Override
  public ISet<V> getSuccessors(V pVertex) {
    final V node=pVertex;
    return new GAdaptorSet<V>(new AbstractSet<V>() {
      @Override
      public Iterator<V> iterator() {
        return new GAbstractIterator<V>() {
          private final Iterator<IEdge<V,E>> iteratorE=adjacencyLists.get(node).iterator();
          @Override
          public boolean hasNext() {
            return iteratorE.hasNext();
          }
          @Override
          public V next() {
            return iteratorE.next().getTarget();
          }
        };
      }
      @Override
      public int size() {
        return adjacencyLists.get(node).size();
      }
    });
  }
  
  @Override
  public ISet<IEdge<V,E>> getOutEdges(V pVertex) {
    return GCollections.unmodifiableSet(adjacencyLists.get(pVertex));
  }
  
  @Override
  public int getOutDegree(V pVertex) {
    return adjacencyLists.get(pVertex).size();
  }
  
  @Override
  public GDirectedGraph<V,E> clone() {
    return new GDirectedGraph<V,E>(this);
  }
  
}
