package gold.structures.heap;

import java.util.*;
import gold.structures.collection.*;
import gold.structures.map.*;
import com.mhhe.clrs2e.*;

public class GBinomialHeap<E> extends GAbstractCollection<E> implements IHeap<E> {
  
  private static final long serialVersionUID=5788516415444841779L;
  
  protected BinomialHeap adaptee;
  protected Map<E,BinomialHeapNode<E>> nodes;
  
  public GBinomialHeap() {
    this(1024);
  }
  
  public GBinomialHeap(int pInitialCapacity) {
    adaptee=new BinomialHeap();
    nodes=new HashMap<E,BinomialHeapNode<E>>(pInitialCapacity);
  }
  
  protected GBinomialHeap(GBinomialHeap<E> pHeap) {
    this(pHeap.nodes.size());
    for (BinomialHeapNode<E> node:pHeap.nodes.values()) {
      insert(node.getData(),node.getKey());
    }
  }
  
  public GBinomialHeap(Map<E,?> pElements) {
    this(pElements.size());
    for (Map.Entry<E,?> entry:pElements.entrySet()) {
      insert(entry.getKey(),((Number)(entry.getValue())).doubleValue());
    }
  }
  
  public GBinomialHeap(IMap<E,?> pElements) {
    this(pElements.toJavaView());
  }
  
  @Override
  public void insert(E pElement, double pKey) {
    BinomialHeapNode<E> node=nodes.get(pElement);
    if (node==null) {
      node=new BinomialHeapNode<E>(pElement,pKey);
      node.handle=adaptee.insert(node);
      nodes.put(pElement,node);
    }
    else if (pKey>node.getKey()) {
      throw new IllegalArgumentException("Cannot increase the key of the element '"+pElement+"'.");
    }
    else if (pKey<node.getKey()) {
      adaptee.decreaseKey(node.handle,pKey);
    }
  }
  
  @Override
  public void decreaseKey(E pElement, double pKey) {
    BinomialHeapNode<E> node=nodes.get(pElement);
    if (node==null) {
      throw new IllegalArgumentException("The element '"+pElement+"' doesn't belong to the heap.");
    }
    else if (pKey>node.getKey()) {
      throw new IllegalArgumentException("Cannot increase the key of the element '"+pElement+"'.");
    }
    else if (pKey<node.getKey()) {
      adaptee.decreaseKey(node.handle,pKey);
    }
  }
  
  @Override
  public void delete(E pElement) {
    BinomialHeapNode<E> node=nodes.remove(pElement);
    if (node==null) {
      throw new IllegalArgumentException("The element '"+pElement+"' doesn't belong to the heap.");
    }
    else {
      adaptee.delete(node.handle);
    }
  }
  
  @Override
  public boolean contains(Object pElement) {
    return nodes.containsKey(pElement);
  }
  
  @Override
  @SuppressWarnings("unchecked")
  public E minimum() {
    if (isEmpty()) throw new NoSuchElementException();
    return ((BinomialHeapNode<E>)(adaptee.minimum())).getData();
  }
  
  @Override
  @SuppressWarnings("unchecked")
  public double minimumKey() {
    if (isEmpty()) throw new NoSuchElementException();
    return ((BinomialHeapNode<E>)(adaptee.minimum())).getKey();
  }
  
  @Override
  @SuppressWarnings("unchecked")
  public E extractMinimum() {
    if (isEmpty()) throw new NoSuchElementException();
    BinomialHeapNode<E> node=(BinomialHeapNode<E>)(adaptee.extractMin());
    nodes.remove(node.getData());
    return node.getData();
  }
  
  @Override
  public int size() {
    return nodes.size();
  }
  
  @Override
  public boolean isEmpty() {
    return nodes.isEmpty();
  }
  
  @Override
  public void clear() {
    adaptee=new BinomialHeap();
    nodes.clear();
  }
  
  @Override
  public Iterator<E> iterator() {
    return nodes.keySet().iterator();
  }
  
  @Override
  public Collection<E> toJavaView() {
    return nodes.keySet();
  }
  
  @Override
  public IHeap<E> clone() {
    return new GBinomialHeap<E>(this);
  }
  
  protected static class BinomialHeapNode<E> implements DynamicSetElement {
    
    protected final E data;
    protected double key;
    protected Object handle=null;
    
    public BinomialHeapNode(E pData, double pKey) {
      data=pData;
      key=pKey;
    }
    
    public E getData() {
      return data;
    }
    
    @Override
    public Double getKey() {
      return key;
    }
    
    @Override
    @SuppressWarnings("rawtypes")
    public void setKey(Comparable pKey) {
      key=(Double)pKey;
    }
    
    @Override
    public int compareTo(Object pObject) {
      return DynamicSetElement.Helper.compareTo(this,pObject);
    }
    
  }
  
}
