package gold.structures.map;

import java.io.*;
import java.util.*;
import gold.util.*;

// Basado en una implementación de Alejandro Sotelo para el Politécnico Grancolombiano.

public class GAVLTreeMap<K,V> extends GAdaptorMap<K,V> {
  
  private static final long serialVersionUID=3384216838183089232L;
  
  public GAVLTreeMap(Comparator<? super K> pComparator) {
    super(new AVLTree<K,V>(pComparator));
  }
  
  public GAVLTreeMap() {
    this(GComparators.<K>genericComparator());
  }
  
  public GAVLTreeMap(Map<? extends K,? extends V> pMap) {
    this();
    toJavaView().putAll(pMap);
  }
  
  public GAVLTreeMap(IMap<? extends K,? extends V> pMap) {
    this(pMap.toJavaView());
  }
  
  @SuppressWarnings("unchecked")
  public static class AVLTree<K,V> extends AbstractMap<K,V> implements Cloneable,Serializable {
    
    private static final long serialVersionUID=7263785976168031497L;
    
    protected final Comparator<? super K> comparator;
    protected K key=null;
    protected V value=null;
    protected AVLTree<K,V> left=null;
    protected AVLTree<K,V> right=null;
    protected int heightAVL=0;
    
    public AVLTree(Comparator<? super K> pComparator) {
      comparator=pComparator;
    }
    
    public AVLTree(AVLTree<K,V> pMap) {
      comparator=pMap.comparator;
      putAll(pMap);
    }
    
    public AVLTree(Map<? extends K,? extends V> pMap) {
      if (pMap instanceof SortedMap) {
        comparator=((SortedMap<K,V>)pMap).comparator();
      }
      else {
        comparator=GComparators.<K>genericComparator();
      }
      putAll(pMap);
    }
    
    @Override
    public boolean isEmpty() {
      return key==null;
    }
    
    @Override
    public void clear() {
      key=null;
      value=null;
      left=null;
      right=null;
    }
    
    @Override
    public int size() {
      return isEmpty()?0:1+left.size()+right.size();
    }
    
    protected int height() {
      return isEmpty()?0:1+Math.max(left.height(),right.height());
    }
    
    protected boolean isLeaf() {
      return !isEmpty()&&left.isEmpty()&&right.isEmpty();
    }
    
    protected AVLTree<K,V> max() {
      return isEmpty()?null:(right.isEmpty()?this:right.max());
    }
    
    @Override
    public Set<Map.Entry<K,V>> entrySet() {
      final Collection<Map.Entry<K,V>> result=inorderEntries(new ArrayList<Map.Entry<K,V>>(size()));
      return new AbstractSet<Map.Entry<K,V>>() {
        @Override
        public Iterator<Map.Entry<K,V>> iterator() {
          return result.iterator();
        }
        @Override
        public int size() {
          return result.size();
        }
        @Override
        public boolean contains(Object pObject) {
          return result.contains(pObject);
        }
      };
    }
    
    protected Collection<Map.Entry<K,V>> inorderEntries(Collection<Map.Entry<K,V>> pCollection) {
      if (!isEmpty()) {
        left.inorderEntries(pCollection);
        pCollection.add(new AVLEntryMap<K,V>(this));
        right.inorderEntries(pCollection);
      }
      return pCollection;
    }
    
    @Override
    public boolean containsKey(Object pKey) {
      if (isEmpty()) {
        return false;
      }
      else {
        int c=comparator.compare((K)pKey,key);
        return c==0||(c<0?left:right).containsKey(pKey);
      }
    }
    
    @Override
    public boolean containsValue(Object pValue) {
      if (isEmpty()) {
        return false;
      }
      else {
        return GToolkit.equals(value,pValue)||left.containsValue(pValue)||right.containsValue(pValue);
      }
    }
    
    @Override
    public V get(Object pKey) {
      if (isEmpty()) {
        return null;
      }
      else {
        int c=comparator.compare((K)pKey,key);
        return c==0?value:(c<0?left:right).get(pKey);
      }
    }
    
    @Override
    public V put(K pKey, V pValue) {
      V previous=null;
      if (isEmpty()) {
        key=pKey;
        value=pValue;
        left=new AVLTree<K,V>(comparator);
        right=new AVLTree<K,V>(comparator);
      }
      else {
        int c=comparator.compare(pKey,key);
        if (c==0) {
          previous=value;
          value=pValue;
        }
        else {
          previous=(c<0?left:right).put(pKey,pValue);
        }
      }
      balance();
      return previous;
    }
    
    @Override
    public void putAll(Map<? extends K,? extends V> pMap) {
      for (Map.Entry<? extends K,? extends V> entry:pMap.entrySet()) {
        put(entry.getKey(),entry.getValue());
      }
    }
    
    @Override
    public V remove(Object pKey) {
      V previous=null;
      if (!isEmpty()) {
        int c=comparator.compare((K)pKey,key);
        if (c==0) {
          previous=value;
          if (isLeaf()) {
            clear();
          }
          else if (left.isEmpty()) {
            AVLTree<K,V> tmp=right;
            key=tmp.key;
            value=tmp.value;
            left=tmp.left;
            right=tmp.right;
          }
          else {
            AVLTree<K,V> tmp=left.max();
            key=tmp.key;
            value=tmp.value;
            left.remove(key);
          }
        }
        else {
          previous=(c<0?left:right).remove(pKey);
        }
      }
      balance();
      return previous;
    }
    
    protected void updateHeight() {
      heightAVL=isEmpty()?0:1+Math.max(left.heightAVL,right.heightAVL);
    }
    
    protected void rotateLeft() {
      K tmp=key;
      key=right.key;
      right.key=tmp;
      AVLTree<K,V> tree1=this,tree2=right,a=left,b=right.left,c=right.right;
      tree1.left=tree2;
      tree1.right=c;
      tree2.left=a;
      tree2.right=b;
      left.updateHeight();
      updateHeight();
    }
    
    protected void rotateRight() {
      K tmp=key;
      key=left.key;
      left.key=tmp;
      AVLTree<K,V> tree1=this,tree2=left,a=left.left,b=left.right,c=right;
      tree1.left=a;
      tree1.right=tree2;
      tree2.left=b;
      tree2.right=c;
      right.updateHeight();
      updateHeight();
    }
    
    protected void balance() {
      if (!isEmpty()) {
        int factor=right.heightAVL-left.heightAVL;
        if (factor==-2) {
          if (left.left.heightAVL>left.right.heightAVL) {
            rotateRight();
          }
          else {
            left.rotateLeft();
            rotateRight();
          }
        }
        else if (factor==+2) {
          if (right.right.heightAVL>right.left.heightAVL) {
            rotateLeft();
          }
          else {
            right.rotateRight();
            rotateLeft();
          }
        }
      }
      updateHeight();
    }
    
    @Override
    public AVLTree<K,V> clone() {
      return new AVLTree<K,V>(this);
    }
    
    // Basado en java.util.TreeMap.Entry
    public static class AVLEntryMap<K,V> implements Map.Entry<K,V>,Serializable {
      
      private static final long serialVersionUID=-9091948189477971215L;
      
      protected final AVLTree<K,V> tree;
      
      public AVLEntryMap(AVLTree<K,V> pTree) {
        tree=pTree;
      }
      
      @Override
      public K getKey() {
        return tree.key;
      }
      
      @Override
      public V getValue() {
        return tree.value;
      }
      
      @Override
      public V setValue(V pValue) {
        V oldValue=getValue();
        tree.value=pValue;
        return oldValue;
      }
      
      @Override
      public boolean equals(Object pObject) {
        if (!(pObject instanceof Map.Entry)) return false;
        Map.Entry<?,?> entry=(Map.Entry<?,?>)pObject;
        return GToolkit.equals(getKey(),entry.getKey())&&GToolkit.equals(getValue(),entry.getValue());
      }
      
      @Override
      public int hashCode() {
        K key=getKey();
        V value=getValue();
        return (key==null?0:key.hashCode())^(value==null?0:value.hashCode());
      }
      
      @Override
      public String toString() {
        return getKey()+"="+getValue();
      }
      
    }
    
  }
  
}
