package gold.structures.set;

import java.io.*;
import java.util.*;
import gold.structures.collection.*;
import gold.util.*;

// Basado en una implementación de Alejandro Sotelo para el Politécnico Grancolombiano.

public class GAVLTreeSet<E> extends GAdaptorSet<E> {
  
  private static final long serialVersionUID=-3433899492806706635L;
  
  public GAVLTreeSet(Comparator<? super E> pComparator) {
    super(new AVLTree<E>(pComparator));
  }
  
  public GAVLTreeSet() {
    this(GComparators.<E>genericComparator());
  }
  
  public GAVLTreeSet(Collection<? extends E> pCollection) {
    this();
    toJavaView().addAll(pCollection);
  }
  
  public GAVLTreeSet(ICollection<? extends E> pCollection) {
    this(pCollection.toJavaView());
    if (!pCollection.isFinite()) complement();
  }
  
  @SuppressWarnings("unchecked")
  public static class AVLTree<E> extends AbstractSet<E> implements Cloneable,Serializable {
    
    private static final long serialVersionUID=1689766976416924262L;
    
    protected final Comparator<? super E> comparator;
    protected E key=null;
    protected AVLTree<E> left=null;
    protected AVLTree<E> right=null;
    protected int heightAVL=0;
    
    public AVLTree(Comparator<? super E> pComparator) {
      comparator=pComparator;
    }
    
    public AVLTree(AVLTree<E> pSet) {
      comparator=pSet.comparator;
      addAll(pSet);
    }
    
    public AVLTree(Collection<? extends E> pCollection) {
      if (pCollection instanceof SortedSet) {
        comparator=((SortedSet<E>)pCollection).comparator();
      }
      else {
        comparator=GComparators.<E>genericComparator();
      }
      addAll(pCollection);
    }
    
    @Override
    public boolean isEmpty() {
      return key==null;
    }
    
    @Override
    public void clear() {
      key=null;
      left=null;
      right=null;
    }
    
    @Override
    public int size() {
      return isEmpty()?0:1+left.size()+right.size();
    }
    
    protected int height() {
      return isEmpty()?0:1+Math.max(left.height(),right.height());
    }
    
    protected boolean isLeaf() {
      return !isEmpty()&&left.isEmpty()&&right.isEmpty();
    }
    
    protected E max() {
      return isEmpty()?null:(right.isEmpty()?key:right.max());
    }
    
    @Override
    public Iterator<E> iterator() {
      return inorder(new ArrayList<E>(size())).iterator();
    }
    
    protected List<E> inorder(List<E> pList) {
      if (!isEmpty()) {
        left.inorder(pList);
        pList.add(key);
        right.inorder(pList);
      }
      return pList;
    }
    
    @Override
    public boolean contains(Object pElement) {
      if (isEmpty()) {
        return false;
      }
      else {
        int c=comparator.compare((E)pElement,key);
        return c==0||(c<0?left:right).contains(pElement);
      }
    }
    
    @Override
    public boolean add(E pElement) {
      boolean modified=false;
      if (isEmpty()) {
        key=pElement;
        left=new AVLTree<E>(comparator);
        right=new AVLTree<E>(comparator);
        modified=true;
      }
      else {
        int c=comparator.compare(pElement,key);
        modified=c!=0&&(c<0?left:right).add(pElement);
      }
      balance();
      return modified;
    }
    
    @Override
    public boolean remove(Object pElement) {
      boolean modified=false;
      if (isEmpty()) {
        modified=false;
      }
      else {
        int c=comparator.compare((E)pElement,key);
        if (c==0) {
          if (isLeaf()) {
            clear();
          }
          else if (left.isEmpty()) {
            AVLTree<E> tmp=right;
            key=tmp.key;
            left=tmp.left;
            right=tmp.right;
          }
          else {
            key=left.max();
            left.remove(key);
          }
          modified=true;
        }
        else {
          modified=(c<0?left:right).remove(pElement);
        }
      }
      balance();
      return modified;
    }
    
    protected void updateHeight() {
      heightAVL=isEmpty()?0:1+Math.max(left.heightAVL,right.heightAVL);
    }
    
    protected void rotateLeft() {
      E tmp=key;
      key=right.key;
      right.key=tmp;
      AVLTree<E> tree1=this,tree2=right,a=left,b=right.left,c=right.right;
      tree1.left=tree2;
      tree1.right=c;
      tree2.left=a;
      tree2.right=b;
      left.updateHeight();
      updateHeight();
    }
    
    protected void rotateRight() {
      E tmp=key;
      key=left.key;
      left.key=tmp;
      AVLTree<E> tree1=this,tree2=left,a=left.left,b=left.right,c=right;
      tree1.left=a;
      tree1.right=tree2;
      tree2.left=b;
      tree2.right=c;
      right.updateHeight();
      updateHeight();
    }
    
    protected void balance() {
      if (!isEmpty()) {
        int factor=right.heightAVL-left.heightAVL;
        if (factor==-2) {
          if (left.left.heightAVL>left.right.heightAVL) {
            rotateRight();
          }
          else {
            left.rotateLeft();
            rotateRight();
          }
        }
        else if (factor==+2) {
          if (right.right.heightAVL>right.left.heightAVL) {
            rotateLeft();
          }
          else {
            right.rotateRight();
            rotateLeft();
          }
        }
      }
      updateHeight();
    }
    
    @Override
    public AVLTree<E> clone() {
      return new AVLTree<E>(this);
    }
    
  }
  
}
