package gold.structures.tree.nary;

import java.util.*;
import gold.structures.automaton.*;
import gold.structures.collection.*;
import gold.structures.list.*;
import gold.structures.queue.*;
import gold.structures.stack.*;
import gold.util.*;

public class GTrie extends GAbstractNaryTree<String> {
  
  private static final long serialVersionUID=-4705501154961861969L;
  
  public static final char ROOT_SYMBOL='\0';
  
  protected final IAlphabet alphabet;
  protected final GTrie[] children;
  protected char symbol;
  protected boolean mark=false;
  
  public GTrie(Iterable<Character> pAlphabet) {
    this(new GAlphabet(pAlphabet),ROOT_SYMBOL);
  }
  
  public GTrie(Iterable<Character> pAlphabet, Iterable<String> pIterable) {
    this(pAlphabet);
    toJavaView().addAll(GCollections.asJavaCollection(pIterable));
  }
  
  protected GTrie(IAlphabet pAlphabet, char pSymbol) {
    alphabet=pAlphabet;
    children=new GTrie[alphabet.size()];
    symbol=pSymbol;
    mark=false;
  }
  
  public boolean isRoot() {
    return symbol==ROOT_SYMBOL;
  }
  
  public boolean isMarked() {
    return mark;
  }
  
  @Override
  public int size() {
    int c=0;
    IQueue<GTrie> queue=new GArrayQueue<GTrie>(64);
    queue.enqueue(this);
    while (!queue.isEmpty()) {
      GTrie trie=queue.dequeue();
      if (trie.mark) c++;
      for (GTrie child:trie.children) {
        if (child!=null) queue.enqueue(child);
      }
    }
    return c;
  }
  
  @Override
  public boolean isEmpty() {
    return !mark&&isRoot()&&getSubtrees().isEmpty();
  }
  
  @Override
  public void clear() {
    if (!isRoot()) throw new UnsupportedOperationException();
    GArrays.fill(children,null);
    mark=false;
  }
  
  @Override
  public Iterator<String> iterator() {
    return new GAbstractIterator<String>() {
      private final char[] string=new char[Math.max(1,GTrie.this.getHeight())];
      private final IStack<GTrie> stack=new GArrayStack<GTrie>(string.length);
      private int length=0;
      private GTrie nextTrie=null;
      {
        stack.push(GTrie.this);
      }
      @Override
      public boolean hasNext() {
        if (nextTrie!=null) return true;
        while (!stack.isEmpty()) {
          nextTrie=stack.pop();
          if (nextTrie==null) {
            length--;
          }
          else {
            string[length++]=nextTrie.symbol;
            stack.push(null);
            for (int index=nextTrie.children.length-1; index>=0; index--) {
              if (nextTrie.children[index]!=null) stack.push(nextTrie.children[index]);
            }
            if (nextTrie.mark) return true;
          }
        }
        nextTrie=null;
        return false;
      }
      @Override
      public String next() {
        nextTrie=null;
        return new String(string,1,length-1);
      }
    };
  }
  
  @Override
  public boolean contains(Object pWord) {
    if (!isRoot()) throw new UnsupportedOperationException();
    if (!(pWord instanceof String)) return false;
    GTrie trie=this;
    for (char c:((String)pWord).toCharArray()) {
      int k=trie.alphabet.indexOf(c);
      if (k==-1||trie.children[k]==null) return false;
      trie=trie.children[k];
    }
    return trie.mark;
  }
  
  public boolean add(String pWord) {
    if (!isRoot()) throw new UnsupportedOperationException();
    GTrie trie=this;
    for (char c:pWord.toCharArray()) {
      int k=trie.alphabet.indexOf(c);
      if (k==-1) return false;
      if (trie.children[k]==null) trie.children[k]=new GTrie(trie.alphabet,c);
      trie=trie.children[k];
    }
    if (trie.mark) {
      return false;
    }
    else {
      trie.mark=true;
      return true;
    }
  }
  
  public boolean remove(String pWord) {
    if (!isRoot()) throw new UnsupportedOperationException();
    GTrie trie=this;
    if (!trie.contains(pWord)) return false;
    IStack<GTrie> stack=new GArrayStack<GTrie>(pWord.length());
    stack.push(trie);
    for (char c:pWord.toCharArray()) {
      trie=trie.children[trie.alphabet.indexOf(c)];
      stack.push(trie);
    }
    trie.mark=false;
    while (stack.size()>=2&&!stack.peek().mark&&stack.peek().isLeaf()) {
      trie=stack.pop();
      stack.peek().children[stack.peek().alphabet.indexOf(trie.symbol)]=null;
    }
    return true;
  }
  
  @Override
  public int hashCode() {
    return toJavaView().hashCode();
  }
  
  @Override
  public String toString() {
    return toJavaView().toString();
  }
  
  @Override
  public Set<String> toJavaView() {
    return new AbstractSet<String>() {
      @Override
      public Iterator<String> iterator() {
        return GTrie.this.iterator();
      }
      @Override
      public int size() {
        return GTrie.this.size();
      }
      @Override
      public boolean add(String pElement) {
        return GTrie.this.add(pElement);
      }
      @Override
      public boolean remove(Object pElement) {
        return GTrie.this.remove((String)pElement);
      }
    };
  }
  
  @Override
  public GTrie clone() {
    return new GTrie(alphabet,this);
  }
  
  @Override
  public String getRoot() {
    return String.valueOf(symbol);
  }
  
  @Override
  public IList<INaryTree<String>> getSubtrees() {
    return new GAdaptorList<INaryTree<String>>(new AbstractList<INaryTree<String>>() {
      @Override
      public Iterator<INaryTree<String>> iterator() {
        return new GAbstractIterator<INaryTree<String>>() {
          private int index=0;
          @Override
          public boolean hasNext() {
            while (index<GTrie.this.children.length&&GTrie.this.children[index]==null) {
              index++;
            }
            return index<GTrie.this.children.length;
          }
          @Override
          public INaryTree<String> next() {
            return GTrie.this.children[index++];
          }
        };
      }
      @Override
      public INaryTree<String> get(int pIndex) {
        int index=0;
        for (GTrie child:GTrie.this.children) {
          if (child!=null) {
            if (index++==pIndex) return child;
          }
        }
        throw new IndexOutOfBoundsException();
      }
      @Override
      public int size() {
        int index=0;
        for (GTrie child:GTrie.this.children) {
          if (child!=null) index++;
        }
        return index;
      }
    });
  }
  
  @Override
  public void setRoot(String pRoot) {
    throw new UnsupportedOperationException();
  }
  
  @Override
  public void setSubtrees(IList<INaryTree<String>> pSubtrees) {
    throw new UnsupportedOperationException();
  }
  
  @Override
  public void addSubtree(INaryTree<String> pSubtree) {
    throw new UnsupportedOperationException();
  }
  
  @Override
  public void addSubtree(int pIndex, INaryTree<String> pSubtree) {
    throw new UnsupportedOperationException();
  }
  
  @Override
  public void removeSubtree(int pIndex) {
    throw new UnsupportedOperationException();
  }
  
  @Override
  public INaryTree<String> getSubtree(int pIndex) {
    return getSubtrees().get(pIndex);
  }
  
}
