package gold.swing.util;

import java.awt.*;
import java.awt.event.*;
import java.text.*;
import java.util.*;
import java.util.List;
import java.util.logging.*;
import javax.swing.*;
import javax.swing.border.*;
import gold.swing.look.*;

public final class GUtilities {
  
  public static final String OPERATIVE_SYSTEM=System.getProperty("os.name");
  public static final Locale DEFAULT_LOCALE=Locale.US;
  public static final Collator DEFAULT_COLLATOR=Collator.getInstance(DEFAULT_LOCALE);
  public static final Map<String,ImageIcon> ICONS=new HashMap<String,ImageIcon>(256);
  
  private GUtilities() {
  }
  
  public static void installGoldLookAndFeel() {
    try {
      if (!UIManager.getLookAndFeel().getName().equals("GoldLookAndFeel")) {
        // Look and Feel
        Locale.setDefault(DEFAULT_LOCALE);
        JFrame.setDefaultLookAndFeelDecorated(true);
        JDialog.setDefaultLookAndFeelDecorated(true);
        UIManager.setLookAndFeel(new GLookAndFeel());
      }
    }
    catch (Throwable th) {
      th.printStackTrace();
    }
  }
  
  public static Set<String> getUIManagerProperties() {
    Set<String> set=new TreeSet<String>();
    Enumeration<?> enumeration=UIManager.getLookAndFeelDefaults().keys();
    while (enumeration.hasMoreElements()) {
      set.add(enumeration.nextElement().toString());
    }
    return set;
  }
  
  public static Logger getLogger() {
    return Logger.getLogger("gold");
  }
  
  public static ImageIcon getIcon(String pName) {
    try {
      ImageIcon icon=ICONS.get(pName);
      if (icon!=null) return icon;
      String path="/gold/swing/icon/"+pName;
      if (GUtilities.class.getResource(".")!=null) path="../icon/"+pName;
      ICONS.put(pName,icon=new ImageIcon(GUtilities.class.getResource(path)));
      return icon;
    }
    catch (Throwable th) {
      getLogger().severe("Icon not found: "+pName);
      return null;
    }
  }
  
  public static DecimalFormat getDecimalFormat(String pPattern) {
    DecimalFormatSymbols symbols=new DecimalFormatSymbols(DEFAULT_LOCALE);
    symbols.setDecimalSeparator('.');
    return new DecimalFormat(pPattern,symbols);
  }
  
  public static Window getWindowForComponent(Component pComponent) {
    for (Component component=pComponent; component!=null; component=component.getParent()) {
      if (component instanceof Window) return (Window)component;
    }
    return JOptionPane.getFrameForComponent(pComponent);
  }
  
  public static Graphics2D setRenderingHints(Graphics pGraphics) {
    Graphics2D graphics=(Graphics2D)pGraphics;
    graphics.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION,RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
    graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
    graphics.setRenderingHint(RenderingHints.KEY_COLOR_RENDERING,RenderingHints.VALUE_COLOR_RENDER_QUALITY);
    graphics.setRenderingHint(RenderingHints.KEY_DITHERING,RenderingHints.VALUE_DITHER_ENABLE);
    graphics.setRenderingHint(RenderingHints.KEY_FRACTIONALMETRICS,RenderingHints.VALUE_FRACTIONALMETRICS_DEFAULT);
    graphics.setRenderingHint(RenderingHints.KEY_INTERPOLATION,RenderingHints.VALUE_INTERPOLATION_BICUBIC);
    graphics.setRenderingHint(RenderingHints.KEY_RENDERING,RenderingHints.VALUE_RENDER_QUALITY);
    graphics.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL,RenderingHints.VALUE_STROKE_PURE);
    graphics.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
    return graphics;
  }
  
  public static String escapeHTML(String pText) {
    StringBuffer sb=new StringBuffer((int)(pText.length()*1.5+1));
    for (int i=0,t=pText.length(); i<t; i++) {
      boolean b=i==0||(i+1<t&&pText.charAt(i+1)==' ');
      char c=pText.charAt(i);
      switch (c) {
        case ' ':
          sb.append(b?"&nbsp;":" ");
          break;
        case '<':
          sb.append("&lt;");
          break;
        case '>':
          sb.append("&gt;");
          break;
        case '&':
          sb.append("&amp;");
          break;
        case '\"':
          sb.append("&quot;");
          break;
        case '\r':
          break;
        case '\n':
          sb.append("<br/>");
          break;
        default:
          if (c>'\u007F') {
            sb.append("&#");
            sb.append((int)c);
            sb.append(";");
          }
          else {
            sb.append(c);
          }
          break;
      }
    }
    if (pText.isEmpty()) sb.append("&nbsp;");
    return sb.toString();
  }
  
  public static void configure(JDialog pDialog, JPanel pPanel) {
    pDialog.setResizable(false);
    pDialog.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
    JPanel panel=new JPanel(new BorderLayout());
    panel.setBorder(BorderFactory.createEmptyBorder(10,10,10,10));
    panel.add(pPanel,BorderLayout.CENTER);
    pDialog.setContentPane(panel);
  }
  
  public static void center(Window pWindow) {
    Dimension d=Toolkit.getDefaultToolkit().getScreenSize();
    pWindow.setLocation(new Point((d.width-pWindow.getWidth())/2,(d.height-pWindow.getHeight())/2));
  }
  
  public static void show(Window pWindow) {
    GUtilities.show(pWindow,null);
  }
  
  public static void show(Window pWindow, Component pParentComponent) {
    pWindow.pack();
    Component parent=pParentComponent!=null?getWindowForComponent(pParentComponent):null;
    if (parent!=null&&parent.isShowing()) {
      Dimension d=parent.getSize();
      Dimension s=Toolkit.getDefaultToolkit().getScreenSize();
      Point q=parent.getLocationOnScreen();
      Point p=new Point(q.x+(d.width-pWindow.getWidth())/2,q.y+(d.height-pWindow.getHeight())/2);
      pWindow.setLocation(Math.max(0,Math.min(s.width-pWindow.getWidth(),p.x)),Math.max(0,Math.min(s.height-pWindow.getHeight(),p.y)));
    }
    else {
      center(pWindow);
    }
    pWindow.setVisible(true);
    pWindow.requestFocus();
  }
  
  public static void resize(Dimension pDimension, Component...pComponents) {
    for (Component component:pComponents) {
      if (component!=null) {
        component.setPreferredSize(pDimension);
        component.setMinimumSize(pDimension);
        component.setMaximumSize(pDimension);
        component.setSize(pDimension);
      }
    }
  }
  
  public static void enable(boolean pEnabled, Component...pComponents) {
    for (Component component:pComponents) {
      if (component!=null) {
        if (component instanceof JPanel) {
          Border border=((JPanel)component).getBorder();
          if (border!=null&&(border instanceof TitledBorder)) ((TitledBorder)border).setTitleColor(pEnabled?SystemColor.textText:SystemColor.textInactiveText);
        }
        component.setEnabled(pEnabled);
      }
    }
  }
  
  public static JMenu getMenu(ActionListener pActionListener, String pTitle, char pMnemonic, Icon pIcon, String pActionCommand, JMenuItem[] pMenuItems) {
    JMenu menu=(JMenu)(configureMenuItem(pActionListener,pMnemonic,new JMenu(pTitle),pIcon,pActionCommand));
    for (JMenuItem mi:pMenuItems) {
      if (mi==null) menu.addSeparator();
      if (mi!=null) menu.add(mi);
    }
    return menu;
  }
  
  public static JMenuItem getMenuItem(ActionListener pActionListener, String pTitle, char pMnemonic, Icon pIcon, String pActionCommand) {
    return configureMenuItem(pActionListener,pMnemonic,new JMenuItem(pTitle),pIcon,pActionCommand);
  }
  
  public static JMenuItem getRadioButtonMenuItem(ActionListener pActionListener, String pTitle, char pMnemonic, String pActionCommand) {
    return configureMenuItem(pActionListener,pMnemonic,new JRadioButtonMenuItem(pTitle,false),null,pActionCommand);
  }
  
  public static JMenuItem getCheckBoxMenuItem(ActionListener pActionListener, String pTitle, char pMnemonic, String pActionCommand) {
    return configureMenuItem(pActionListener,pMnemonic,new JCheckBoxMenuItem(pTitle,false),null,pActionCommand);
  }
  
  private static JMenuItem configureMenuItem(ActionListener pActionListener, char pMnemonic, JMenuItem pMenuItem, Icon pIcon, String pActionCommand) {
    pMenuItem.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
    if (pIcon!=null) pMenuItem.setIcon(pIcon);
    if (pMnemonic!='\0') pMenuItem.setMnemonic(pMnemonic);
    if (pActionCommand!=null) pMenuItem.setActionCommand(pActionCommand);
    if (pActionListener!=null) pMenuItem.addActionListener(pActionListener);
    return pMenuItem;
  }
  
  public static JPanel getBoxPaneX(int pSpacing, Component...pComponents) {
    return getBoxPane(BoxLayout.X_AXIS,pSpacing,pComponents);
  }
  
  public static JPanel getBoxPaneY(int pSpacing, Component...pComponents) {
    return getBoxPane(BoxLayout.Y_AXIS,pSpacing,pComponents);
  }
  
  private static JPanel getBoxPane(int pAxis, int pSpacing, Component...pComponents) {
    JPanel panel=new JPanel();
    panel.setLayout(new BoxLayout(panel,pAxis));
    for (Component component:pComponents) {
      if (component!=null) {
        if (pSpacing>0&&panel.getComponentCount()>0) panel.add(pAxis==BoxLayout.X_AXIS?Box.createHorizontalStrut(pSpacing):Box.createVerticalStrut(pSpacing));
        panel.add(component);
      }
    }
    return panel;
  }
  
  public static JPanel getGridPane(int pSpacing, int pNumColumns, Component...pComponents) {
    List<Component> components=Arrays.asList(pComponents);
    List<JPanel> rows=new ArrayList<JPanel>(1+components.size()/pNumColumns);
    for (int i=0; i<components.size(); i+=pNumColumns) {
      JPanel panel=getBoxPaneX(pSpacing,components.subList(i,Math.min(i+pNumColumns,components.size())).toArray(new Component[0]));
      panel.setAlignmentX(0.0f);
      rows.add(panel);
    }
    return getBoxPaneY(pSpacing,rows.toArray(new Component[0]));
  }
  
  public static JPanel getFlowPane(int pAlignment, Component...pComponents) {
    JPanel panel=new JPanel(new FlowLayout(pAlignment,0,0));
    for (Component component:pComponents) {
      if (component!=null) panel.add(component);
    }
    return panel;
  }
  
  public static JPanel getGridBagPane(Object[][] pComponents, int pSpacing, double[] pWeights) {
    JPanel panel=new JPanel(new GridBagLayout());
    panel.setBorder(BorderFactory.createEmptyBorder(0,0,pSpacing,0));
    for (int i=0,t=pComponents.length; i<=t; i++) {
      Object[] row=i<t?pComponents[i]:new Object[]{new JSeparator(SwingConstants.HORIZONTAL)};
      double weight=pWeights!=null&&i<pWeights.length?pWeights[i]:0.0;
      for (int j=0,u=row.length; j<u; j++) {
        Component component=(Component)(row[j]==null?new JLabel():((row[j] instanceof Component)?row[j]:new JLabel(row[j].toString(),SwingConstants.RIGHT)));
        GridBagConstraints constraints=null;
        if (j==u-1) {
          constraints=new GridBagConstraints(j,i,u==1?2:1,1,1,weight,GridBagConstraints.LINE_START,GridBagConstraints.BOTH,new Insets(pSpacing,u==1?0:pSpacing,pSpacing,0),0,0);
        }
        else {
          constraints=new GridBagConstraints(j,i,1,1,0,weight,GridBagConstraints.LINE_END,(component instanceof JLabel)?GridBagConstraints.NONE:GridBagConstraints.BOTH,new Insets(pSpacing,0,pSpacing,pSpacing),0,0);
        }
        panel.add(component,constraints);
      }
    }
    return panel;
  }
  
  public static JPanel getBorderPane(Component...pComponents) {
    return getBorderPane(0,0,0,0,pComponents);
  }
  
  public static JPanel getBorderPane(int pTopInsets, int pLeftInsets, int pBottomInsets, int pRightInsets, Component...pComponents) {
    JPanel panel=new JPanel(new BorderLayout());
    panel.setBorder(BorderFactory.createEmptyBorder(pTopInsets,pLeftInsets,pBottomInsets,pRightInsets));
    if (pComponents[0]!=null) panel.add(pComponents[0],BorderLayout.NORTH);
    if (pComponents[1]!=null) panel.add(pComponents[1],BorderLayout.WEST);
    if (pComponents[2]!=null) panel.add(pComponents[2],BorderLayout.CENTER);
    if (pComponents[3]!=null) panel.add(pComponents[3],BorderLayout.EAST);
    if (pComponents[4]!=null) panel.add(pComponents[4],BorderLayout.SOUTH);
    return panel;
  }
  
  public static JScrollPane getScrollPane(Component pComponent, int pVerticalScrollPolicy, int pHorizontalScrollPolicy, int pPreferredWidth, int pPreferredHeight) {
    if (pComponent==null) {
      return null;
    }
    else {
      JScrollPane scrollPane=new JScrollPane(pComponent,pVerticalScrollPolicy,pHorizontalScrollPolicy);
      if (pPreferredWidth>0||pPreferredHeight>0) scrollPane.setPreferredSize(new Dimension(pPreferredWidth<=0?scrollPane.getPreferredSize().width:pPreferredWidth,pPreferredHeight<=0?scrollPane.getPreferredSize().height:pPreferredHeight));
      return scrollPane;
    }
  }
  
  public static JSplitPane getSplitPane(int pOrientation, Component pComponent1, Component pComponent2) {
    if (pComponent1==null||pComponent2==null) {
      return null;
    }
    else {
      JSplitPane splitPane=new JSplitPane(pOrientation,true,pComponent1,pComponent2);
      splitPane.setOneTouchExpandable(true);
      return splitPane;
    }
  }
  
}
