package gold.util;

import java.awt.geom.*;
import java.io.*;
import java.text.*;
import java.util.*;
import java.util.regex.*;
import gold.structures.automaton.*;
import gold.structures.disjointset.*;
import gold.structures.map.*;
import gold.structures.multimap.*;
import gold.structures.queue.*;
import gold.structures.set.*;
import gold.structures.tuple.*;
import gold.swing.util.*;
import gold.visualization.automaton.*;
import gold.visualization.util.*;

public final class GAutomata {
  
  public static enum Format {
    JFLAP
  }
  
  private GAutomata() {
  }
  
  public static IAutomaton<String> parse(String pAlphabet, String pInitialState, String pAcceptStates, String pTransitionFunction) {
    if (!pTransitionFunction.matches("(\\w+\\-.\\-\\w+(\\,\\w+\\-.\\-\\w+)*)?")) throw new IllegalArgumentException("The string that describes the transition function doesn't satisfy the format.");
    if (!pInitialState.matches("\\w+")) throw new IllegalArgumentException("The string that describes the initial state doesn't satisfy the format.");
    if (!pAcceptStates.matches("(\\w+(\\,\\w+)*)?")) throw new IllegalArgumentException("The string that describes the accept states doesn't satisfy the format.");
    ISet<String> states=new GHashTableSet<String>(pTransitionFunction.length()/4);
    states.union(GCollections.asSet(pTransitionFunction.split("\\-.\\-|\\,")));
    states.union(GCollections.asSet(pInitialState));
    states.union(GCollections.asSet(pAcceptStates.split("\\,")));
    states.remove("");
    GNondeterministicAutomaton<String> automaton=new GNondeterministicAutomaton<String>(states,new GAlphabet(pAlphabet));
    if (!pTransitionFunction.isEmpty()) {
      for (String string:pTransitionFunction.split("\\,")) {
        int i=string.indexOf('-'),j=string.lastIndexOf('-');
        String input=string.substring(0,i);
        char action=string.charAt(i+1);
        String output=string.substring(j+1);
        automaton.addDelta(input,action,output);
      }
    }
    automaton.setInitialState(pInitialState);
    if (!pAcceptStates.isEmpty()) {
      automaton.setAcceptStates(GCollections.asSet(pAcceptStates.split("\\,")));
    }
    return automaton;
  }
  
  public static IAutomaton<String> read(InputStream pInputStream) throws IOException {
    return read(pInputStream,(GLayoutConfiguration)null);
  }
  
  public static IAutomaton<String> read(InputStream pInputStream, GLayoutConfiguration pLayoutConfigurationResult) throws IOException {
    return read(new InputStreamReader(pInputStream,"UTF-8"),pLayoutConfigurationResult);
  }
  
  public static IAutomaton<String> read(Reader pReader) throws IOException {
    return read(pReader,(GLayoutConfiguration)null);
  }
  
  public static IAutomaton<String> read(Reader pReader, GLayoutConfiguration pLayoutConfigurationResult) throws IOException {
    final BufferedReader reader=new BufferedReader(pReader,1024);
    class ReaderConverter {
      private List<String> read(Iterator<String> pIterator, String pTitle, String pRegex) {
        String next=pIterator.hasNext()?pIterator.next():null;
        if (next==null) throw new IllegalArgumentException(pTitle+" must be specified.");
        if (!next.matches(pRegex)) throw new IllegalArgumentException(pTitle+" '"+next+"' doesn't satisfy the format '"+pRegex+"'.");
        return list(next.replaceAll("^[^\\ \\\"]+ *\\= *","").trim());
      }
      private List<String> list(String pText) {
        return Arrays.asList(pText.startsWith("{")&&pText.endsWith("}")?split(pText.substring(1,pText.length()-1)):new String[]{member(pText)});
      }
      private String lambda(String pText) {
        return pText.equals("λ")?"":pText;
      }
      private String member(String pText) {
        return GToolkit.convert(pText);
      }
      private List<Character> symbols(List<String> pList) {
        List<Character> list=new ArrayList<Character>(pList.size());
        for (String string:pList) {
          list.add(symbol(string));
        }
        return list;
      }
      private char symbol(String pText) {
        if (pText.startsWith("\"")&&pText.endsWith("\"")) pText=pText.substring(1,pText.length()-1);
        if (pText.length()!=1) throw new IllegalArgumentException("'"+pText+"' is not a single character.");
        return pText.charAt(0);
      }
      private int indexOf(String pText, String pFragment, int pStart) {
        for (int i=pStart,c=0; i<pText.length(); i++) {
          if (pText.charAt(i)=='\"') {
            c++;
          }
          else if ((c%2)==0&&pText.startsWith(pFragment,i)) {
            return i;
          }
        }
        return -1;
      }
      private String[] split(String pText) {
        if (pText.isEmpty()) return new String[0];
        List<String> list=new ArrayList<String>();
        int j=0,i=0,t=pText.length();
        for (int c=0; i<t; i++) {
          if (pText.charAt(i)=='\"') {
            c++;
          }
          else if ((c%2)==0&&pText.charAt(i)==',') {
            list.add(member(pText.substring(j,i)));
            j=i+1;
          }
        }
        list.add(member(pText.substring(j,i)));
        return list.toArray(new String[0]);
      }
    }
    final ReaderConverter z=new ReaderConverter();
    try {
      List<String> lines=new ArrayList<String>();
      for (String s; (s=reader.readLine())!=null;) {
        s=(s.startsWith("\uFEFF")?s.substring(1):s).trim();
        if (!s.isEmpty()&&!s.startsWith("//")) lines.add(s);
      }
      String regex="([^\\(\\)\\<\\>\\{\\}\\\"\\-\\,\\=\\t\\ ]*|\\\"([^\\\"]|\\\"\\\")*\\\")",regexL=regex+"(\\,"+regex+")*";
      Iterator<String> it=lines.iterator();
      String type=z.read(it,"The automaton type (ψ)","ψ *\\= *(DFA|DFT|NFA|PDA)").get(0);
      List<String> states=z.read(it,"The set of states (Q)","Q *\\= *\\{"+regexL+"\\}");
      List<Character> inputAlphabet=z.symbols(z.read(it,"The input alphabet (Σ)","Σ *\\= *\\{"+regexL+"\\}"));
      List<Character> outputAlphabet=type.equals("DFT")?z.symbols(z.read(it,"The output alphabet (Σ')","Σ' *\\= *\\{"+regexL+"\\}")):new ArrayList<Character>();
      List<Character> stackAlphabet=type.equals("PDA")?z.symbols(z.read(it,"The stack alphabet (Γ)","Γ *\\= *\\{"+regexL+"\\}")):new ArrayList<Character>();
      String initialState=z.read(it,"The initial state (q\u2080)","q\u2080 *\\= *"+regex).get(0);
      List<String> acceptStates=z.read(it,"The set of accept states (F)","F *\\= *\\{("+regexL+")?\\}");
      if (type.equals("DFA")||type.equals("DFT")||type.equals("NFA")||type.equals("PDA")) {
        boolean isDeterministic=type.equals("DFA")||type.equals("DFT");
        boolean isTransducer=type.equals("DFT");
        boolean isPushdownAutomaton=type.equals("PDA");
        IAutomaton<String> automaton=null;
        if (type.equals("DFA")) {
          automaton=new GDeterministicAutomaton<String>(states,inputAlphabet,initialState,acceptStates);
        }
        else if (type.equals("DFT")) {
          automaton=new GDeterministicTransducer<String>(states,inputAlphabet,outputAlphabet,initialState,acceptStates);
        }
        else if (type.equals("PDA")) {
          automaton=new GPushdownAutomaton<String>(states,inputAlphabet,stackAlphabet,initialState,acceptStates);
        }
        else {
          automaton=new GNondeterministicAutomaton<String>(states,inputAlphabet,initialState,acceptStates);
        }
        Map<Character,Pattern> patterns=new HashMap<Character,Pattern>(8);
        if (isPushdownAutomaton) {
          String regexB=regex+"\\-\\>"+regex+" +with +"+regex+"\\,";
          String regex0=regexB+"ignore";
          String regex1=regexB+"(skip|push|pop)\\("+regex+"\\)";
          String regex2=regexB+"(pushOn|changeTop)\\("+regex+"\\,"+regex+"\\)";
          patterns.put('-',Pattern.compile("("+regex0+")|("+regex1+")|("+regex2+")"));
        }
        else {
          patterns.put('δ',Pattern.compile("δ\\("+regex+"\\,"+regex+"\\) *\\= *"+regex));
          patterns.put('Δ',Pattern.compile("Δ\\("+regex+"\\,"+regex+"\\) *\\= *\\{("+regexL+")?\\}"));
          patterns.put('g',Pattern.compile("g\\("+regex+"\\) *\\= *"+regex));
          patterns.put('h',Pattern.compile("h\\("+regex+"\\,"+regex+"\\) *\\= *"+regex));
        }
        List<String> layout=new ArrayList<String>(10);
        while (it.hasNext()) {
          String command=it.next();
          if (command.startsWith("ξ(")) {
            layout.add(command);
          }
          else if (isPushdownAutomaton) {
            Pattern pattern=patterns.get('-');
            IPushdownAutomaton<String> pushdownAutomaton=(IPushdownAutomaton<String>)automaton;
            if (!pattern.matcher(command).matches()) throw new IllegalArgumentException("The command '"+command+"' doesn't satisfy the format '"+pattern.pattern()+"'.");
            int p1=z.indexOf(command,"->",0),p2=p1+"->".length(),p3=z.indexOf(command," with ",p2),p4=p3+" with ".length(),p5=z.indexOf(command,",",p4);
            String input=z.member(command.substring(0,p1).trim());
            String output=z.member(command.substring(p2,p3).trim());
            String prefix=z.member(z.lambda(command.substring(p4,p5).trim()));
            IPushdownDelta<String> delta=pushdownAutomaton.delta(input,output,prefix);
            String operation=command.substring(p5+1).trim();
            if (operation.equals("ignore")) {
              delta.ignore();
            }
            else if (operation.startsWith("skip(")||operation.startsWith("push(")||operation.startsWith("pop(")) {
              int p6=z.indexOf(operation,"(",0),p7=z.indexOf(operation,")",p6+1);
              String arg=z.member(z.lambda(operation.substring(p6+1,p7).trim()));
              if (operation.startsWith("skip(")) {
                delta.skip(arg);
              }
              else if (operation.startsWith("push(")) {
                delta.push(arg);
              }
              else if (operation.startsWith("pop(")) {
                delta.pop(arg);
              }
            }
            else if (operation.startsWith("pushOn(")||operation.startsWith("changeTop(")) {
              int p6=z.indexOf(operation,"(",0),p7=z.indexOf(operation,",",p6+1),p8=z.indexOf(operation,")",p7+1);
              String arg1=z.member(z.lambda(operation.substring(p6+1,p7).trim())),arg2=z.member(z.lambda(operation.substring(p7+1,p8).trim()));
              if (operation.startsWith("pushOn(")) {
                delta.pushOn(arg1,arg2);
              }
              else if (operation.startsWith("changeTop(")) {
                delta.changeTop(arg1,arg2);
              }
            }
          }
          else {
            Pattern pattern=null;
            if (command.startsWith("δ")&&isDeterministic) pattern=patterns.get(command.charAt(0));
            if (command.startsWith("Δ")&&!isDeterministic) pattern=patterns.get(command.charAt(0));
            if (command.startsWith("h")&&isTransducer) pattern=patterns.get(command.charAt(0));
            if (command.startsWith("g")&&isTransducer) pattern=patterns.get(command.charAt(0));
            if (pattern==null) throw new IllegalArgumentException("The first letter of the command '"+command+"' must be "+(isTransducer?"δ, g or h":(isDeterministic?"δ":"Δ"))+".");
            if (!pattern.matcher(command).matches()) throw new IllegalArgumentException("The command '"+command+"' doesn't satisfy the format '"+pattern.pattern()+"'.");
            if (command.startsWith("g")) {
              int p1=z.indexOf(command,"(",0),p2=z.indexOf(command,")",p1+1),p3=z.indexOf(command,"=",p2+1);
              String input=z.member(command.substring(p1+1,p2).trim());
              String string=z.member(command.substring(p3+1).trim());
              ((GDeterministicTransducer<String>)automaton).setStateOutput(input,string);
            }
            else {
              int p1=z.indexOf(command,"(",0),p2=z.indexOf(command,",",p1+1),p3=z.indexOf(command,")",p2+1),p4=z.indexOf(command,"=",p3+1);
              String input=z.member(command.substring(p1+1,p2).trim());
              char action=z.symbol(z.member(command.substring(p2+1,p3).trim()));
              if (command.startsWith("h")) {
                String string=z.member(command.substring(p4+1).trim());
                ((GDeterministicTransducer<String>)automaton).setTransitionOutput(input,action,string);
              }
              else {
                List<String> outputs=z.list(command.substring(p4+1).trim());
                ((GNondeterministicAutomaton<String>)automaton).setDeltaSet(input,action,outputs);
              }
            }
          }
        }
        if (pLayoutConfigurationResult!=null&&!layout.isEmpty()) {
          String regexN="(\\-?[0-9]+(\\.[0-9]+)?)";
          String regexP="("+regexN+"\\,"+regexN+")";
          String regex1="ξ\\("+regex+"\\)\\=\\("+regexP+"\\)";
          String regex2="ξ\\("+regex+"\\,"+regex+"\\)\\=\\("+regexP+"(\\,"+regexP+")*\\)";
          Pattern pattern=Pattern.compile("("+regex1+")|("+regex2+")");
          for (String command:layout) {
            if (!pattern.matcher(command).matches()) throw new IllegalArgumentException("The command '"+command+"' doesn't satisfy the format '"+pattern.pattern()+"'.");
            int p1=z.indexOf(command,"(",0),pC=z.indexOf(command,",",p1+1),p2=z.indexOf(command,")",p1+1);
            int p3=z.indexOf(command,"(",p2+1),p4=z.indexOf(command,")",p3+1);
            String fragments[]=command.substring(p3+1,p4).trim().split("\\,");
            List<Point2D> points=new ArrayList<Point2D>(4);
            for (int i=0; i+1<fragments.length; i+=2) {
              points.add(new Point2D.Double(Double.parseDouble(fragments[i]),Double.parseDouble(fragments[i+1])));
            }
            if (pC==-1||pC>p2) {
              String state=command.substring(p1+1,p2).trim();
              if (automaton.getStates().contains(state)) {
                pLayoutConfigurationResult.getVertexLayout().put(state,points.get(0));
              }
            }
            else {
              String state1=command.substring(p1+1,pC).trim();
              String state2=command.substring(pC+1,p2).trim();
              if (automaton.getStates().contains(state1)&&automaton.getStates().contains(state2)) {
                pLayoutConfigurationResult.getEdgeLayout().put(new GPair<String>(state1,state2),points.toArray(new Point2D[0]));
              }
            }
          }
        }
        return automaton;
      }
      else {
        throw new IllegalArgumentException("Invalid type of automata: "+type);
      }
    }
    finally {
      reader.close();
    }
  }
  
  public static <V> void write(OutputStream pOutputStream, IAutomaton<V> pAutomaton) throws IOException {
    write(pOutputStream,pAutomaton,null);
  }
  
  public static <V> void write(OutputStream pOutputStream, GAutomatonPainter<V> pPainter) throws IOException {
    write(pOutputStream,pPainter.getAutomaton(),pPainter.getLayoutConfiguration());
  }
  
  public static <V> void write(OutputStream pOutputStream, IAutomaton<V> pAutomaton, GLayoutConfiguration pLayoutConfiguration) throws IOException {
    write(new OutputStreamWriter(pOutputStream,"UTF-8"),pAutomaton,pLayoutConfiguration);
  }
  
  public static <V> void write(Writer pWriter, IAutomaton<V> pAutomaton) throws IOException {
    write(pWriter,pAutomaton,null);
  }
  
  public static <V> void write(Writer pWriter, GAutomatonPainter<V> pPainter) throws IOException {
    write(pWriter,pPainter.getAutomaton(),pPainter.getLayoutConfiguration());
  }
  
  public static <V> void write(Writer pWriter, IAutomaton<V> pAutomaton, GLayoutConfiguration pLayoutConfiguration) throws IOException {
    Set<String> strings=new HashSet<String>(pAutomaton.getStates().size());
    for (V state:pAutomaton.getStates()) {
      if (!strings.add(state.toString())) throw new IllegalArgumentException("The automaton has states with the same textual representation.");
    }
    final BufferedWriter writer=new BufferedWriter(pWriter,1024);
    class WriterEscaper {
      private final IMap<Object,String> cache=new GHashTableMap<Object,String>(16);
      private String list(Iterable<?> pObjects) {
        Object[] array=GCollections.unmodifiableCollection(pObjects).toArray();
        for (int i=0; i<array.length; i++) {
          array[i]=member(array[i]);
        }
        return "{"+GToolkit.separate(",",array)+"}";
      }
      private String set(Iterable<?> pObjects) {
        return list(sort(pObjects));
      }
      private String member(Object pObject) {
        if (pObject==null) return "";
        String string=cache.get(pObject);
        if (string==null) cache.put(pObject,string=GToolkit.escape(pObject.toString()));
        return string;
      }
      private <T> Iterable<T> sort(Iterable<T> pObjects) {
        List<T> list=new ArrayList<T>(GCollections.asJavaCollection(pObjects));
        Collections.sort(list,new Comparator<T>() {
          private final Comparator<String> comparator=GComparators.shortlexStringComparator();
          @Override
          public int compare(T pObject1, T pObject2) {
            return comparator.compare(member(pObject1),member(pObject2));
          }
        });
        return list;
      }
    }
    final WriterEscaper z=new WriterEscaper();
    try {
      boolean isDeterministic=pAutomaton.isDeterministic();
      writer.write("// Automaton type (ψ):\r\n");
      writer.write("ψ = "+type(pAutomaton)+"\r\n");
      writer.write("// Set of states (Q):\r\n");
      writer.write("Q = "+z.set(pAutomaton.getStates())+"\r\n");
      writer.write("// Input alphabet (Σ):\r\n");
      writer.write("Σ = "+z.list(pAutomaton.getAlphabet())+"\r\n");
      if (pAutomaton instanceof ITransducer) {
        writer.write("// Output alphabet (Σ'):\r\n");
        writer.write("Σ' = "+z.list(((ITransducer<V>)pAutomaton).getOutputAlphabet())+"\r\n");
      }
      else if (pAutomaton instanceof IPushdownAutomaton) {
        writer.write("// Stack alphabet (Γ):\r\n");
        writer.write("Γ = "+z.list(((IPushdownAutomaton<V>)pAutomaton).getStackAlphabet())+"\r\n");
      }
      writer.write("// Initial state (q\u2080):\r\n");
      writer.write("q\u2080 = "+z.member(pAutomaton.getInitialState())+"\r\n");
      writer.write("// Set of accept states (F):\r\n");
      writer.write("F = "+z.set(pAutomaton.getAcceptStates())+"\r\n");
      writer.write("// Transition function ("+(isDeterministic?"δ":"Δ")+"):\r\n");
      if (pAutomaton instanceof IPushdownAutomaton) {
        IPushdownAutomaton<V> pushdownAutomaton=(IPushdownAutomaton<V>)pAutomaton;
        for (V input:z.sort(pAutomaton.getStates())) {
          List<IPushdownTransition<V>> transitions=new ArrayList<IPushdownTransition<V>>(GCollections.asJavaCollection(pushdownAutomaton.getTransitions(input)));
          Collections.sort(transitions,new Comparator<IPushdownTransition<V>>() {
            private final Comparator<String> comparator=GComparators.shortlexStringComparator();
            @Override
            public int compare(IPushdownTransition<V> pTransition1, IPushdownTransition<V> pTransition2) {
              int c=comparator.compare(z.member(pTransition1.getOutput()),z.member(pTransition2.getOutput()));
              return c!=0?c:comparator.compare(pTransition1.toString(),pTransition2.toString());
            }
          });
          for (IPushdownTransition<V> transition:transitions) {
            writer.write(z.member(transition.getInput())+"->"+z.member(transition.getOutput())+" with "+transition.toString()+"\r\n");
          }
        }
      }
      else {
        for (V input:z.sort(pAutomaton.getStates())) {
          for (char action:pAutomaton.getActions()) {
            ISet<V> outputs=pAutomaton.getDeltaSet(input,action);
            if (!outputs.isEmpty()) {
              writer.write((isDeterministic?"δ":"Δ")+"("+z.member(input)+","+z.member(action)+") = ");
              writer.write(isDeterministic?z.member(outputs.pick()):z.set(outputs));
              writer.write("\r\n");
            }
          }
        }
        if (pAutomaton instanceof ITransducer) {
          ITransducer<V> transducer=(ITransducer<V>)pAutomaton;
          writer.write("// State output function (g):\r\n");
          for (V input:z.sort(transducer.getStates())) {
            writer.write("g("+z.member(input)+") = "+z.member(transducer.getStateOutput(input))+"\r\n");
          }
          writer.write("// Transition output function (h):\r\n");
          for (V input:z.sort(transducer.getStates())) {
            for (char action:transducer.getActions()) {
              writer.write("h("+z.member(input)+","+z.member(action)+") = "+z.member(transducer.getTransitionOutput(input,action))+"\r\n");
            }
          }
        }
      }
      if (pLayoutConfiguration!=null&&!pLayoutConfiguration.isEmpty()) {
        writer.write("// Automaton layout (ξ)\r\n");
        DecimalFormat decimalFormat=GUtilities.getDecimalFormat("0.00");
        for (V state:z.sort(pAutomaton.getStates())) {
          Point2D location=pLayoutConfiguration.getVertexLayout().get(state.toString());
          if (location!=null) {
            writer.write("ξ("+z.member(state)+")=("+decimalFormat.format(location.getX()).replace(',','.')+","+decimalFormat.format(location.getY()).replace(',','.')+")\r\n");
          }
        }
        for (Map.Entry<IPair<String>,Point2D[]> entry:pLayoutConfiguration.getEdgeLayout().toJavaView().entrySet()) {
          StringBuilder sb=new StringBuilder();
          for (Point2D point:entry.getValue()) {
            if (sb.length()>0) sb.append(",");
            sb.append(decimalFormat.format(point.getX()));
            sb.append(",");
            sb.append(decimalFormat.format(point.getY()));
          }
          String state1=entry.getKey().getFirst(),state2=entry.getKey().getSecond();
          writer.write("ξ("+z.member(state1)+","+z.member(state2)+")=("+sb+")\r\n");
        }
      }
    }
    finally {
      writer.close();
    }
  }
  
  public static <V> void write(OutputStream pOutputStream, Format pFormat, IAutomaton<V> pAutomaton) throws IOException {
    write(pOutputStream,pFormat,pAutomaton,null);
  }
  
  public static <V> void write(OutputStream pOutputStream, Format pFormat, GAutomatonPainter<V> pPainter) throws IOException {
    write(pOutputStream,pFormat,pPainter.getAutomaton(),pPainter.getLayoutConfiguration());
  }
  
  public static <V> void write(OutputStream pOutputStream, Format pFormat, IAutomaton<V> pAutomaton, GLayoutConfiguration pLayoutConfiguration) throws IOException {
    Set<String> strings=new HashSet<String>(pAutomaton.getStates().size());
    for (V state:pAutomaton.getStates()) {
      if (!strings.add(state.toString())) throw new IllegalArgumentException("The automaton has states with the same textual representation.");
    }
    Random random=new Random(System.currentTimeMillis());
    switch (pFormat) {
      case JFLAP:
        IAutomaton<V> automaton=pAutomaton;
        IPushdownAutomaton<V> pushdownAutomaton=(automaton instanceof IPushdownAutomaton)?(IPushdownAutomaton<V>)automaton:null;
        ITransducer<V> transducer=(automaton instanceof ITransducer)?(ITransducer<V>)automaton:null;
        Writer writer=new OutputStreamWriter(pOutputStream,"UTF-8");
        try {
          String endl="&#13;\r\n";
          IMap<V,Integer> identifiers=new GHashTableMap<V,Integer>(automaton.getStates().size());
          for (V state:automaton.getStates()) {
            identifiers.put(state,identifiers.size());
          }
          boolean isMoore=false,isMealy=false;
          if (transducer!=null) {
            isMoore=isMealy=true;
            for (V state:automaton.getStates()) {
              if (isMealy&&!transducer.getStateOutput(state).isEmpty()) isMealy=false;
              for (char symbol:automaton.getAlphabet()) {
                if (isMoore&&!transducer.getTransitionOutput(state,symbol).isEmpty()) isMoore=false;
              }
            }
            if (!isMoore&&!isMealy) {
              throw new IllegalArgumentException("The transducer cannot be exported to JFlap because it's not a Moore nor a Mealy Machine (it outputs in both states and transitions).");
            }
            if (isMoore&&isMealy) {
              transducer=null;
              isMoore=isMealy=false;
            }
          }
          if (pushdownAutomaton!=null) {
            if (pushdownAutomaton.getStackAlphabet().contains('Z')) throw new IllegalArgumentException("The pushdown automaton cannot be exported to JFlap because the character 'Z' belongs to the automaton's stack alphabet.");
            if (strings.contains("")) throw new IllegalArgumentException("The pushdown automaton cannot be exported to JFlap because it contains a state with an empty textual representation.");
          }
          writer.write("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?><!--Created with GOLD "+GAboutDialog.getMinorVersion()+". for JFLAP 6.4.--><structure>"+endl);
          writer.write("  <type>"+(pushdownAutomaton!=null?"pda":(isMoore?"moore":(isMealy?"mealy":"fa")))+"</type>"+endl);
          writer.write("  <automaton>"+endl);
          writer.write("    <!--The list of states.-->"+endl);
          DecimalFormat decimalFormat=GUtilities.getDecimalFormat("0.00");
          for (V state:automaton.getStates()) {
            Point2D location=pLayoutConfiguration==null?null:pLayoutConfiguration.getVertexLayout().get(state.toString());
            if (location==null) location=new Point2D.Double(random.nextInt(300),random.nextInt(300));
            String output=transducer!=null&&isMoore?transducer.getStateOutput(state):"";
            writer.write("    <state id=\""+identifiers.get(state)+"\" name=\""+GToolkit.escapeXML(state.toString())+"\">"+endl);
            writer.write("      <x>"+decimalFormat.format(location.getX()).replace(',','.')+"</x>"+endl);
            writer.write("      <y>"+decimalFormat.format(location.getY()).replace(',','.')+"</y>"+endl);
            if (GToolkit.equals(state,automaton.getInitialState())) writer.write("      <initial/>"+endl);
            if (pushdownAutomaton==null&&automaton.isAcceptState(state)) writer.write("      <final/>"+endl);
            if (isMoore) writer.write("      "+(output.isEmpty()?"<output/>":"<output>"+GToolkit.escapeXML(output)+"</output>")+endl);
            writer.write("    </state>"+endl);
          }
          if (pushdownAutomaton!=null) {
            writer.write("    <state id=\""+identifiers.size()+"\" name=\"\">"+endl);
            writer.write("      <x>0.00</x>"+endl);
            writer.write("      <y>0.00</y>"+endl);
            writer.write("      <final/>"+endl);
            writer.write("    </state>"+endl);
          }
          writer.write("    <!--The list of transitions.-->"+endl);
          for (V source:automaton.getStates()) {
            if (pushdownAutomaton!=null) {
              for (IPushdownTransition<V> transition:pushdownAutomaton.getTransitions(source)) {
                writer.write("    <transition>"+endl);
                writer.write("      <from>"+identifiers.get(transition.getInput())+"</from>"+endl);
                writer.write("      <to>"+identifiers.get(transition.getOutput())+"</to>"+endl);
                String prefix=transition.getPrefix(),phi=GToolkit.reverse(transition.getPhi()),rho=GToolkit.reverse(transition.getRho());
                writer.write("      "+(prefix.isEmpty()?"<read/>":"<read>"+GToolkit.escapeXML(prefix)+"</read>")+endl);
                writer.write("      "+(phi.isEmpty()?"<pop/>":"<pop>"+GToolkit.escapeXML(phi)+"</pop>")+endl);
                writer.write("      "+(rho.isEmpty()?"<push/>":"<push>"+GToolkit.escapeXML(rho)+"</push>")+endl);
                writer.write("    </transition>"+endl);
              }
            }
            else {
              for (char action:automaton.getActions()) {
                for (V destination:automaton.getDeltaSet(source,action)) {
                  String output=transducer!=null&&isMoore?transducer.getStateOutput(destination):(transducer!=null&&isMealy?transducer.getTransitionOutput(source,action):"");
                  writer.write("    <transition>"+endl);
                  writer.write("      <from>"+identifiers.get(source)+"</from>"+endl);
                  writer.write("      <to>"+identifiers.get(destination)+"</to>"+endl);
                  writer.write("      "+(action==IAutomaton.LAMBDA?"<read/>":"<read>"+GToolkit.escapeXML(String.valueOf(action))+"</read>")+endl);
                  if (isMoore||isMealy) writer.write("      "+(output.isEmpty()?"<transout/>":"<transout>"+GToolkit.escapeXML(output)+"</transout>")+endl);
                  writer.write("    </transition>"+endl);
                }
              }
            }
          }
          if (pushdownAutomaton!=null) {
            for (V state:pushdownAutomaton.getAcceptStates()) {
              writer.write("    <transition>"+endl);
              writer.write("      <from>"+identifiers.get(state)+"</from>"+endl);
              writer.write("      <to>"+identifiers.size()+"</to>"+endl);
              writer.write("      <read/>"+endl);
              writer.write("      <pop>Z</pop>"+endl);
              writer.write("      <push/>"+endl);
              writer.write("    </transition>"+endl);
            }
          }
          writer.write("  </automaton>"+endl);
          writer.write("</structure>");
        }
        finally {
          writer.close();
        }
        break;
    }
  }
  public static <V> String type(IAutomaton<V> pAutomaton) {
    if (pAutomaton instanceof ITransducer) return "DFT"; // Deterministic finite transducer
    if (pAutomaton instanceof IPushdownAutomaton) return "PDA"; // Pushdown finite automaton
    if (!pAutomaton.isDeterministic()) return "NFA"; // Nondeterministic finite automaton
    return "DFA"; // Deterministic finite automaton
  }
  
  public static <V> IAutomaton<?> simplify(IAutomaton<V> pAutomaton) {
    checkPrecondition(pAutomaton);
    ISet<V> states=pAutomaton.getStates();
    V initialState=pAutomaton.getInitialState();
    ISet<V> usefulStates=pAutomaton.getUsefulStates();
    if (!usefulStates.contains(initialState)) {
      GDeterministicAutomaton<V> result=new GDeterministicAutomaton<V>(GCollections.singletonSet(initialState),pAutomaton.getAlphabet());
      result.setInitialState(initialState);
      for (char symbol:result.getAlphabet()) {
        result.setDelta(initialState,symbol,initialState);
      }
      return result;
    }
    else if (states.size()==usefulStates.size()) {
      return new GNondeterministicAutomaton<V>(pAutomaton);
    }
    else {
      GNondeterministicAutomaton<V> result=new GNondeterministicAutomaton<V>(usefulStates,pAutomaton.getAlphabet());
      result.setInitialState(initialState);
      result.setAcceptStates(GCollections.intersection(pAutomaton.getAcceptStates(),usefulStates));
      for (V input:usefulStates) {
        for (char action:pAutomaton.getActions()) {
          result.setDeltaSet(input,action,GCollections.intersection(pAutomaton.getDeltaSet(input,action),usefulStates));
        }
      }
      return result;
    }
  }
  
  public static <V> IAutomaton<?> determinize(IAutomaton<V> pAutomaton) {
    checkPrecondition(pAutomaton);
    ISet<V> states=pAutomaton.getStates();
    if (pAutomaton.isDeterministic()) {
      return new GDeterministicAutomaton<V>(pAutomaton);
    }
    else if (states.size()>16) {
      throw new UnsupportedOperationException("The result of the determinization exceeds the maximum number of states ("+(1<<16)+").");
    }
    else {
      ISet<ISet<V>> rStates=new GHashTableSet<ISet<V>>(1<<states.size());
      for (ISet<V> subset:GCollections.powerSet(states)) {
        rStates.add(pAutomaton.applyLambdasRecursively(subset));
      }
      GDeterministicAutomaton<ISet<V>> result=new GDeterministicAutomaton<ISet<V>>(rStates,pAutomaton.getAlphabet());
      result.setInitialState(pAutomaton.applyLambdasRecursively(GCollections.singletonSet(pAutomaton.getInitialState())));
      for (ISet<V> state:result.getStates()) {
        if (!state.isDisjointFrom(pAutomaton.getAcceptStates())) result.addAcceptState(state);
        for (char symbol:result.getAlphabet()) {
          result.setDelta(state,symbol,pAutomaton.applyAction(state,symbol));
        }
      }
      return result;
    }
  }
  
  private static <V> IAutomaton<?> minimizeAuxiliaryMethod(IAutomaton<V> pAutomaton) {
    checkPrecondition(pAutomaton);
    assert pAutomaton.isDeterministic();
    IAlphabet alphabet=pAutomaton.getAlphabet();
    ISet<V> states=pAutomaton.getStates(),acceptStates=pAutomaton.getAcceptStates();
    IDisjointSets<V> forest=new GForestDisjointSets<V>(states);
    ISet<V> acceptSet=acceptStates,rejectSet=GCollections.difference(states,acceptStates);
    for (V state:states) {
      forest.union(state,(pAutomaton.isAcceptState(state)?acceptSet:rejectSet).pick());
    }
    boolean modified=true;
    while (modified) {
      modified=false;
      IDisjointSets<V> nextForest=new GForestDisjointSets<V>(states);
      for (V state1:states) {
        IDisjointSet<V> set1=forest.findSet(state1),nextSet1=nextForest.findSet(state1);
        for (V state2:states) {
          IDisjointSet<V> set2=forest.findSet(state2),nextSet2=nextForest.findSet(state2);
          if (set1==set2&&nextSet1!=nextSet2) {
            boolean equivalent=true;
            for (char symbol:alphabet) {
              if (forest.findSet(pAutomaton.getDelta(state1,symbol))!=forest.findSet(pAutomaton.getDelta(state2,symbol))) {
                equivalent=false;
                modified=true;
                break;
              }
            }
            if (equivalent) nextForest.union(state1,state2);
          }
        }
      }
      forest=nextForest;
    }
    IMultiMap<V,V> partition=new GHashTableMultiMap<V,V>(states.size());
    IMap<V,V> representants=new GHashTableMap<V,V>(states.size());
    for (V state:states) {
      V representant=forest.findSet(state).getData();
      partition.add(representant,state);
      representants.put(state,representant);
    }
    GDeterministicAutomaton<ISet<V>> result=new GDeterministicAutomaton<ISet<V>>(partition.values(),alphabet);
    result.setInitialState(partition.get(representants.get(pAutomaton.getInitialState())));
    for (ISet<V> state:result.getStates()) {
      if (!state.isDisjointFrom(acceptStates)) result.addAcceptState(state);
      for (char symbol:result.getAlphabet()) {
        V output=pAutomaton.getDelta(state.pick(),symbol);
        result.setDelta(state,symbol,partition.get(representants.get(output)));
      }
    }
    return result;
  }
  
  public static <V> IAutomaton<?> minimize(IAutomaton<V> pAutomaton) {
    checkPrecondition(pAutomaton);
    return minimizeAuxiliaryMethod(determinize(pAutomaton));
  }
  
  public static <V,W> IAutomaton<?> union(IAutomaton<V> pAutomaton1, IAutomaton<W> pAutomaton2) {
    checkPrecondition(pAutomaton1,pAutomaton2);
    if (!pAutomaton1.getAlphabet().equals(pAutomaton2.getAlphabet())) throw new IllegalArgumentException("The alphabets of both automata aren't equal.");
    if (!pAutomaton1.isDeterministic()||!pAutomaton2.isDeterministic()) {
      return GAutomata.union(pAutomaton1.minimize(),pAutomaton2.minimize());
    }
    else {
      GNondeterministicAutomaton<ICouple<V,W>> result=new GDeterministicAutomaton<ICouple<V,W>>(GCollections.binaryCartesianProduct(pAutomaton1.getStates(),pAutomaton2.getStates()),pAutomaton1.getAlphabet());
      result.setInitialState(new GCouple<V,W>(pAutomaton1.getInitialState(),pAutomaton2.getInitialState()));
      ISet<ICouple<V,W>> acceptStates1=GCollections.binaryCartesianProduct(pAutomaton1.getAcceptStates(),pAutomaton2.getStates());
      ISet<ICouple<V,W>> acceptStates2=GCollections.binaryCartesianProduct(pAutomaton1.getStates(),pAutomaton2.getAcceptStates());
      result.setAcceptStates(GCollections.union(acceptStates1,acceptStates2));
      for (ICouple<V,W> input:result.getStates()) {
        for (char action:result.getAlphabet()) {
          result.setDelta(input,action,new GCouple<V,W>(pAutomaton1.getDelta(input.getFirst(),action),pAutomaton2.getDelta(input.getSecond(),action)));
        }
      }
      return result;
    }
  }
  
  public static <V,W> IAutomaton<?> intersection(IAutomaton<V> pAutomaton1, IAutomaton<W> pAutomaton2) {
    checkPrecondition(pAutomaton1,pAutomaton2);
    if (!pAutomaton1.getAlphabet().equals(pAutomaton2.getAlphabet())) throw new IllegalArgumentException("The alphabets of both automata aren't equal.");
    GNondeterministicAutomaton<ICouple<V,W>> result=new GNondeterministicAutomaton<ICouple<V,W>>(GCollections.binaryCartesianProduct(pAutomaton1.getStates(),pAutomaton2.getStates()),pAutomaton1.getAlphabet());
    result.setInitialState(new GCouple<V,W>(pAutomaton1.getInitialState(),pAutomaton2.getInitialState()));
    result.setAcceptStates(GCollections.binaryCartesianProduct(pAutomaton1.getAcceptStates(),pAutomaton2.getAcceptStates()));
    for (ICouple<V,W> input:result.getStates()) {
      for (char action:result.getActions()) {
        result.setDeltaSet(input,action,GCollections.binaryCartesianProduct(pAutomaton1.getDeltaSet(input.getFirst(),action),pAutomaton2.getDeltaSet(input.getSecond(),action)));
      }
    }
    return result;
  }
  
  public static <V,W> IAutomaton<?> concatenation(IAutomaton<V> pAutomaton1, IAutomaton<W> pAutomaton2) {
    checkPrecondition(pAutomaton1,pAutomaton2);
    if (!pAutomaton1.getAlphabet().equals(pAutomaton2.getAlphabet())) throw new IllegalArgumentException("The alphabets of both automata aren't equal.");
    ISet<ICouple<V,W>> states1=GCollections.binaryCartesianProduct(pAutomaton1.getStates(),GCollections.<W>singletonSet(null));
    ISet<ICouple<V,W>> states2=GCollections.binaryCartesianProduct(GCollections.<V>singletonSet(null),pAutomaton2.getStates());
    GNondeterministicAutomaton<ICouple<V,W>> result=new GNondeterministicAutomaton<ICouple<V,W>>(GCollections.union(states1,states2),pAutomaton1.getAlphabet());
    result.setInitialState(new GCouple<V,W>(pAutomaton1.getInitialState(),null));
    result.setAcceptStates(GCollections.binaryCartesianProduct(GCollections.<V>singletonSet(null),pAutomaton2.getAcceptStates()));
    for (V input1:pAutomaton1.getStates()) {
      for (char action:result.getActions()) {
        for (V output1:pAutomaton1.getDeltaSet(input1,action)) {
          result.addDelta(new GCouple<V,W>(input1,null),action,new GCouple<V,W>(output1,null));
        }
      }
    }
    for (W input2:pAutomaton2.getStates()) {
      for (char action:result.getActions()) {
        for (W output2:pAutomaton2.getDeltaSet(input2,action)) {
          result.addDelta(new GCouple<V,W>(null,input2),action,new GCouple<V,W>(null,output2));
        }
      }
    }
    for (V acceptState1:pAutomaton1.getAcceptStates()) {
      result.addDelta(new GCouple<V,W>(acceptState1,null),IAutomaton.LAMBDA,new GCouple<V,W>(null,pAutomaton2.getInitialState()));
    }
    return result;
  }
  
  public static <V,W> IAutomaton<?> difference(IAutomaton<V> pAutomaton1, IAutomaton<W> pAutomaton2) {
    checkPrecondition(pAutomaton1,pAutomaton2);
    return intersection(pAutomaton1,complement(pAutomaton2));
  }
  
  public static <V> IAutomaton<?> transpose(IAutomaton<V> pAutomaton) {
    checkPrecondition(pAutomaton);
    ISet<V> specialState=GCollections.<V>emptySet();
    GNondeterministicAutomaton<ISet<V>> result=new GNondeterministicAutomaton<ISet<V>>(GCollections.union(GCollections.singletons(pAutomaton.getStates()),GCollections.singletonSet(specialState)),pAutomaton.getAlphabet());
    result.setInitialState(specialState);
    result.addAcceptState(GCollections.singletonSet(pAutomaton.getInitialState()));
    for (V input:pAutomaton.getStates()) {
      for (char action:result.getActions()) {
        for (V output:pAutomaton.getDeltaSet(input,action)) {
          result.addDelta(GCollections.singletonSet(output),action,GCollections.singletonSet(input));
        }
      }
    }
    for (V acceptState:pAutomaton.getAcceptStates()) {
      result.addDelta(specialState,IAutomaton.LAMBDA,GCollections.singletonSet(acceptState));
    }
    return result;
  }
  
  private static <V> IAutomaton<V> complementAuxiliaryMethod(IAutomaton<V> pAutomaton) {
    checkPrecondition(pAutomaton);
    assert pAutomaton.isDeterministic();
    GDeterministicAutomaton<V> automaton=new GDeterministicAutomaton<V>(pAutomaton);
    automaton.setAcceptStates(GCollections.difference(automaton.getStates(),automaton.getAcceptStates()));
    return automaton;
  }
  
  public static <V> IAutomaton<?> complement(IAutomaton<V> pAutomaton) {
    checkPrecondition(pAutomaton);
    return complementAuxiliaryMethod(pAutomaton.determinize());
  }
  
  public static <V> IAutomaton<?> optionality(IAutomaton<V> pAutomaton) {
    checkPrecondition(pAutomaton);
    ISet<V> specialState=GCollections.<V>emptySet();
    GNondeterministicAutomaton<ISet<V>> result=new GNondeterministicAutomaton<ISet<V>>(GCollections.union(GCollections.singletons(pAutomaton.getStates()),GCollections.singletonSet(specialState)),pAutomaton.getAlphabet());
    result.setInitialState(specialState);
    result.setAcceptStates(GCollections.singletons(pAutomaton.getAcceptStates()));
    for (V input:pAutomaton.getStates()) {
      for (char action:result.getActions()) {
        result.setDeltaSet(GCollections.singletonSet(input),action,GCollections.singletons(pAutomaton.getDeltaSet(input,action)));
      }
    }
    result.addDelta(specialState,IAutomaton.LAMBDA,GCollections.singletonSet(pAutomaton.getInitialState()));
    result.addAcceptState(specialState);
    return result;
  }
  
  public static <V> IAutomaton<?> kleenePlus(IAutomaton<V> pAutomaton) {
    checkPrecondition(pAutomaton);
    GNondeterministicAutomaton<V> result=new GNondeterministicAutomaton<V>(pAutomaton);
    for (V acceptState:result.getAcceptStates()) {
      result.addDelta(acceptState,IAutomaton.LAMBDA,result.getInitialState());
    }
    return result;
  }
  
  public static <V> IAutomaton<?> kleeneStar(IAutomaton<V> pAutomaton) {
    checkPrecondition(pAutomaton);
    return optionality(kleenePlus(pAutomaton));
  }
  
  private static <V,W> boolean areEquivalentAuxiliaryMethod(IAutomaton<V> pAutomaton1, IAutomaton<W> pAutomaton2) {
    checkPrecondition(pAutomaton1,pAutomaton2);
    assert pAutomaton1.isDeterministic()&&pAutomaton2.isDeterministic();
    if (!pAutomaton1.getAlphabet().equals(pAutomaton2.getAlphabet())) return false;
    if (pAutomaton1.getStates().size()!=pAutomaton2.getStates().size()) return false;
    if (pAutomaton1.getAcceptStates().size()!=pAutomaton2.getAcceptStates().size()) return false;
    int size=pAutomaton1.getStates().size();
    IMap<V,W> map=new GHashTableMap<V,W>(size);
    ISet<V> visited=new GHashTableSet<V>(size);
    IQueue<V> queue=new GArrayQueue<V>(size/2);
    visited.add(pAutomaton1.getInitialState());
    queue.enqueue(pAutomaton1.getInitialState());
    map.put(pAutomaton1.getInitialState(),pAutomaton2.getInitialState());
    while (!queue.isEmpty()) {
      V input1=queue.dequeue();
      W input2=map.get(input1);
      if (pAutomaton1.isAcceptState(input1)!=pAutomaton2.isAcceptState(input2)) {
        return false;
      }
      for (char symbol:pAutomaton1.getAlphabet()) {
        V output1=pAutomaton1.getDelta(input1,symbol);
        W output2=pAutomaton2.getDelta(input2,symbol);
        if (!map.containsKey(output1)) {
          map.put(output1,output2);
        }
        else if (!GToolkit.equals(map.get(output1),output2)) {
          return false;
        }
        if (!visited.contains(output1)) {
          visited.add(output1);
          queue.enqueue(output1);
        }
      }
    }
    return true;
  }
  
  public static <V,W> boolean areEquivalent(IAutomaton<V> pAutomaton1, IAutomaton<W> pAutomaton2) {
    checkPrecondition(pAutomaton1,pAutomaton2);
    return areEquivalentAuxiliaryMethod(pAutomaton1.minimize(),pAutomaton2.minimize());
  }
  
  public static <V,W> boolean areDisjoint(IAutomaton<V> pAutomaton1, IAutomaton<W> pAutomaton2) {
    checkPrecondition(pAutomaton1,pAutomaton2);
    return intersection(pAutomaton1,pAutomaton2).acceptsNoStrings();
  }
  
  public static <V,W> boolean isSubset(IAutomaton<V> pAutomaton1, IAutomaton<W> pAutomaton2) {
    checkPrecondition(pAutomaton1,pAutomaton2);
    return areEquivalent(intersection(pAutomaton1,pAutomaton2),pAutomaton1);
  }
  
  private static void checkPrecondition(IAutomaton<?>...pAutomata) {
    for (IAutomaton<?> automaton:pAutomata) {
      if (automaton instanceof IPushdownAutomaton) throw new UnsupportedOperationException("The operation is not implemented over pushdown automata.");
    }
  }
  
}
