package gold.visualization.automaton;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.concurrent.*;
import javax.swing.*;
import javax.swing.event.*;
import gold.structures.automaton.*;
import gold.structures.graph.*;
import gold.structures.list.*;
import gold.structures.set.*;
import gold.structures.stack.*;
import gold.structures.tuple.*;
import gold.swing.util.*;
import gold.util.*;
import gold.visualization.util.*;
import edu.uci.ics.jung.algorithms.layout.*;
import edu.uci.ics.jung.visualization.*;

@SuppressWarnings("serial")
public class GAutomatonInternalFrame extends JInternalFrame implements Comparable<GAutomatonInternalFrame> {
  
  protected GAutomatonPainter<?> painter=null;
  protected File file=null;
  
  public GAutomatonInternalFrame(String pTitle, IAutomaton<?> pAutomaton) {
    this(pTitle,pAutomaton,null);
  }
  
  public GAutomatonInternalFrame(String pTitle, IAutomaton<?> pAutomaton, GLayoutConfiguration pLayoutConfiguration) {
    super(GAutomata.type(pAutomaton)+" - "+pTitle,true,true,true,true);
    setPreferredSize(new Dimension(600,400));
    setSize(new Dimension(600,400));
    pack();
    setAutomaton(pAutomaton,pLayoutConfiguration);
  }
  
  public IAutomaton<?> getAutomaton() {
    return painter.getAutomaton();
  }
  
  public GAutomatonPainter<?> getPainter() {
    return painter;
  }
  
  public File getFile() {
    return file;
  }
  
  public <V> void setAutomaton(IAutomaton<V> pAutomaton, GLayoutConfiguration pLayoutConfiguration) {
    GAutomatonPainter<?> oldPainter=painter;
    GAutomatonPainter<V> newPainter=new GAutomatonPainter<V>(pAutomaton,getSize());
    if (oldPainter!=null) {
      getContentPane().remove(oldPainter.getCanvas());
      copyLayout(oldPainter,newPainter);
    }
    if (pLayoutConfiguration!=null&&!pLayoutConfiguration.isEmpty()) {
      newPainter.setLayoutConfiguration(pLayoutConfiguration);
    }
    getContentPane().add(newPainter.getCanvas(),BorderLayout.CENTER);
    painter=newPainter;
    revalidate();
    repaint();
  }
  
  private <U,V> void copyLayout(GAutomatonPainter<U> pOldPainter, GAutomatonPainter<V> pNewPainter) {
    pNewPainter.setLayoutType(pOldPainter.getLayoutType());
    pNewPainter.setMode(pOldPainter.getMode());
    Layout<U,?> oldLayout=pOldPainter.getLayout();
    Layout<V,?> newLayout=pNewPainter.getLayout();
    ISet<V> newStates=pNewPainter.getAutomaton().getStates();
    Map<String,Point2D> map=new HashMap<String,Point2D>();
    for (U vertex:oldLayout.getGraph().getVertices()) {
      Point2D point=vertex!=null?oldLayout.transform(vertex):null;
      if (vertex!=null&&point!=null) map.put(vertex.toString(),point);
    }
    MultiLayerTransformer oldTransformer=pOldPainter.getRenderContext().getMultiLayerTransformer();
    for (V state:newStates) {
      try {
        String string=state.toString();
        if (map.containsKey(string)) newLayout.setLocation(state,oldTransformer.transform(Layer.LAYOUT,map.get(string)));
      }
      catch (Throwable th) {
      }
    }
  }
  
  public void setFile(File pFile) {
    file=pFile;
    setTitle(GAutomata.type(getAutomaton())+" - "+(file!=null?file.getAbsolutePath():"Untitled automaton"));
  }
  
  public void save(File pFile) {
    try {
      final File newFile=pFile;
      GWaitDialog.wait(this,new Callable<Void>() {
        @Override
        public Void call() throws Exception {
          GAutomata.write(new FileOutputStream(newFile),getPainter());
          return null;
        }
      });
      setFile(newFile);
      GOptionDialog.showInfo(this,"The automaton was saved successfully.");
    }
    catch (Throwable th) {
      GOptionDialog.showError(this,th);
    }
  }
  
  public void displayBasicInformationDialog() {
    if (painter.getAutomaton() instanceof IPushdownAutomaton) throw new UnsupportedOperationException("The operation is not available over pushdown automata.");
    new GAutomatonBasicInformationDialog().display();
  }
  
  public void displayAcceptedRejectedDialog() {
    if (painter.getAutomaton() instanceof IPushdownAutomaton) throw new UnsupportedOperationException("The operation is not available over pushdown automata.");
    new GAutomatonAcceptedRejectedDialog().display();
  }
  
  public void displayRunDialog() {
    displayRunDialog(painter);
  }
  
  private <V> void displayRunDialog(GAutomatonPainter<V> pPainter) {
    new GAutomatonRunDialog<V>(pPainter).display();
  }
  
  public Object displayOperationsDialog(IList<GAutomatonInternalFrame> pInternalFrames) {
    return new GAutomatonOperationsDialog().display(pInternalFrames);
  }
  
  @Override
  public int compareTo(GAutomatonInternalFrame pInternalFrame) {
    return GUtilities.DEFAULT_COLLATOR.compare(getTitle(),pInternalFrame.getTitle());
  }
  
  @Override
  public String toString() {
    return getTitle();
  }
  
  private static Method findMethod(String pMethodSignature) {
    int q=pMethodSignature.lastIndexOf('('),p=pMethodSignature.substring(0,q).lastIndexOf(' ');
    return GReflection.$searchMethod(IAutomaton.class,pMethodSignature.substring(p,q).trim(),false,(Class<?>[])null);
  }
  
  private final class GAutomatonBasicInformationDialog extends JDialog {
    
    private final JEditorPane informationLabel=new GEditorPane("",false);
    
    public GAutomatonBasicInformationDialog() {
      super(GUtilities.getWindowForComponent(GAutomatonInternalFrame.this),"Automaton basic information",JDialog.ModalityType.TOOLKIT_MODAL);
      JButton closeButton=new JButton("Close");
      closeButton.addActionListener(new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent pEvent) {
          GAutomatonBasicInformationDialog.this.dispose();
        }
      });
      informationLabel.setBackground(Color.WHITE);
      informationLabel.setFont(new Font(Font.MONOSPACED,Font.PLAIN,14));
      JScrollPane scrollPane=GUtilities.getScrollPane(informationLabel,ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS,ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED,800,300);
      GUtilities.configure(this,GUtilities.getBorderPane(5,5,0,8,null,null,GUtilities.getBoxPaneY(0,scrollPane,Box.createVerticalStrut(5)),null,GUtilities.getFlowPane(FlowLayout.RIGHT,closeButton)));
      setResizable(true);
    }
    
    public void display() {
      IList<IPair<String>> methods=new GArrayList<IPair<String>>(10);
      methods.addLast(new GPair<String>("public String toString()","Properties"));
      methods.addLast(new GPair<String>("public ISet<V> getReachableStates()","Reachable states"));
      methods.addLast(new GPair<String>("public ISet<V> getLiveStates()","Live states"));
      methods.addLast(new GPair<String>("public ISet<V> getUsefulStates()","Useful states"));
      methods.addLast(new GPair<String>("public boolean isDeterministic()","¿The automaton is deterministic?"));
      methods.addLast(new GPair<String>("public boolean acceptsNoStrings()","¿The automaton accepts no strings?"));
      methods.addLast(new GPair<String>("public boolean acceptsAllStrings()","¿The automaton accepts all strings?"));
      methods.addLast(new GPair<String>("public boolean acceptsEmptyString()","¿The automaton accepts the empty string?"));
      StringBuilder sb=new StringBuilder(4096);
      sb.append("<html>");
      for (IPair<String> method:methods) {
        try {
          sb.append("<font color='#008000'><b>"+GUtilities.escapeHTML(method.getSecond())+":</b></font><br/>");
          sb.append(GUtilities.escapeHTML((""+GAutomatonInternalFrame.findMethod(method.getFirst()).invoke(painter.getAutomaton())).trim()));
        }
        catch (Throwable th) {
          sb.append("<font color='#FF0000'>"+GUtilities.escapeHTML((th instanceof InvocationTargetException)?th.getCause().toString():th.toString())+"</font>");
        }
        sb.append("<br/><br/>");
      }
      sb.append("</html>");
      informationLabel.setText(sb.toString());
      informationLabel.setCaretPosition(0);
      GUtilities.show(this,getOwner());
    }
    
  }
  
  private final class GAutomatonAcceptedRejectedDialog extends JDialog {
    
    private final JLabel acceptedLabel=new JLabel();
    private final JLabel rejectedLabel=new JLabel();
    private final JList acceptedList=new JList();
    private final JList rejectedList=new JList();
    private final JSpinner lengthSpinner=new JSpinner(new SpinnerNumberModel(2,0,8,1));
    
    public GAutomatonAcceptedRejectedDialog() {
      super(GUtilities.getWindowForComponent(GAutomatonInternalFrame.this),"Automaton accepted/rejected strings",JDialog.ModalityType.TOOLKIT_MODAL);
      JButton closeButton=new JButton("Close");
      closeButton.addActionListener(new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent pEvent) {
          GAutomatonAcceptedRejectedDialog.this.dispose();
        }
      });
      lengthSpinner.addChangeListener(new ChangeListener() {
        @Override
        public void stateChanged(ChangeEvent pEvent) {
          updateInformation();
        }
      });
      JScrollPane scrollPaneAccepted=GUtilities.getScrollPane(acceptedList,ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS,ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED,250,400);
      JScrollPane scrollPaneRejected=GUtilities.getScrollPane(rejectedList,ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS,ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED,250,400);
      JPanel panelAccepted=GUtilities.getBorderPane(acceptedLabel,null,scrollPaneAccepted,null,null);
      JPanel panelRejected=GUtilities.getBorderPane(rejectedLabel,null,scrollPaneRejected,null,null);
      JPanel panelLength=GUtilities.getBoxPaneY(0,GUtilities.getFlowPane(FlowLayout.LEFT,new JLabel("Maximum length: "),lengthSpinner),Box.createVerticalStrut(5));
      GUtilities.configure(this,GUtilities.getBorderPane(5,5,0,8,panelLength,panelAccepted,Box.createHorizontalStrut(10),panelRejected,GUtilities.getBoxPaneY(0,Box.createVerticalStrut(5),GUtilities.getFlowPane(FlowLayout.RIGHT,closeButton))));
      setResizable(true);
    }
    
    private void configure(Iterable<?> pItems, JList pList, String pTitle, JLabel pLabel) {
      DefaultListModel listModel=new DefaultListModel();
      for (Object item:pItems) {
        listModel.addElement("<html>"+GUtilities.escapeHTML("\""+item+"\"")+"</html>");
      }
      pList.setModel(listModel);
      pLabel.setText(pTitle+" ("+listModel.size()+"):");
    }
    
    private void updateInformation() {
      int n=((Number)(lengthSpinner.getValue())).intValue();
      configure(painter.getAutomaton().getAcceptedStrings(n),acceptedList,"Accepted strings",acceptedLabel);
      configure(painter.getAutomaton().getRejectedStrings(n),rejectedList,"Rejected strings",rejectedLabel);
    }
    
    public void display() {
      updateInformation();
      GUtilities.show(this,getOwner());
    }
    
  }
  
  private final class GAutomatonRunDialog<V> extends JDialog {
    
    private final GAutomatonPainter<V> painter;
    private final IAutomaton<V> automaton;
    private final GTapeLabel inputTape=new GTapeLabel("Input string");
    private final GTapeLabel outputTape=new GTapeLabel("Output string");
    private final GTapeLabel stackTape=new GTapeLabel("Stack content");
    private final JTextField stringField=new JTextField(16);
    private final JLabel progressLabel=new JLabel("<html><i>pending ...</i></html>");
    private final JButton firstButton=new JButton(GUtilities.getIcon("jlfgr/Rewind16.gif"));
    private final JButton backButton=new JButton(GUtilities.getIcon("jlfgr/Back16.gif"));
    private final JButton nextButton=new JButton(GUtilities.getIcon("jlfgr/Forward16.gif"));
    private final JButton lastButton=new JButton(GUtilities.getIcon("jlfgr/FastForward16.gif"));
    private final IStack<GAutomatonRunConfiguration> configurations=new GArrayStack<GAutomatonRunConfiguration>();
    private final IList<IPushdownConfiguration<V>> pushdownTrace=new GArrayList<IPushdownConfiguration<V>>();
    private boolean disabled=true;
    
    public GAutomatonRunDialog(GAutomatonPainter<V> pPainter) {
      super(GUtilities.getWindowForComponent(GAutomatonInternalFrame.this),"Automaton execution",JDialog.ModalityType.TOOLKIT_MODAL);
      painter=pPainter;
      automaton=painter.getAutomaton();
      stringField.setFont(stringField.getFont().deriveFont(16f));
      JButton closeButton=new JButton("Close");
      closeButton.addActionListener(new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent arg0) {
          GAutomatonRunDialog.this.dispose();
        }
      });
      firstButton.addActionListener(new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent pEvent) {
          animate(0);
        }
      });
      backButton.addActionListener(new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent pEvent) {
          animate(configurations.isEmpty()?0:configurations.peek().step-1);
        }
      });
      nextButton.addActionListener(new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent pEvent) {
          if (disabled) {
            disabled=false;
            configure();
          }
          else {
            animate(configurations.isEmpty()?1:configurations.peek().step+1);
          }
        }
      });
      lastButton.addActionListener(new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent pEvent) {
          if (disabled) {
            disabled=false;
            if (!configure()) return;
          }
          if (!(getAutomaton() instanceof IPushdownAutomaton)) {
            animate(inputTape.maxStep());
          }
          else {
            animate(pushdownTrace.size());
          }
          stringField.requestFocus();
        }
      });
      stringField.getDocument().addDocumentListener(new DocumentListener() {
        @Override
        public void changedUpdate(DocumentEvent pEvent) {
          disabled=true;
          configure();
        }
        @Override
        public void insertUpdate(DocumentEvent pEvent) {
          disabled=true;
          configure();
        }
        @Override
        public void removeUpdate(DocumentEvent pEvent) {
          disabled=true;
          configure();
        }
      });
      JLabel titleLabels[]={new JLabel("Input string: "),new JLabel("Input tape: "),new JLabel("Output tape: "),new JLabel("Stack tape: "),new JLabel("Progress: ")};
      GUtilities.resize(new Dimension(21,21),firstButton,backButton,nextButton,lastButton);
      for (JLabel label:titleLabels) {
        GUtilities.resize(new Dimension(95,21),label);
        label.setHorizontalAlignment(SwingConstants.RIGHT);
      }
      JPanel panelParameters=GUtilities.getBorderPane(null,titleLabels[0],stringField,GUtilities.getBoxPaneX(0,Box.createHorizontalStrut(8),firstButton,backButton,nextButton,lastButton),null);
      JPanel panelInputTape=GUtilities.getBorderPane(5,0,0,0,null,titleLabels[1],inputTape,null,null);
      JPanel panelOutputTape=GUtilities.getBorderPane(5,0,0,0,null,titleLabels[2],outputTape,null,null);
      JPanel panelStackTape=GUtilities.getBorderPane(5,0,0,0,null,titleLabels[3],stackTape,null,null);
      JPanel panelProgress=GUtilities.getBorderPane(5,0,5,0,null,titleLabels[4],progressLabel,null,null);
      JPanel panelLabels=GUtilities.getBoxPaneY(5,panelInputTape,panelOutputTape,panelStackTape);
      JPanel panelCenter=GUtilities.getBorderPane(5,0,0,0,null,null,panelLabels,null,panelProgress);
      GUtilities.configure(this,GUtilities.getBorderPane(5,5,0,8,panelParameters,null,panelCenter,null,GUtilities.getFlowPane(FlowLayout.RIGHT,closeButton)));
      setResizable(true);
      configure();
    }
    
    public void clean() {
      painter.getVertexFillPaintTransformer().clearAll();
      painter.getEdgeDrawPaintTransformer().clearAll();
      painter.getArrowFillPaintTransformer().clearAll();
      painter.clearSelection();
    }
    
    public void display() {
      clean();
      GUtilities.show(this,getOwner());
      clean();
      painter.repaint();
    }
    
    private boolean configure() {
      if (disabled) {
        GUtilities.enable(false,firstButton,backButton);
        GUtilities.enable(true,nextButton,lastButton);
        setInputString("");
        return false;
      }
      else {
        String string=stringField.getText();
        try {
          GUtilities.enable(true,firstButton,backButton,nextButton,lastButton);
          if (string.indexOf(IAutomaton.LAMBDA)!=-1) {
            throw new IllegalArgumentException("The input string cannot have lambda symbols.");
          }
          else if (!automaton.getAlphabet().containsAll(string)) {
            Set<Character> set=new TreeSet<Character>(Arrays.asList(GArrays.toObjectArray(string.toCharArray())));
            set.removeAll(GCollections.asJavaCollection(automaton.getAlphabet()));
            if (set.size()==1) throw new IllegalArgumentException("The character '"+set.iterator().next()+"' isn't a symbol of the automaton's alphabet.");
            throw new IllegalArgumentException("The characters '"+GToolkit.separate("','",set.toArray())+"' aren't symbols of the automaton's alphabet.");
          }
          setInputString(string);
          animate(0);
          return true;
        }
        catch (Throwable th) {
          setInputString("");
          stringField.requestFocus();
          GUtilities.enable(false,firstButton,backButton,nextButton,lastButton);
          progressLabel.setText("<html><i><font color=red>"+th.getMessage()+"</font></i></html>");
          return false;
        }
      }
    }
    
    private void setInputString(String pInputString) {
      inputTape.string=pInputString;
      outputTape.string="";
      stackTape.string="";
      inputTape.head=0;
      outputTape.head=0;
      stackTape.head=0;
      configurations.clear();
      progressLabel.setText("<html><i><font color=\"#808080\">Click the &gt; button to begin the simulation.</font></i></html>");
      painter.repaint();
      inputTape.repaint();
      outputTape.repaint();
      clean();
    }
    
    private void animate(int pHead) {
      ITransducer<V> transducer=(automaton instanceof ITransducer)?(ITransducer<V>)automaton:null;
      IPushdownAutomaton<V> pushdownAutomaton=(automaton instanceof IPushdownAutomaton)?(IPushdownAutomaton<V>)automaton:null;
      if (pushdownAutomaton!=null) {
        if (pHead<=0) {
          pushdownTrace.clear();
          for (IPushdownConfiguration<V> c=pushdownAutomaton.trace(inputTape.string); c!=null; c=c.getPreviousConfiguration()) {
            pushdownTrace.addLast(c);
          }
          pushdownTrace.reverse();
          if (pushdownTrace.isEmpty()) pushdownTrace.addLast(null);
        }
        int step=Math.max(0,Math.min(pushdownTrace.size(),pHead));
        while (!configurations.isEmpty()&&configurations.peek().step>step) {
          configurations.pop();
        }
        while ((step>0&&configurations.isEmpty())||(!configurations.isEmpty()&&configurations.peek().step<step)) {
          if (configurations.isEmpty()) {
            IPushdownConfiguration<V> first=!pushdownTrace.isEmpty()?pushdownTrace.getFirst():null;
            configurations.push(new GAutomatonRunConfiguration(1,first!=null?GCollections.singletonSet(first.getState()):GCollections.<V>emptySet(),first!=null?first.getInputString():"","",first!=null?first.getStack():""));
          }
          else {
            IPushdownConfiguration<V> current=step-1<pushdownTrace.size()?pushdownTrace.get(step-1):null;
            configurations.push(new GAutomatonRunConfiguration(configurations.peek().step+1,current!=null?GCollections.singletonSet(current.getState()):GCollections.<V>emptySet(),current!=null?current.getInputString():"","",current!=null?current.getStack():""));
          }
        }
      }
      else {
        int head=Math.max(0,Math.min(inputTape.maxStep(),pHead));
        for (; inputTape.head>head; inputTape.head--) {
          configurations.pop();
        }
        for (; inputTape.head<head; inputTape.head++) {
          if (configurations.isEmpty()) {
            ISet<V> states=automaton.applyLambdasRecursively(GCollections.singletonSet(automaton.getInitialState()));
            String output=transducer!=null?transducer.getStateOutput(automaton.getInitialState()):"";
            configurations.push(new GAutomatonRunConfiguration(1,states,"",output,""));
          }
          else {
            GAutomatonRunConfiguration c=configurations.peek();
            ISet<V> states=automaton.applyAction(c.states,inputTape.currentSymbol());
            String output=transducer!=null?c.output+transducer.getTransitionOutput(c.states.pick(),inputTape.currentSymbol())+transducer.getStateOutput(states.pick()):"";
            configurations.push(new GAutomatonRunConfiguration(c.step+1,states,"",output,""));
          }
        }
      }
      clean();
      ITransformer<V,Paint> vertexFillPaintTransformer=painter.getVertexFillPaintTransformer();
      ITransformer<IEdge<V,ISet<?>>,Paint> edgeDrawPaintTransformer=painter.getEdgeDrawPaintTransformer();
      ITransformer<IEdge<V,ISet<?>>,Paint> edgeArrowFillPaintTransformer=painter.getArrowFillPaintTransformer();
      if (!configurations.isEmpty()) {
        IPushdownConfiguration<V> pushdownConfiguration=pushdownAutomaton!=null&&configurations.peek().step<pushdownTrace.size()?pushdownTrace.get(configurations.peek().step):null;
        Object action=pushdownAutomaton==null?inputTape.currentSymbol():(pushdownConfiguration!=null?pushdownConfiguration.getTransition():null);
        vertexFillPaintTransformer.setAll(configurations.peek().states,Color.CYAN);
        for (V state:configurations.peek().states) {
          for (IEdge<V,ISet<?>> edge:painter.getGraph().getOutEdges(state)) {
            if (action!=null&&edge.getData().contains(action)) {
              edgeDrawPaintTransformer.set(edge,Color.CYAN);
              edgeArrowFillPaintTransformer.set(edge,Color.CYAN);
            }
          }
        }
        if (pushdownAutomaton!=null) inputTape.head=inputTape.string.length()-configurations.peek().fragment.length()+1;
        outputTape.string=configurations.peek().output;
        outputTape.head=outputTape.maxStep();
        stackTape.string=configurations.peek().stack;
        stackTape.head=stackTape.maxStep();
      }
      else if (pushdownAutomaton!=null) {
        inputTape.head=0;
        outputTape.head=0;
        stackTape.head=0;
      }
      else {
        outputTape.head=0;
        stackTape.head=0;
      }
      if (inputTape.head==inputTape.maxStep()&&(pushdownAutomaton==null||(!configurations.isEmpty()&&configurations.peek().step==pushdownTrace.size()))) {
        boolean accepted=!configurations.isEmpty()&&!automaton.getAcceptStates().isDisjointFrom(configurations.peek().states);
        String additionalText=accepted?(pushdownAutomaton!=null?" (string processed, final state reached, empty stack)":" (string processed, final state reached)"):"";
        progressLabel.setText("<html><i>"+(accepted?"<font color=green>The input string was accepted"+additionalText+"</font>":"<font color=red>The input string was rejected</font>")+"</i></html>");
      }
      else {
        progressLabel.setText("<html><i>Performing the step #"+pHead+" ...</i></html>");
      }
      if (pushdownAutomaton==null) {
        GUtilities.enable(inputTape.head>inputTape.minStep(),firstButton,backButton);
        GUtilities.enable(inputTape.head<inputTape.maxStep(),nextButton,lastButton);
      }
      else {
        GUtilities.enable(!pushdownTrace.isEmpty()&&!configurations.isEmpty(),firstButton,backButton);
        GUtilities.enable(!pushdownTrace.isEmpty()&&(configurations.isEmpty()||configurations.peek().step<pushdownTrace.size()),nextButton,lastButton);
      }
      painter.repaint();
      inputTape.repaint();
      outputTape.repaint();
    }
    
    private class GTapeLabel extends JLabel {
      
      private String string="";
      private int head=0;
      
      public GTapeLabel(String pTitle) {
        final String title=pTitle;
        setPreferredSize(new Dimension(600,30));
        setToolTipText("Click to copy text");
        addMouseListener(new MouseAdapter() {
          @Override
          public void mouseClicked(MouseEvent pEvent) {
            JEditorPane editorPane=new JEditorPane("text/plain",string);
            editorPane.setEditable(false);
            editorPane.setFont(editorPane.getFont().deriveFont(18f));
            JScrollPane scrollPane=new JScrollPane(editorPane,JScrollPane.VERTICAL_SCROLLBAR_NEVER,JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
            scrollPane.setPreferredSize(new Dimension(500,45));
            // GOptionDialog.showInfo(GTapeLabel.this,scrollPane);
            JOptionPane.showMessageDialog(GTapeLabel.this,scrollPane,title,JOptionPane.INFORMATION_MESSAGE);
          }
        });
      }
      
      public int minStep() {
        return 0;
      }
      
      public int maxStep() {
        return string.length()+1;
      }
      
      public char currentSymbol() {
        return symbolAt(head);
      }
      
      public char symbolAt(int pIndex) {
        return pIndex<=0?'\u22B3':(pIndex>string.length()?'\u2423':string.charAt(pIndex-1));
      }
      
      @Override
      public void paint(Graphics pGraphics) {
        super.paint(pGraphics);
        int B=8,Z=getHeight()-1-B;
        setFont(new Font(Font.MONOSPACED,Font.PLAIN,Z*8/10));
        Color background=GAutomatonRunDialog.this.getBackground();
        Graphics2D g=(Graphics2D)pGraphics;
        FontMetrics fontMetrics=g.getFontMetrics();
        for (int k=0,u=getWidth()/Z; k<=u; k++) {
          g.setPaint(k==head?Color.CYAN:(k<=string.length()?(k<head?new Color(255,255,230):new Color(190,190,190)):background));
          g.fillRect(k*Z,0,Z,Z);
          g.setPaint(Color.BLACK);
          g.drawRect(k*Z,0,Z,Z);
          String s=String.valueOf(symbolAt(k));
          g.drawString(s,(float)(k*Z+(Z-fontMetrics.stringWidth(s))/2),(float)((Z+fontMetrics.getAscent()-fontMetrics.getDescent())/2));
        }
        double x=head*Z+Z*0.5,y=Z;
        g.draw(GShapes.polygonalChain(GShapes.points(x-3,y+4,x,y,x+3,y+4,x,y,x,y+B)));
      }
      
    }
    
    private class GAutomatonRunConfiguration {
      
      private int step;
      private ISet<V> states;
      private String fragment;
      private String output;
      private String stack;
      
      public GAutomatonRunConfiguration(int pStep, ISet<V> pStates, String pFragment, String pOutput, String pStack) {
        step=pStep;
        states=pStates;
        fragment=pFragment;
        output=pOutput;
        stack=pStack;
      }
      
    }
    
  }
  
  private final class GAutomatonOperationsDialog extends JDialog {
    
    private final JComboBox operand1ComboBox=new JComboBox();
    private final JComboBox operatorComboBox=new JComboBox();
    private final JComboBox operand2ComboBox=new JComboBox();
    private Object methodResult=null;
    
    public GAutomatonOperationsDialog() {
      super(GUtilities.getWindowForComponent(GAutomatonInternalFrame.this),"Automaton operations",JDialog.ModalityType.TOOLKIT_MODAL);
      JButton acceptButton=new JButton("Accept");
      JButton cancelButton=new JButton("Cancel");
      {
        Vector<GAutomatonOperationItem> operators=new Vector<GAutomatonOperationItem>();
        operators.add(new GAutomatonOperationItem("public IAutomaton<?> simplify()","Simplify"));
        operators.add(new GAutomatonOperationItem("public IAutomaton<?> determinize()","Determinize"));
        operators.add(new GAutomatonOperationItem("public IAutomaton<?> minimize()","Minimize"));
        operators.add(new GAutomatonOperationItem("public IAutomaton<?> union(IAutomaton<?> pAutomaton)","Union"));
        operators.add(new GAutomatonOperationItem("public IAutomaton<?> intersection(IAutomaton<?> pAutomaton)","Intersection"));
        operators.add(new GAutomatonOperationItem("public IAutomaton<?> concatenation(IAutomaton<?> pAutomaton)","Concatenation"));
        operators.add(new GAutomatonOperationItem("public IAutomaton<?> difference(IAutomaton<?> pAutomaton)","Difference"));
        operators.add(new GAutomatonOperationItem("public IAutomaton<?> transpose()","Transpose"));
        operators.add(new GAutomatonOperationItem("public IAutomaton<?> complement()","Complement"));
        operators.add(new GAutomatonOperationItem("public IAutomaton<?> optionality()","Optionality"));
        operators.add(new GAutomatonOperationItem("public IAutomaton<?> kleenePlus()","Kleene plus"));
        operators.add(new GAutomatonOperationItem("public IAutomaton<?> kleeneStar()","Kleene star"));
        operators.add(new GAutomatonOperationItem("public boolean isEquivalentTo(IAutomaton<?> pAutomaton)","Is equivalent to"));
        operators.add(new GAutomatonOperationItem("public boolean isDisjointFrom(IAutomaton<?> pAutomaton)","Is disjoint from"));
        operators.add(new GAutomatonOperationItem("public boolean isSubset(IAutomaton<?> pAutomaton)","Is subset"));
        operatorComboBox.setModel(new DefaultComboBoxModel(operators));
        operatorComboBox.addActionListener(new ActionListener() {
          @Override
          public void actionPerformed(ActionEvent pEvent) {
            GAutomatonOperationItem operator=(GAutomatonOperationItem)(operatorComboBox.getSelectedItem());
            operand2ComboBox.setEnabled(operator.methodReference.getParameterTypes().length==1);
          }
        });
        operatorComboBox.setSelectedIndex(0);
      }
      {
        acceptButton.addActionListener(new ActionListener() {
          @Override
          public void actionPerformed(ActionEvent pEvent) {
            GAutomatonOperationsDialog dialog=GAutomatonOperationsDialog.this;
            try {
              GWaitDialog.wait(GAutomatonOperationsDialog.this,new Callable<Void>() {
                @Override
                public Void call() throws Exception {
                  GAutomatonInternalFrame operand1=(GAutomatonInternalFrame)(operand1ComboBox.getSelectedItem());
                  GAutomatonOperationItem operator=(GAutomatonOperationItem)(operatorComboBox.getSelectedItem());
                  GAutomatonInternalFrame operand2=(GAutomatonInternalFrame)(operand2ComboBox.getSelectedItem());
                  methodResult=null;
                  if (operator.methodReference.getParameterTypes().length==1) {
                    methodResult=operator.methodReference.invoke(operand1.getAutomaton(),operand2.getAutomaton());
                  }
                  else {
                    methodResult=operator.methodReference.invoke(operand1.getAutomaton());
                  }
                  return null;
                }
              });
              dialog.dispose();
            }
            catch (InvocationTargetException ex) {
              GOptionDialog.showError(dialog,ex.getCause());
            }
            catch (Throwable th) {
              GOptionDialog.showError(dialog,th);
            }
          }
        });
        cancelButton.addActionListener(new ActionListener() {
          @Override
          public void actionPerformed(ActionEvent pEvent) {
            GAutomatonOperationsDialog dialog=GAutomatonOperationsDialog.this;
            dialog.dispose();
          }
        });
        operand1ComboBox.setPreferredSize(new Dimension(650,21));
        operatorComboBox.setPreferredSize(new Dimension(650,21));
        operand2ComboBox.setPreferredSize(new Dimension(650,21));
        JPanel panel=GUtilities.getGridBagPane(new Object[][]{{"Operand 1:",operand1ComboBox},{"Operation:",operatorComboBox},{"Operand 2:",operand2ComboBox}},4,null);
        GUtilities.configure(this,GUtilities.getBorderPane(5,5,0,5,null,null,panel,null,GUtilities.getBoxPaneX(0,Box.createHorizontalGlue(),acceptButton,Box.createHorizontalStrut(5),cancelButton)));
      }
    }
    
    public Object display(IList<GAutomatonInternalFrame> pInternalFrames) {
      if (pInternalFrames.size()==0) throw new IllegalArgumentException("There should be at least one active automaton in the desktop.");
      operand1ComboBox.setModel(new DefaultComboBoxModel(new Vector<GAutomatonInternalFrame>(pInternalFrames.toJavaView())));
      operand2ComboBox.setModel(new DefaultComboBoxModel(new Vector<GAutomatonInternalFrame>(pInternalFrames.toJavaView())));
      operand1ComboBox.setSelectedItem(GAutomatonInternalFrame.this);
      GUtilities.show(this,getOwner());
      return methodResult;
    }
    
    private class GAutomatonOperationItem {
      
      private final Method methodReference;
      private final String methodTitle;
      
      public GAutomatonOperationItem(String pMethodSignature, String pMethodTitle) {
        methodReference=GAutomatonInternalFrame.findMethod(pMethodSignature);
        methodTitle=pMethodTitle;
      }
      
      @Override
      public String toString() {
        return "<html><font color='#008000'>"+GUtilities.escapeHTML(methodTitle)+"</font></html>";
      }
      
    }
    
  }
  
}
