package gold.visualization.util;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import java.util.List;
import gold.util.*;

public final class GShapes {
  
  private GShapes() {
  }
  
  public static Rectangle2D point(double pX, double pY) {
    return rectangle(pX,pY,0,0);
  }
  
  public static Line2D line(double pX1, double pY1, double pX2, double pY2) {
    return new Line2D.Double(pX1,pY1,pX2,pY2);
  }
  
  public static Rectangle2D rectangle(double pX, double pY, double pWidth, double pHeight) {
    return new Rectangle2D.Double(pX,pY,pWidth,pHeight);
  }
  
  public static RoundRectangle2D roundRectangle(double pX, double pY, double pWidth, double pHeight, double pArcWidth, double pArcHeight) {
    return new RoundRectangle2D.Double(pX,pY,pWidth,pHeight,pArcWidth,pArcHeight);
  }
  
  public static Rectangle2D square(double pX, double pY, double pSideLength) {
    return new Rectangle2D.Double(pX,pY,pSideLength,pSideLength);
  }
  
  public static GeneralPath polygon(double pX, double pY, double pRadius, double pAngleOffset, int pSideCount) {
    double angleDelta=Math.PI*2/pSideCount,points[][]=new double[pSideCount][2];
    for (int i=0; i<pSideCount; i++) {
      double angle=Math.PI/2+(i+0.5)*angleDelta+pAngleOffset;
      points[i][0]=pX+pRadius*Math.cos(angle);
      points[i][1]=pY+pRadius*Math.sin(angle);
    }
    return polygon(points);
  }
  
  public static GeneralPath polygon(double[][] pPoints) {
    GeneralPath generalPath=polygonalChain(pPoints);
    generalPath.closePath();
    return generalPath;
  }
  
  public static GeneralPath polygonalChain(double[][] pPoints) {
    GeneralPath generalPath=new GeneralPath(Path2D.WIND_EVEN_ODD);
    generalPath.moveTo(pPoints[0][0],pPoints[0][1]);
    for (int i=1; i<pPoints.length; i++) {
      generalPath.lineTo(pPoints[i][0],pPoints[i][1]);
    }
    return generalPath;
  }
  
  public static QuadCurve2D quadraticCurve(double pStartX, double pStartY, double pControlX, double pControlY, double pEndX, double pEndY) {
    return new QuadCurve2D.Double(pStartX,pStartY,pControlX,pControlY,pEndX,pEndY);
  }
  
  public static CubicCurve2D cubicCurve(double pStartX, double pStartY, double pControlX1, double pControlY1, double pControlX2, double pControlY2, double pEndX, double pEndY) {
    return new CubicCurve2D.Double(pStartX,pStartY,pControlX1,pControlY1,pControlX2,pControlY2,pEndX,pEndY);
  }
  
  public static Ellipse2D ellipse(Shape pBoundingShape) {
    Rectangle2D bounds=pBoundingShape.getBounds2D();
    return new Ellipse2D.Double(bounds.getX(),bounds.getY(),bounds.getWidth(),bounds.getHeight());
  }
  
  public static Ellipse2D ellipse(double pCenterX, double pCenterY, double pRadiusX, double pRadiusY) {
    return new Ellipse2D.Double(pCenterX-pRadiusX,pCenterY-pRadiusY,pRadiusX*2,pRadiusY*2);
  }
  
  public static Ellipse2D circle(double pCenterX, double pCenterY, double pRadius) {
    return ellipse(pCenterX,pCenterY,pRadius,pRadius);
  }
  
  public static Area area(Shape pShape) {
    return (pShape instanceof Area)?(Area)pShape:new Area(pShape);
  }
  
  public static Area union(Shape...pShapes) {
    if (pShapes.length==0) throw new IllegalArgumentException("The number of shapes cannot be zero.");
    Area result=new Area(pShapes[0]);
    for (int i=1; i<pShapes.length; i++) {
      result.add(area(pShapes[i]));
    }
    return result;
  }
  
  public static Area intersection(Shape...pShapes) {
    if (pShapes.length==0) throw new IllegalArgumentException("The number of shapes cannot be zero.");
    Area result=new Area(pShapes[0]);
    for (int i=1; i<pShapes.length; i++) {
      result.intersect(area(pShapes[i]));
    }
    return result;
  }
  
  public static Area difference(Shape pShape1, Shape pShape2) {
    Area result=new Area(pShape1);
    result.subtract(area(pShape2));
    return result;
  }
  
  public static Area symmetricDifference(Shape...pShapes) {
    if (pShapes.length==0) throw new IllegalArgumentException("The number of shapes cannot be zero.");
    Area result=new Area(pShapes[0]);
    for (int i=1; i<pShapes.length; i++) {
      result.exclusiveOr(area(pShapes[i]));
    }
    return result;
  }
  
  public static Shape transform(Shape pShape, AffineTransform pTransform) {
    return pTransform.createTransformedShape(pShape);
  }
  
  public static double[][] points(double...pCoordinates) {
    int t=pCoordinates.length/2;
    double[][] points=new double[t][2];
    for (int i=0; i<t; i++) {
      points[i][0]=pCoordinates[i*2];
      points[i][1]=pCoordinates[i*2+1];
    }
    return points;
  }
  
  public static double[][] points(Iterable<?> pCoordinates) {
    List<double[]> points=new ArrayList<double[]>();
    for (Object object:pCoordinates) {
      Iterator<?> it=GCollections.unmodifiableCollection(object).iterator();
      points.add(new double[]{((Number)(it.next())).doubleValue(),((Number)(it.next())).doubleValue()});
    }
    return points.toArray(new double[0][]);
  }
  
  public static double[][] points(Shape pPath) {
    return points(pPath.getPathIterator(null));
  }
  
  public static double[][] points(PathIterator pPathIterator) {
    List<double[]> points=new ArrayList<double[]>();
    double[] c=new double[6];
    for (PathIterator it=pPathIterator; !it.isDone(); it.next()) {
      if (it.currentSegment(c)!=PathIterator.SEG_CLOSE) points.add(new double[]{c[0],c[1]});
    }
    return points.toArray(new double[0][]);
  }
  
}
